/*
   CREATEREINDEX.PRG
   This program is designed to do what the version that
   Romain once did does -- it creates a program for you
   that will re-build your index tags on your tables --
   but this one uses some special features of Visual dBASE 7
   to do it ...

   This will create a single program file that should regenerate
   your index tags for all tables (after deleting them) -- 
   if you add a table, or add a new index tag, or change an 
   index tag definition, you should re-run this routine!

   // NOTE: If anyone has suggestions for improvements, 
   // feel free to contact me. This is something I threw
   // together in about an hour of tinkering ...

   Programmer: Ken Mayer
   Updates:
   08/23/1998 -- added addition of a try/catch around
                 the DROP INDEX command, to catch
                 an error if the index tag doesn't
                 exist ... this adds heavily to the 
                 length of the code, because the try/catch
                 is created for *each* index tag that
                 we attempt to drop, but it should
                 avoid the user being interrupted
                 by an error that won't stop the
                 program from doing it's job ...

   9/29/1998 -- Gary White added code to handle primary key
                index tags.

   10/6/1998 -- GW added a minor fix ...

   January 26, 1999 -- Added optional database parameter,
                which can be used to pass the name of
                a database to this routine. If used,
                you MUST be sure to have access to 
                PATHDATA.CC (also in the dUFLP library),
                otherwise it won't function.

                The advantage to this is that you can 
                use this routine to generate the 
                correct database object and references
                so that it will use that information,
                rather than the default database
                (_app.databases[1]), and will therefore
                find tables stored in a database
                folder other than the current directory.

   February 1, 1999 -- minor modification -- this will
                make sure that the new program created
                by this one will close any open databases 
                before it tries to do anything -- otherwise
                the user will get "file in use" errors.

   March/April 1999 -- Gary White added more code, and
                program renamed to "CreateReindex"
                No longer requires the use of "PathData.cc"
                as well ...

   5/24/1999 -- GW added small fix to primary keys where
   				 the primary key was a single field with
                a space in the field name.

   June 23, 1999 -- Lawrence Bain modified to accept an array 
                of database names as a parameter.

   April 2, 2003 -- Geoff Wass and Ivar Jessen ran into some
                difficulties with string delimiters in expressions
                and forExpressions for index tags, which we have
                attempted to deal with here. If an index tag
                has an expression or forExpression using all three 
                string delimiters: ', ", or [], this code will
                not generate code to generate that index tag. 

                In addition, it will attempt to generate a string
                with an appropriate set of delimiters to avoid 
                problems where in the past the code here assumed
                you were not using ' as a delimiter in your index
                expression and/or forExpression ...

   July, 2005 -- Ivar B. Jessen noted two things:
                * A table can have two index tags that
                  have the exact same expression. The 
                  problem then becomes if one of them
                  is the Primary key. The old code would
                  attempt to create two primary keys in
                  this instance. Changing this program
                  to use BDEIndex.cc appears to resolve
                  the problem, as this class uses the
                  BDE API to determine if the tag is
                  the primary key (the tableDef object
                  does not allow this).

                * The second problem is that if a corrupt
                  index tag exists, the old code would 
                  have heartburn when deleting indexes.
                  This code now uses dBF7File, and the
                  SetMDXFlag method to turn off the byte
                  in the table header that determines if
                  the table has indexes, then we delete
                  the .mdx file. Much more efficient ...

   August, 2005 -- Fixed a bug found by Howard Mintzer, and 
                  dealt with hard-coded path when deleting
                  the .MDX file ... can't have that in the
                  final program of course, because the
                  path may change ...

   November, 2006 -- added code to call _app.executeMessages()
                  if using 2.5 or later ...

   August, 2007  -- Several updates:
                 * Added a bit more code to insert some blanks
                   between tables and between indexes to make
                   individual tables and indexes easier to find,
                   and make the code generated more readable.
                 * Added a form that is part of the code. It's
                   based on :dUFLP:Message.wfm, but adds a 
                   progress bar.
                 * Added the use of a session object, and
                   the ability to display some progress
                   on a progress bar ... useful for those
                   large tables where the indexes take awhile
                   to re-create.

   Usage:
         set procedure to :dUFLP:CreateReindex.prg additive
         CreateReindex()
            or
         do :dUFLP:CreateReindex

         ** OR:

         set procedure to :dUFLP:CreateReindex.prg additive
         CreateReindex( "databaseName")
            or
         do :dUFLP:CreateReindex with "databaseName"
            // where "databaseName" is the BDE Alias
            // where the tables are stored

         ** OR:

         set procedure to :dUFLP:CreateReindex.prg additive
         CreateReindex( arrayOfDatabaseNames )
            or
         do :dUFLP:CreateReindex"with arrayOfDatabaseName

   DEPENDENCIES:
      :dUFLP:BDEIndex.cc -- used to obtain detailed information
      about the indexes in the table, this is vital for the
      determination of the primary key, which is not really possible
      through the TableDef object as we were doing before. The problem
      with the TableDef is that the primaryKey expression can be
      obtained, but it is possible to have two index tags with the
      same expression, but one of them be a primary key. The code
      that was being used before would attempt to create BOTH
      tags as primary keys, which is not possible ...

   ** Note that this program uses Source Aliasing, specifically the
   ** dUFLP source Alias. If you do not know how to set up
   ** a Source Alias, see the instructions in WHATS.NEW at
   ** the top of the file, or in README.TXT.
   ** Any files referenced by :dUFLP:filename  must be
   ** included in an executable if you are building 
   ** one ... make sure that your project includes these
   ** files.
   **
   ** This particular program uses:
   **     :dUFLP:message.wfm
   **     :dUFLP:BDEIndex.cc
   **     :dUFLP:dBF7File.cc
   **     :dUFLP:BFile.cc
   **     :dUFLP:BDEAlias.cc

*/

parameter aDatabase  // changed to array of databases 6/22/99

   //new section to handle character rather than array parameter 6/22/99
   if type("aDatabase")=="C"
      local cTmp
      cTmp=aDatabase
      aDatabase=new array()
      aDatabase.add(cTmp)
   endif

   close database // just to be sure no tables or 
                  // databases are open ...

   // Need this array:
   aTables = new Array()
   // we'll be checking this throughout ...:
   lDatabase = false

   if argCount() > 0
      for x=1 to aDatabase.size	// added loop for number of databases 6/22/99
         // if we're using a database, we need to get the
         // table names into an array ...:
         lDatabase := true
         try
            dMyData = new database()
            dMyData.databaseName := aDatabase[x]//changed to array 6/22/99
            dMyData.active := true
         catch (exception e)
            msgbox( "BDE Alias error: "+e.code+", "+e.message, ;
                    "Problem!", 16 )
            dMyData.active := false
            return
         endtry
         aTables = dMyData.getSchema("tables")

         set procedure to :dUFLP:BDEAlias.cc additive
         oBDE = new BDEAlias()
         cDataBasePath = oBDE.databaseDir( dMyData.databaseName )
         // add trailing backslash
         cDatabasePath += "\"

         //new section - 6/22/99
         if anyFiles()   //check for files
            fProgram=iif(type("fProgram")="O",fProgram,openfile()) // opens 'rebuild indes.prg'
            makeFile(lDatabase,aTables,fProgram,aDatabase[x],cDatabasePath) // writes to 'rebuild index.prg'
         endif
      endfor
      cleanup(fProgram)  // created cleanup function 6/23/99
   else
      // we're NOT using a database, so check the current directory:
      aTables = _app.databases[1].getSchema("tables")
      cDataBasePath = set("dire")

      //new section - 6/22/99
      if anyFiles() //check for files
         fProgram=openfile() // opens 'rebuild indes.prg'
         makeFile(lDatabase,aTables,fProgram,"",cDatabasePath)  // writes to 'rebuild index.prg'
         cleanup( fProgram )
      endif
   endif

return

function cleanup(fProgram)   // new function - old code 6/23/99
   // cleanup ... //pulled parts of following from end of old program 6/22/99

   // and turn off the code associated with the session object:
   fProgram.puts( "" )
   cOutString = "// Turn off code for Session object:"
   fProgram.puts( cOutString )
   cOutString = "_app.session.onProgress := null"
   fProgram.puts( cOutString )

   // end comment:
   cOutString = ""
   fProgram.puts( cOutString )
   cOutString = "/* -- End of Program: RebuildIndexes.prg   */"
   fProgram.puts( cOutString )

   // add more code at bottom
   EndFile( fProgram )

   // NOW close the program file
   fProgram.close()
   msgbox( "The program is RebuildIndexes.prg -- note if you used"+chr(13)+;
           "earlier versions of this program, the space is gone from the filename.",;
           "Finished!",64)
return

function Anyfiles // new function - old code 6/22/99
   // check to see if we're wasting our time:
   if aLen( aTables ) == 0
      msgbox( "There are no .DBF tables in this directory!",;
              "Nothing to do", 16)
      return false
   endif
return true

function openFile // new function - old code 6/22/99
   // create file object:
   fProgram = new File()
   // should just overwrite one if it exists
   fProgram.create( "RebuildIndexes.prg", "W" )
return (fProgram)

// new function - old code 6/22/99
// function called as many times as there are database names in the aDatabase array paramater
function makeFile(lDatabase,aTables,fProgram,cDatabase,cDatabasePath)
   // sort it ...
   aTables.sort()

   // Store some basic stuff into beginning of program:
   cOutString = "/*"
   fProgram.puts( cOutString )
   cOutString = "   RebuildIndexes.prg"
   fProgram.puts( cOutString )
   cOutString = "   Program Generated by 'CreateReindex.prg' in the dUFLP library."
   fProgram.puts( cOutString )
   cOutString = "   Date generated: "+date()
   fProgram.puts( cOutString )
   cOutString = "   This will delete the .mdx file and rebuild the index tags from scratch."
   fProgram.puts( cOutString )
   cOutString = ""
   fProgram.puts( cOutString )
   cOutString = "   ** Make sure you deploy: "
   fProgram.puts( cOutString )
   cOutString = "      :dUFLP:dBF7File.cc"
   fProgram.puts( cOutString )
   cOutString = "      :dUFLP:BFile.cc"
   fProgram.puts( cOutString )
   cOutString = "      :dUFLP:BDEAlias.cc"
   fProgram.puts( cOutString )
   cOutString = "*/"
   fProgram.puts( cOutString )
   cOutString = ""
   fProgram.puts( cOutString )
   cOutString = "close database // Will close tables/databases just to be safe"
   fProgram.puts( cOutString )
   cOutString = ""
   fProgram.puts( cOutString )
   cOutString = "// set code for Session object:"
   fProgram.puts( cOutString )
   cOutString = "_app.session.onProgress := ShowProgress"
   fProgram.puts( cOutString )
   fProgram.puts( "" )

   if lDatabase
      cOutString = "// set up the database if needed"
      fProgram.puts( cOutString )
      cOutString = "d = new Database()"
      fProgram.puts( cOutString )
      cOutString = "d.databaseName := '"+cDatabase+"'"
      fProgram.puts( cOutString )
      cOutString = "d.active       := true"
      fProgram.puts( cOutString )
      cOutString = ""
      fProgram.puts( cOutString )
   else
      cOutString = "// need a database reference"
      fProgram.puts( cOutString )
      cOutString = "d = _app.databases[1]"
      fProgram.puts( cOutString )
      cOutString = ""
      fProgram.puts( cOutString )
   endif
//   cOutString = "set procedure to :dUFLP:message.wfm additive"
//   fProgram.puts( cOutString )
   fProgram.puts( "// set up message form (below this code)" )
   cOutString = "fMessage = new MessageForm()"
   fProgram.puts( cOutString )
   cOutString = "fMessage.text = 'Processing Tables'"
   fProgram.puts( cOutString )
   cOutString = "fMessage.title.text = 'Recreating index tags'"
   fProgram.puts( cOutString )
   cOutString = "fMessage.open()"
   fProgram.puts( cOutString )

   // add a blank line
   cOutString = ""
   fProgram.puts( cOutString )

   // open a message dialog to show user something's
   // happening:
   set procedure to :dUFLP:message.wfm additive
   fMessage = new MessageForm()
   fMessage.text         := "Processing tables"
   fMessage.title.text   := "Getting index tags"
   fMessage.message.text := upper( aTables[ 1 ] )
   fMessage.Open()

   // need to get path to tables, so we can delete .mdx files 
   // as needed. Currently we're outputting a hard-coded path ...
   cOutString = "// Need the path to the tables"
   fProgram.puts( cOutString )
   if lDatabase
      cOutString = "set procedure to :dUFLP:BDEAlias.cc additive"
      fProgram.puts( cOutString )
      cOutString = "oBDE = new BDEAlias()"
      fProgram.puts( cOutString )
      cOutString = [cDataBasePath = oBDE.databaseDir( d.databaseName )]
      fProgram.puts( cOutString )
      cOutString = "// add trailing backslash"
      fProgram.puts( cOutString )
      cOutString = [cDatabasePath += "\"]
      fProgram.puts( cOutString )
   else
      cOutString = [cDatabasePath = set("dire")+"\"]
      fProgram.puts( cOutString )
   endif
   fProgram.puts( " " )


   // now, do it!
   fProgram.puts( "// --------------------------------------" )
   fProgram.puts( "// Start processing individual tables ..." )
   fProgram.puts( "// --------------------------------------" )

   for tableNumber = 1 to aLen( aTables, 1 )
       // Display information in message form:
       fMessage.message.text := upper( aTables[ tableNumber ] )

       // Instance of TableDef object
       tTable = new TableDef()
       if lDatabase
          tTable.database := dMyData
       endif
       // Give it the table name to use
       tTable.tableName := aTables[tableNumber]
       // Load the table
       tTable.load()
       // new for the update -- July 2005:
       // and we need a query:
       qTable = new query() 
       if lDatabase
          qTable.database := dMyData
       endif
       qTable.sql := [select * from ']+aTables[TableNumber]+[']
       qTable.active := true

       // shorten coding a bit:
       cTableName = aTables[ tableNumber ].toUpperCase()
       // output name of table and message instructions ...
       // add more spacing/comments between tables:
       fProgram.puts( "" )
       fProgram.puts( "   // --------------------------------------" )
       cOutString =   "   // -- "+cTableName
       fProgram.puts( cOutString )
       fProgram.puts( "   // --------------------------------------" )
       cOutString = "   fMessage.message.text = '"+cTableName+"'"
       fProgram.puts( cOutString )
       fProgram.puts( "   fMessage.pb.value := 0" )
       cOutString = "   fMessage.repaint()"
       fProgram.puts( cOutString )
       cOutString = "   if __version__ => 2.5"
       fProgram.puts( cOutString )
       cOutString = "      _app.executeMessages()"
       fProgram.puts( cOutString )
       cOutString = "   endif"
       fProgram.puts( cOutString )

       // Deal with index tags, if any. 
       if tTable.indexes.size > 0

          // Open the index ...
          set procedure to :dUFLP:BDEIndex.cc additive
          oInd = new BDEIndex()

          // code to unhook the indexes using :dUFLP:dBF7File.cc
          cOutString = "   /"+"/ Disconnecting index flag for table"
          fProgram.puts( cOutString )
          cOutString = "   try"
          fProgram.puts( cOutString )
          cOutString = "      set procedure to :dUFLP:dBF7File.cc additive"
          fProgram.puts( cOutString )
          if lDatabase // database? 
             cOutString = [      oDBF7File = new DBF7File( ":]+dMyData.databaseName+[:]+aTables[TableNumber]+[" )]
          else // no database
             cOutString = [      oDBF7File = new DBF7File( "]+aTables[TableNumber]+[" )]
          endif
          fProgram.puts( cOutString )
          // Disconnect MDX tags:
          cOutString = "      nRet = oDBF7File.SetMDXFlag(false) // turn off internal flag in table"
          fProgram.puts( cOutString )
          cOutString = "      if nRet > 0 // Error!"
          fProgram.puts( cOutString )
          // note, inserted 'nRet' into the msgbox string ...
          cOutString = [         msgbox( "An error ("+nRet+") occurred disconnecting MDX flag for table ']+;
                                 aTables[TableNumber]+['", "RebuildIndex.prg Error", 16 )]
          fProgram.puts( cOutString )
          cOutString = "         return"
          fProgram.puts( cOutString )
          cOutString = "      endif"
          fProgram.puts( cOutString )
          cOutString = "      release object oDBF7File"
          fProgram.puts( cOutString )
          cOutString = "      close procedure :dUFLP:dBF7File.cc"
          fProgram.puts( cOutString )
          // add code to delete the .MDX file -- this should hopefully get rid of any issues
          // with the primary index existing ...?
          cTableName = left( aTables[TableNumber], at( ".", aTables[TableNumber] )-1 )
          cIndexFile = cTableName+".mdx"
          cOutString = "      // Erase the .mdx file:"
          fProgram.puts( cOutString )
          cOutString = [      cFullIndex = cDatabasePath+"]+cIndexFile+["]
          fProgram.puts( cOutString )
          cOutString = [      if file( cFullIndex )]
          fProgram.puts( cOutString )
          cOutString = [         erase ( cFullIndex )]
          fProgram.puts( cOutString )
          cOutString = "      endif"
          fProgram.puts( cOutString )
          cOutString = "   catch( Exception E )"
          fProgram.puts( cOutString )
          cOutString = "      msgbox( 'Error: '+e.code+chr(13)+e.message )"
          fProgram.puts( cOutString )
          cOutString = "      return"
          fProgram.puts( cOutString )
          cOutString = "   endtry"
          fProgram.puts( cOutString )
          cOutString = "" // blank line
          fProgram.puts( cOutString )
       else
          cOutString = "   // No Indexes to Process"
          fProgram.puts( cOutString )
       endif

       // at this point there should be no .MDX file and hence
       // no index tags. Next, we need to rebuild ...

       // are there any index tags in the first place?
       if tTable.indexes.size > 0
          cOutString = "   // rebuild index tags ..."
          fProgram.puts( cOutString )
          fProgram.puts( "   fMessage.pb.value := 0" )

          for indexNumber = 1 to tTable.indexes.size

               // use BDEIndex.cc, we can get *ALL* the properties of the index tag:
               aTag = oInd.TagProperties( qTable, indexNumber )
               // see if this is a primary key
               if aTag["Primary"]
                  fProgram.puts( "" ) // blank line
                  cOutString = "   // index tag: "+aTag["IndexName"]+" (primary key)"
                  fProgram.puts( cOutString )
                  cOutString = "   // primary key, we must use XDML to create it:"
                  fProgram.puts( cOutString )
                  cOutString = "   try"
                  fProgram.puts( cOutString )

                  if lDatabase
                     cOutString = "      use :"+cDatabase+":"+cTableName+;
                                  " exclusive"
                  else
                     cOutString = "   use " + cTableName + " exclusive"
                  endif
                  fProgram.puts( cOutString )

                  // fix for single field primary key with space in field name
                  // added 5/24/99 by gw
                  cTemp = aTag["Expression"]
                  for i = 1 to tTable.fields.size
                      if trim(upper(tTable.fields[i].fieldName)) == ;
                             trim(upper(cTemp)) and ;
                             ' ' $ cTemp
                         cTemp := ':' + cTemp + ':'
                      endif
                  endfor
                  // end of fix

                  cOutString = "      index on " + ;
                        cTemp + ;
                        " tag '" + ;
                        aTag["IndexName"]+"'"

                  if aTag["Descending"]
                     cOutString += " descending"
                  endif

                  cOutString += " primary"
                  fProgram.puts( cOutString )

                  cOutString = "   catch ( exception e )"
                  fProgram.puts( cOutString )

                  cOutString = ;
                     "     // dBASE may throw an exception about " + ;
                     "invalid file name here " + chr(13) + ;
                     "     // (presumably for the key violation table)" + chr(13) + ;
                     "     // of course it will also fail if we didn't" + chr(13) + ;
                     "     // get exclusive use of the table."
                  fProgram.puts( cOutString )

                  cOutString = "   endtry"
                  fProgram.puts( cOutString )

                  cOutString = "   use"
                  fProgram.puts( cOutString )

                  cOutString = ""
                  fProgram.puts( cOutString )

               else
                  /*  
                     Modified, April 2, 2003 -- changed function call for
                     expression, forExpression and indexName, in case of problems
                     with the use of various string delimiters the problem
                     is that the table designer allows you to use any 
                     combination of the three possible delimiters, and this 
                     caused a problem in some testing Geoff Wass and Ivar Jessen 
                     were doing. We also need to check to see if the expression
                     and forExpression contain, by chance, all three delimiters,
                     in which case we cannot create this index tag ...
                  */
                  cExp    = aTag["Expression"] 
                  cForExp = aTag["ForExpression"]

                  // check to see if the expression contains all three
                  // OR the forExpression contains all three ...
                  if ( cExp.indexOf( ["] ) > -1 AND ;
                       cExp.indexOf( ['] ) > -1 AND ;
                       cExp.indexOf( "[" ) > -1 ) OR ;
                     ( cForExp.indexOf( ["] ) > -1 AND ;
                       cForExp.indexOf( ['] ) > -1 AND ;
                       cForExp.indexOf( "[" ) > -1 ) 

                     msgbox( "Index: "+aTag["IndexName"] +;
                             "has an expression or forExpression containing "+chr(13)+;
                             "all three string delimiters, which this program cannot "+;
                             "work with."+chr(13)+;
                             "Please use a maximum of two string delimiters ... "+chr(13)+;
                             "This index tag will not be created.",;
                             "CreateReindex Error", 16 )

                  else // it's okay
                     fProgram.puts( "" ) // blank line
                     cOutString = "   // index tag: "+aTag["IndexName"]
                     fProgram.puts( cOutString )
                     cOutString = "   i = new DBFIndex()"
                     fProgram.puts( cOutString )
                     cOutString = "   i.expression    := "+CheckDelim( aTag["Expression"] )
                     fProgram.puts( cOutString )
                     cOutString = "   i.descending    := "+aTag["Descending"]
                     fProgram.puts( cOutString )
                     cOutString = "   i.unique        := "+aTag["Unique"]
                     fProgram.puts( cOutString )
                     cOutString = "   i.forExpression := "+CheckDelim( aTag["ForExpression"] )
                     fProgram.puts( cOutString )
                     cOutString = "   i.indexName     := "+CheckDelim( aTag["IndexName"] )
                     fProgram.puts( cOutString )
                     // update message:
                     cOutString = "   fMessage.message.text := ["+cTableName+"]+chr(13)+;"
                     fProgram.puts( cOutString )
                     cOutString = "                            ["+aTag["IndexName"]+"]"
                     fProgram.puts( cOutString )
                     cOutString = "   if __version__ => 2.5"
                     fProgram.puts( cOutString )
                     cOutString = "      _app.executeMessages()"
                     fProgram.puts( cOutString )
                     cOutString = "   endif"
                     fProgram.puts( cOutString )

                     // actually create it:
                     cOutString = "   d.createIndex( '"+cTableName+"', i ) // create it"
                     fProgram.puts( cOutString )

                  endif // check for all three string delimiters

             endif // skip primary key

          next // indexNumber
       endif

       cOutString = ""
       fProgram.puts( cOutString )
       release object tTable
       tTable = null
   next

   if lDatabase
      // deactivate the database we used in this program:
      dMyData.active := false
      // output code to deactivate it in the new program:
      cOutString = "// deactivate database:"
      fProgram.puts( cOutString )   
      cOutString = "d.active := false"
      fProgram.puts( cOutString )
   endif

   // the rest ... (cleanup ...)
   cOutString = ""
   fProgram.puts( cOutString )
   cOutString = "// remove the message dialog"
   fProgram.puts( cOutString )
   cOutString = "fMessage.release()"
   fProgram.puts( cOutString )

   // Add code to the end of the program:
   fMessage.message.text := "Finishing ..."

   // cleanup
   fMessage.release()                  //left this here.

return

function EndFile( fProgram )
   // code added to the end of the RebuildIndexes.prg file:

   fProgram.puts( "" )
   fProgram.puts( "// -------------------------------------------" )
   fProgram.puts( "// More code needed, including Message form" )
   fProgram.puts( "// Do not remove!" )
   fProgram.puts( "// -------------------------------------------" )

   fProgram.puts( [class messageForm of FORM])
   fProgram.puts( [   with (this)])
   fProgram.puts( [      open = class::FORM_OPEN])
   fProgram.puts( [      scaleFontBold = false])
   fProgram.puts( [      height = 5.2273])
   fProgram.puts( [      left = 41.8571])
   fProgram.puts( [      top = 9.7273])
   fProgram.puts( [      width = 45.0])
   fProgram.puts( [      text = "Processing"])
   fProgram.puts( [      autoCenter = true])
   fProgram.puts( [      sizeable = false])
   fProgram.puts( [      maximize = false])
   fProgram.puts( [      minimize = false])
   fProgram.puts( [      icon = "resource #210"])
   fProgram.puts( [   endwith])
   fProgram.puts( [])
   fProgram.puts( [   this.TITLE = new TEXT(this)])
   fProgram.puts( [   with (this.TITLE)])
   fProgram.puts( [      height = 1.5455])
   fProgram.puts( [      left = 0.0])
   fProgram.puts( [      top = 0.0])
   fProgram.puts( [      width = 45.0])
   fProgram.puts( [      border = true])
   fProgram.puts( [      anchor = 2	/]+[/ Top])
   fProgram.puts( [      colorNormal = "Blue/BtnFace"])
   fProgram.puts( [      alignVertical = 1	/]+[/ Middle])
   fProgram.puts( [      alignHorizontal = 1	/]+[/ Center])
   fProgram.puts( [      fontSize = 14.0])
   fProgram.puts( [      text = "Title Message Here"])
   fProgram.puts( [      borderStyle = 1	/]+[/ Raised])
   fProgram.puts( [   endwith])
   fProgram.puts( [])
   fProgram.puts( [   this.MESSAGE = new TEXT(this)])
   fProgram.puts( [   with (this.MESSAGE)])
   fProgram.puts( [      height = 2.4091])
   fProgram.puts( [      left = 0.0])
   fProgram.puts( [      top = 1.5455])
   fProgram.puts( [      width = 44.8571])
   fProgram.puts( [      border = true])
   fProgram.puts( [      alignHorizontal = 1	/]+[/ Center])
   fProgram.puts( [      text = "Message Here"])
   fProgram.puts( [      borderStyle = 10	/]+[/ Etched Out])
   fProgram.puts( [   endwith])
   fProgram.puts( [])
   fProgram.puts( [   this.PB = new PROGRESS(this)])
   fProgram.puts( [   with (this.PB)])
   fProgram.puts( [      height = 1.0])
   fProgram.puts( [      left = 0.4286])
   fProgram.puts( [      top = 4.0909])
   fProgram.puts( [      width = 44.0])
   fProgram.puts( [      value = 0])
   fProgram.puts( [      rangeMin = 0])
   fProgram.puts( [      rangeMax = 100])
   fProgram.puts( [   endwith])
   fProgram.puts( [])
   fProgram.puts( [])
   fProgram.puts( [   function form_open])
   fProgram.puts( [      /]+[/ if dBASE Plus or later, we usually don't])
   fProgram.puts( [      /]+[/ want this form to show the button on the])
   fProgram.puts( [      /]+[/ taskbar, so ...])
   fProgram.puts( [      if __version__ => 2.0])
   fProgram.puts( [         this.showTaskbarButton := false])
   fProgram.puts( [      endif])
   fProgram.puts( [   return MESSAGEFORM::open()])
   fProgram.puts( [])
   fProgram.puts( [   function Repaint /]+[/ code by Jim Sare])
   fProgram.puts( [      if type( "UpdateWindow" ) # "FP"])
   fProgram.puts( [         extern CLOGICAL UpdateWindow( CHANDLE ) USER32])
   fProgram.puts( [      endif])
   fProgram.puts( [      /]+[/ if no parameter, use current form])
   fProgram.puts( [      if argCount() = 0])
   fProgram.puts( [         UpdateWindow( form.hWnd )])
   fProgram.puts( [      else])
   fProgram.puts( [         UpdateWindow( argVector(1) )])
   fProgram.puts( [      endif])
   fProgram.puts( [])
   fProgram.puts( [   RETURN ])
   fProgram.puts( [])
   fProgram.puts( [endclass])
   fProgram.puts( "// end of MessageForm class definition" )

   // code for the session object to update the form:
   fProgram.puts( "" )
   fProgram.puts( "// ---------------------------------------------" )
   fProgram.puts( "// Display progress ..." )
   fProgram.puts( "// ---------------------------------------------" )
   fProgram.puts( "function ShowProgress( nPercent, msg )" )
   fProgram.puts( "   if nPercent >= 0" )
   fProgram.puts( "      fMessage.PB.value := nPercent" )
   fProgram.puts( "      if __version__ => 2.5 " )
   fProgram.puts( "         _app.ExecuteMessages()" )
   fProgram.puts( "      endif" )
   fProgram.puts( "   endif" )
   fProgram.puts( "return" )
   fProgram.puts( "// end of ShowProgress function" )
   fProgram.puts( "" )
   fProgram.puts( "// the 'real' end of the program")

return

function CheckDelim( cString )
   /*
      Code provided by Geoff Wass to check for
      delimiters of specific types, and try to
      deal with setting beginning/ending delimiters
      for a string that are not contained IN the
      string ...
   */
   local cStart, cEnd

   cStart = ""
   cEnd   = ""

   if cString.indexOf( ["] ) == -1 // " not found
      cStart := ["]
      cEnd   := ["]
   elseif cString.indexOf( ['] ) == -1 // ' not found
      cStart := [']
      cEnd   := [']
   elseif cString.indexOf( "[" ) == -1 // [ not found
      cStart := "["
      cEnd   := "]"
   endif

return( cStart+cString+cEnd )

/* 
    -- End of Program: CREATEREINDEX.PRG --
*/

