/*
   --------------------------------------------------------------------
   Name.....: FileEx32.cc
   Author...: Michael Nuwer
   Date.....: June 5, 2003 
   Notes....:    
      FileEX32 is a subclass of the dBASE FILE class. As such it 
      inherits all of the properties and methods found in the file
      class. In addition this class includes methods for manipulating 
      text data contained in a file.  

      All methods respect the access rights of the file object.

      Create()         -- Overridden Create method
      CreateTempFile() -- Create's a temporary file in Temp folder
      DeleteStr()      -- Delete a string from inside a file
      GetFileName()    -- Gets the name of the file
      GetFilePath()    -- Gets the path to the file
      Insert()         -- Insert's a string into a file
      IsDir()          -- Bowen Moursund's IsDir() -- checks to see
                          if a path is an existing directory
      Length()         -- Length of the file (bytes/characters)
      LineCount()      -- Number of lines in the file
      Locate()         -- Locates a string in a file
      Open()           -- Overridden Open method
      TokenCount()     -- Number of occurances of a specified string

      Internal Properties:
      FileName         -- Name of the file (including path)
      AccessRights     -- Access rights
      TempFile         -- Name/path of temporary file created
                          by CreateTempFile()

   An Extended Example

   set procedure to :dUFLP:fileEX32.cc additive
   f = new FileEX32()
   if f.exists("TestFile.txt")
      f.delete("TestFile.txt")
   endif

   //  Create a new file
   f.create("TestFile.txt", "RW")

   // insert two lines
   f.Insert( "<!CUTSTART>"+chr(13)+chr(10) ) 
   f.Insert( "<!CUTEND>"+chr(13)+chr(10) )

   // insert a line at the beginning of the file
   f.Insert( replicate("test ",7)+chr(13)+chr(10), 0 )

   // insert a line at the end of the file
   f.Insert( replicate("test ",7)+chr(13)+chr(10), f.length() )

   // goto the beginning of the file
   f.seek(0,0)

   // locate the end of the string "<!CUTSTART>" 
   nInsert = f.Locate("<!CUTSTART>", true) + len("<!CUTSTART>"+chr(13)+chr(10))
   // insert text after this string
   f.Insert( replicate("delete this ",7)+chr(13)+chr(10), nInsert  )
   // Close the file
   f.close()


   // Open the new file and display its contents.
   f.open("TestFile.txt","RW")
   msgbox(f.read(f.length()))
   f.seek(0,0)


   //  Replace all occurrences of "test" with "Test"
   for i=1 to f.tokenCount("test")
      nStart = f.locate("test",false)
      f.deleteStr(len("test"),nStart)
      f.insert("Test",nStart)
   next

   // goto beginning of file and display its contents.
   f.seek(0,0)
   msgbox(f.read(f.length()))
   f.seek(0,0)


   // remove specified text from file.

   // find the spot to begin deleting
   nStart = f.Locate("<!CUTSTART>", true)
   // go to bof
   f.seek(0,0)

   // find the spot to end deleting
   nEnd = f.Locate( "<!CUTEND>", true)

   // compute the number of bytes to delete
   nCut = nEnd-nStart +len("<!CUTEND>"+chr(13)+chr(10))

   // Delete the text
   f.deleteStr(nCut,nStart)

   // goto beginning of file and display the contents.
   f.seek(0,0)
   msgbox(f.read(f.length()))

   f.close()
   f = null

   return
   --------------------------------------------------------------------
*/
class FileEX32 of file

   function deleteStr(nBytes,nStart)
   /*
      -------------------------------------------------------------
       Programmer..: Michael Nuwer
       Date........: June 5, 2003
       Notes.......: Delete a specific number of characters in an 
                     open file.
       Written for.: dBASE Plus
       Rev. History: June 5, 2003 -- Original
       Calls.......: fileex32.CreateTempFile()
       Usage.......: FileEx32.DeleteStr(<nBytes>,<nStart>)
       Example.....: ? oFile.DeleteStr(1000,1)
       Returns.....: Logical
       Parameters..: nBytes  = number of chars (bytes) to delete in 
                               file
                     nStart  = position to start deleting at
       Detailed Example:
          Replace all instances of "test" with "Test". 
          tokenCount() will return the number of occurrences 
          of the word "test". We can then loop for each occurrence 
          of the word: (1) locate the starting position of word; 
          (2) delete the occurrence; and (3) insert the new text.

            for i=1 to oFile.tokenCount("test")
               nStart = oFile.locate("test",false)
               oFile.deleteStr(len("test"),nStart)
               oFile.insert("Test",nStart)
            next
      -------------------------------------------------------------
   */
      // based on code by Adam L. Menkes
      // Remove nBytes from the open file beginning at nStart

      if not "W" $ this.AccessRights
         msgbox("File not open for editing","Error",48)
         return
      endif

      local fTemp,nRead
      this.CreateTempFile("TMP")
      fTemp = new file()
      fTemp.create(this.TempFile,"RW")

      //  ensure the file can be read.
      this.close()
      super::open(this.FileName,"R")

      // read from BOF to start of delete characters
      nRead = this.read( nStart )
      // write characters to temp file
      fTemp.write(nRead)

      // move pointer past text to be removed
      this.seek(nBytes,1)

      //  Read to end of file
      nRead = this.read( this.length() - this.seek(0,1)  )
      // write characters to temp file
      fTemp.write(nRead)

      //  Close both files
      this.close()
      fTemp.close()
      // Replace file with temp file
      fTemp.copy(this.TempFile,this.FileName)
      fTemp.delete(this.TempFile)
      // Open file as before
      this.open(this.fileName,this.AccessRights)
      //  return pointer to start position
      this.seek(nStart,0)
   return ( this.error() == 0 )
   // -- EoM: DeleteStr()

   function Insert(cString,nStart)
   /*
      -------------------------------------------------------------
       Programmer..: Michael Nuwer
       Date........: June 5, 2003
       Notes.......: Insert a string at a specified location in 
                     an open file.
       Written for.: dBASE Plus
       Rev. History: June 5, 2003 -- Original
       Calls.......: fileex32.CreateTempFile()
       Usage.......: FileEx32.Insert(<cString>,<nStart>)
       Example.....: ? oFile.Insert(cString + chr(13) + chr(10), ;
                                    oFile.length())
       Returns.....: Logical
       Parameters..: cString  = character string to insert
                     nStart   = position to start inserting the 
                                string
       Detailed Example:
         Example: Insert a string at the end of the file:
            oFile.Insert( cString + chr(13) + chr(10), oFile.length() )

         Insert a last name after a first name:
            nInsert = oFile.Locate("Michael ", false) + len("Michael ")
            oFile.Insert( "Nuwer" , nInsert  )
      -------------------------------------------------------------
   */
      //  If the file access rights are "A" nStart is ignored
      if not ( "W" $ this.AccessRights ) AND ;
         not ( "A" $ this.AccessRights ) 
         msgbox("File not open for editing","Error",48)
         return
      endif

      local fTemp,nRead

      if "A" $ this.AccessRights or pcount() < 2
         // if file is open in Append mode or
         // if start position is not passed
         // set start position at end of file.
         nStart = this.Length()  
      endif

      this.CreateTempFile("TMP")
      fTemp = new file()
      fTemp.create(this.TempFile,"RW")

      //  ensure the file can be read.
      this.close()
      super::open(this.FileName,"R")

      // read file upto start of insert position
      nRead = this.read( nStart )

      // write this to the temp file
      fTemp.write(nRead)

      // Insert text 
      fTemp.write(cString) 

      if nStart < this.length()
         // If not append mode

         // read the remainder of the file
         nRead = this.read( this.length() - this.seek(0,1) )
         // write this to the temp file
         fTemp.write(nRead)
      endif

      // close both files
      this.close()
      fTemp.close()
      // replace file with temp file
      fTemp.copy(this.TempFile,this.FileName)
      fTemp.delete(this.TempFile)
      // open file as we started.
      this.open(this.fileName,this.AccessRights)

      if "W" $ this.AccessRights
         // move pointer to end of insert string.
         this.seek(nStart + len(cString),0)

         //  If Rights are "A" then the pointer
         //  is at EOF (automatically)
      endif
   return ( this.error() == 0 )
   // EoM: Insert()

   function Locate( cSearch, bMatchCase )
   /*
      -------------------------------------------------------------
       Programmer..: Michael Nuwer
       Date........: June 5, 2003
       Notes.......: Find the occurrence of a string in the open 
                     file. Leaves the file pointer at this 
                     position. 
       Written for.: dBASE Plus
       Rev. History: June 5, 2003 -- Original
       Calls.......: None
       Usage.......: FileEx32.Locate(<cString>,<bMatchCase>)
       Example.....: ? oFile.Locate( "Test", true )
       Returns.....: Numeric - the starting position of the 
                               search string.
       Parameters..: cString    = character string to locate
                     bMatchCase = logical (boolean) - whether 
                                  to match the case of the search 
                                  string. Default is false.
      -------------------------------------------------------------
   */
      //  based on code by Rose Vines

      if not "R" $ this.AccessRights
         msgbox("File not open for reading","Error",48)
         return
      endif

      local cBuffer, nCurPos, nStartPos, nBuffSize, nFlength
      local nBufPos, cTxtBuff, nBuffOffset, nFoundPos, cAddBuf

      nFoundPos  = -2
      
      // Convert buffer to uppercase if required
      if pcount() = 1
         bMatchCase = false
      endif
      
      //  Store our current position in the file,
      //  and check the file length  
      nStartPos   = this.seek(0,1) 
      nFlength    = this.seek(0,2)

      // Now start reading characters into the buffer
      this.seek(nStartPos,0)

      cBuffer = this.read(nFlength - nStartPos )

      if not bMatchCase
         cBuffer = lower(cBuffer)
         cSearch = lower(cSearch)
      endif
      // See if the search string is in the buffer
      nBufPos = at( cSearch, cBuffer )
      // and if it is, store its position in the file
      if nBufPos > 0
         nFoundPos = nStartPos + nBufPos - 1
      endif
      if nFoundPos < 1
         this.seek( nStartPos, 0 )
      else
         this.seek( nFoundPos, 0 )
      endif
   return( nFoundPos )
   // -- EoM: Locate()

   function Length
   /*
      -------------------------------------------------------------
       Programmer..: Michael Nuwer
       Date........: June 5, 2003
       Notes.......: Returns the total number of characters 
                     contained in the file name in the Open() 
                     or Create() methods. 
       Written for.: dBASE Plus
       Rev. History: June 5, 2003 -- Original
       Calls.......: None
       Usage.......: FileEx32.Length()
       Example.....: ? oFile.Length()
       Returns.....: Numeric - number of characters in the file
       Parameters..: None
      -------------------------------------------------------------
   */
      // based on code by Rose Vines
      local nCurPos, nLength

      // Locate current position in file without moving pointer
      nCurPos = this.seek(0,1) 

      // Find the length of the file by shifting the pointer to 
      // the end
      nLength = this.seek( 0, 2 )
      
      // Return the pointer to the original starting point
      this.seek( nCurPos, 0 )
      
   return( nLength )
   // -- EoM: Length()

   function CreateTempFile(cFileExt)
   /*
      -------------------------------------------------------------
       Programmer..: Michael Nuwer
       Date........: June 5, 2003
       Notes.......: Creates a temporary file in the system's 
                     temp folder. The temp file is used by the 
                     deleteStr() and Insert() methods 
       Written for.: dBASE Plus
       Rev. History: June 5, 2003 -- Original
       Calls.......: None
       Usage.......: FileEx32.CreateTempFile( <cFileExt> )
       Example.....: oFile.CreateTempFile( "TMP" )
                     ? oFile.TempFile
       Returns.....: Character - File path/name as property
                                 of the file object: TempFile
       Parameters..: None
      -------------------------------------------------------------
   */
      // based on code by Todd Kreuter
      local cFileExt, cPath, nLen, n
      if pCount() = 0
         cFileExt = ".TMP"
      else
         cFileExt = "."+cFileExt
      endif
      if type("GetTempPath") # "FP"
         extern culong GetTempPath(culong, cstring) kernel32 ;
         from "GetTempPathA"
      endif

      cPath	= SPACE(256)
      nLen	= cPath.length
      n		= GetTempPath(nLen, cPath)

      if n > nLen
         nLen	= n
         n		= GetTempPath(nLen, cPath)
      endif

      if n > 0
         cPath	= cPath.left(nLen)
      else
         cPath	= SET("Directory")
      endif

      this.TempFile = cPath + funique("TMP?????"+cFileExt)
   return
   // -- EoM: CreateTempFile()

   function TokenCount( cToken )
   /*
      -------------------------------------------------------------
       Programmer..: Michael Nuwer
       Date........: June 5, 2003
       Notes.......: Counts the number of occurrences of a string
       Written for.: dBASE Plus
       Rev. History: June 5, 2003 -- Original
       Calls.......: None
       Usage.......: FileEx32.TokenCount( <cToken> )
       Example.....: ? oFile.TokenCount( "Test" )
       Returns.....: Numeric -- Number of occurances of string in 
                                the file.
       Parameters..: cToken -- string to count occurances of
       Extended Example:
            Loop through each occurrence of a string.
            for i = 1 to oFile.TokenCount("Nuwer")
              nStart = oFile.Locate( "Nuwer", false)
              oFile.seek(len("Nuwer"),1)   
            next
      -------------------------------------------------------------
   */
      local nCount,cStr,nCurPos

      if not "R" $ this.AccessRights
         msgbox("File not open for reading","Error",48)
         return
      endif
      // save current cursor position
      nCurPos = this.seek(0,1)

      //  goto BOF
      this.seek(0,0)
      nCount = 1
      cStr   = this.read(this.length())

      // convert case for matching
      cStr   = lower(cStr)
      cToken = lower(cToken)
      if at(cToken, cStr) > 0
         do while at(cToken, cStr, nCount) > 0
            nCount++
         enddo
      endif

      // goto saved cursor position
      this.seek(nCurPos,0)
   return ( nCount-1 )
   // -- EoM: TokenCount()

   function LineCount
   /*
      -------------------------------------------------------------
       Programmer..: Michael Nuwer
       Date........: June 5, 2003
       Notes.......: Counts the number of lines in open text file
       Written for.: dBASE Plus
       Rev. History: June 5, 2003 -- Original
       Calls.......: None
       Usage.......: FileEx32.LineCount()
       Example.....: ? oFile.LineCount()
       Returns.....: Numeric -- Number of occurances lines
       Parameters..: None
      -------------------------------------------------------------
   */
      local nLines, cTemp, nCurPos

      if not "R" $ this.AccessRights
         msgbox("File not open for reading","Error",48)
         return
      endif
      // save current cursor position
      nCurPos = this.seek(0,1)

      //  goto BOF
      this.seek(0,0)
      nLines = 0
      do while not this.eof() 
         this.gets()
         nLines++
      enddo

      // goto saved cursor position
      this.seek(nCurPos,0)
   return nLines
   // -- EoM: LineCount()

   function getFilePath
   /*
      -------------------------------------------------------------
       Programmer..: Michael Nuwer
       Date........: June 5, 2003
       Notes.......: Returns the full path to the file named in 
                     the Open() or Create() methods
       Written for.: dBASE Plus
       Rev. History: June 5, 2003 -- Original
       Calls.......: None
       Usage.......: FileEx32.GetFilePath()
       Example.....: ? oFile.GetFilePath()
       Returns.....: Character -- path
       Parameters..: None
      -------------------------------------------------------------
   */
      local cPath, cCurFolder
      if "\" $ this.FileName

         cPath = this.fileName
         //  strip filename 
         cPath = substr(cPath,1,rat( "\",cPath ))

         //  Deal with relative path if used
         if left(cPath,2)==".."

            // e.g. ..\..\dBASE\
            cCurFolder = set('directory')
            do while at("..",cPath) >0
               cCurFolder = substr(cCurFolder,1,rat( "\",cCurFolder )-1)
               cPath = substr(cPath,4)
            enddo
            cPath = cCurFolder+"\"+substr(cPath,4)

         elseif left(cPath,1)=="."

            //  e.g. .\dBASE\Samples\
            cCurFolder = set('directory') +"\"
            cPath = cCurFolder+substr(cPath,3)
         endif 
      else

         // only a filename was used
         cPath = set('directory') +"\"
      endif 
   return cPath
   // -- EoM: GetFilePath()

   function getFileName
   /*
      -------------------------------------------------------------
       Programmer..: Michael Nuwer
       Date........: June 5, 2003
       Notes.......: Returns the filename only of the file named 
                     in the Open() or Create() methods
       Written for.: dBASE Plus
       Rev. History: June 5, 2003 -- Original
       Calls.......: None
       Usage.......: FileEx32.GetFileName()
       Example.....: ? oFile.GetFileName()
       Returns.....: Character -- Filename
       Parameters..: None
      -------------------------------------------------------------
   */
      local cFileName
      cFileName =  this.fileName
      cFileName = substr(cFileName,rat( "\",cFileName )+1)
   return  cFileName
   // -- EoM: GetFileName()


   function IsDir( cDir )
   /*
      -------------------------------------------------------------
       Programmer..: Bowen Moursund
       Date........: August 15, 2001 
       Notes.......: Determine whether a directory (or folder) 
                     exists.
       Written for.: dBASE Plus
       Rev. History: June 5, 2003 -- added to this class
       Calls.......: None
       Usage.......: FileEx32.IsDir( <cDir> )
       Example.....: ? oFile.IsDir( "C:\Temp" )
       Returns.....: Logical - if directory exists
       Parameters..: cDir - directory to look for
      -------------------------------------------------------------
   */
   return (not empty(funique(cDir+"\isdir???.???")))
   // -- EoM: IsDir()

   function Open(cFile,cRights)
   /*
      -------------------------------------------------------------
       Programmer..: Michael Nuwer
       Date........: June 5, 2003
       Notes.......: Overridden file.open() method
       Written for.: dBASE Plus
       Rev. History: June 5, 2003 -- Original
       Calls.......: None
       Usage.......: FileEx32.Open( <cFile>, <cRights> )
       Example.....: ? oFile.Open( "MyTestFile.txt", "RW" )
       Returns.....: None (stores cFile to object.FileName
                           and cRights to object.AccessRights
                           custom properties)
       Parameters..: cFile   - Filename to open
                     cRights - standard file rights (R = read, etc.)
      -------------------------------------------------------------
   */
      if "B" $ upper(cRights)
         msgbox("This class is intended for text files only.","Error",48)
      endif
      this.FileName = cFile
      this.AccessRights = upper(cRights)
   return super::Open(cFile,cRights)
   // -- EoM: Open()

   function Create(cFile,cRights)
   /*
      -------------------------------------------------------------
       Programmer..: Michael Nuwer
       Date........: June 5, 2003
       Notes.......: Overridden file.Create() method
       Written for.: dBASE Plus
       Rev. History: June 5, 2003 -- Original
       Calls.......: None
       Usage.......: FileEx32.Create( <cFile>, <cRights> )
       Example.....: ? oFile.Create( "MyTestFile.txt", "A" )
       Returns.....: None (stores cFile to object.FileName
                           and cRights to object.AccessRights
                           custom properties)
       Parameters..: cFile   - Filename to open
                     cRights - standard file rights (R = read, etc.)
      -------------------------------------------------------------
   */
      // override of built-in method
      if "B" $ upper(cRights)
         msgbox("This class is intended for text files only.","Error",48)
      endif
      this.FileName = cFile
      this.AccessRights = upper(cRights)
   return super::Create(cFile,cRights)
   // -- EoM: Create()


endclass

