/*

	Filename:   NumericEF.cc
   Version:    1.02 
   Date:       2002 12 26

   For:        dBASE PLUS / dB2K / Visual dBase 7.** 
               Windows 95/98/ME/XP/NT/2000

   Author:     Ronnie MacGregor
   Contact:    NumericEFFeedBack@high-lugtonridge.co.uk
	
         
   Description:

      A custom numeric entryfield with:

         Right hand numeric entry
          
         No mouse selection of it's contents allowed

         "SelectAll = true" behaviour

         Integer portion and Decimal entry validated and restricted by the
         EntryField picture property if there is one defined, otherwise by the
         properties of the datalinked field if there is one.

         Automatic leading zero insertion if the first entered character is
         a decimal point.

         Optional automatic insertion of the decimal point after the maximum
         number of integer portion characters have been entered.

         Backspace key deletes most recently entered character during entry.

         Ordinary delete key deletes the whole entry.

      On opening the control creates it's own Text control which covers the
      EntryField. This prevents user mouse access to the EntryField itself,
      and is used to display the control contents.
 
   
   Revisions:

      2003 01 03 Version 1.01 /////////////////////////////////////////////////

      Macro instantiation of CoverText object dropped in favour of creating the
      CoverText as a property of the EntryField while still using the form as
      the container. Slightly simpler, and slightly faster to open.

      The Control will now fire it's onChange event if it exists.

      The control will now try to add itself, as a reference, to a custom array
      existing as a property of it's own datalinked rowset (if it exists). This
      array is then available to the developer so that multiple instances of
      this control can all easily be refreshed by using code similar to that
      given as an example under the Methods heading.

      2003 02 01 Version 1.02 /////////////////////////////////////////////////

      New FormatDisplay() method added and used to format the display of entered
      characters after the tab or enter key is pressed. This behaviour was
      requested by Peter Balfour.


   Methods:

.     function Refresh()

      This method must be explicitly called after rowset navigation, saving,
      abandoning, deleting, or appending, to refresh the control.

      This custom control will try to add itself to it's own datalinked rowset's
      custom array of NumericEF references, so that multiple instances of this
      control can all be refreshed by using code similar to the following:

.        function rowset_event()
            for nCount = 1 to this.aNumericEFs.size
               this.aFinancialEFs[nCount].Refresh()
            endfor
            return


   Dependancies:

      None


   Acknowledgements:

      A solution created for Peter Balfour in the dBASE newsgroups, with some
      code taken from his original idea.

      Some elements of the cover control instantiation method derived from
      a sample by Marko Mihorko.
         
*/



class NumericEF(parentObj) of EntryField(parentObj) custom


   // Ensure control drops onto form correctly
   form.metric := 6 // Pixels

   
   // Standard and default properties
   with (this)
      onOpen = CLASS::onOpen
      canClose = CLASS::canClose
      onGotFocus = CLASS::onGotFocus
      onLostFocus = CLASS::onLostFocus
      key = CLASS::EF_Key
      height = 22
      width = 100
      value = ""
   endwith

   // Custom properties
   this.Version = 1.02           // Version number
   this.ExistingChars = ""       // The existing string of accepted characters
   this.HasChanged = false       // Whether the control contents have changed
   this.nLength = 0              // Length of picture property or field
   this.nInt = 0                 // Integer portion of picture property
   this.nDec = 0                 // Decimal portion of picture property
   this.AutoDecimalPoint = false // Auto insertion of decimal point


   // Prevent unwanted outside access to internals
   protect UpdateControl(), AddChar(), DelChar(), FireOnChange()
   protect this.Version, this.ExistingChars, this.HasChanged
   protect this.nLength, this.nInt, this.nDec


   ////////////////////////////////////////

   
   function onOpen()

      // Create Cover control
      this.CoverText = new TEXT(this.parent)
      with (this.CoverText)
         onLeftMouseDown = {; this.Owner.setFocus()}
         top = this.top
         left = this.left 
         width = this.width
         height = this.height
         border = true
         colorNormal = "WindowText/Window"
         alignHorizontal = 2 // Right
         text = ""
      endwith

      // Give cover a reference to this control
      this.CoverText.Owner = this

      // Deal with picture property
      if len(this.picture) > 0
         this.nLength := len(this.picture)
         this.nDec := len(this.picture) - at(".", this.picture)
         this.nInt := len(this.picture) - this.nDec - 1
      else
         try
            this.nLength := this.datalink.length
            this.nDec := this.datalink.DecimalLength
            this.nInt := this.datalink.length - this.nDec - 1
         catch (exception e)
            // Datalink doesn't exist
            this.nLength := 20
            this.nDec := 2
            this.nInt := 17
         endtry
      endif

      // Initialise display
      this.Refresh()

      // Add a reference this control to the 
      // custom rowset array of NumericEF controls
      try
         r = this.datalink.parent.parent
      catch (exception e)
         return  // Datalink doesn't exist 
      endtry
      try
         a = r.aNumericEFs
      catch (exception e)
         // Array doesn't exist so create it
         r.aNumericEFs = new array()
      endtry
      r.aNumericEFs.add(this)

      return


   /////////


   function onGotFocus()
      this.start := true
      this.HasChanged := false
      this.CoverText.colornormal := "HighLightText/HighLight"
      return


   /////////


   function onLostFocus()
      this.CoverText.colornormal := "WindowText/Window"
      this.FireOnChange()
      return


   /////////


   function canClose()

      // Remove circular reference
      this.CoverText.Owner := null

      // Remove the reference to this control in the 
      // rowset's custom array of FinancialEF controls
      try
         r = this.datalink.parent.parent
         a = r.aNumericEFs
         for nCount = 1 to  a.size
            if a[nCount].name == this.name
               a.delete(nCount)
            endif
         endfor // next
      catch (exception e)
         return true // entry doesn't exist 
      endtry

      return true


   /////////


   function EF_Key(nChar, nRepCnt, nFlags)

      // Deal with return or tab key
      if nChar == 13 or nChar == 9
         this.FormatDisplay()
         return true
      endif

      // Prevent Alpha Entry
      if nChar > 47 and nChar < 58 or nChar == 46 or nChar == 45 or nChar == 8 or nChar == 127
         // Allowable key - just continue
      else
         ? chr(7)
         return false
      endif

      // force over-write 
      if this.start == true 
         this.ExistingChars := ""
         this.start = false
      endif

      // Get existing entry 
      c = this.ExistingChars

      // Prevent unwanted minus signs
      if nChar == 45 and len(c) > 0
         ? chr(7)
         return false
      endif

      // Deal with decimal point
      if nChar == 46 and "." $ c
         ? chr(7)
         return false
      endif

      // Deal with required leading zero
      if nChar == 46 and len(c) == 0
         c = "0"
      endif
      if nChar == 46 and len(c) == 1 and c == "-"
         c = c + "0"
      endif

      // Limit entry to allowed integer places
      if not nChar == 46 // Decimal point
         if ( not "." $ c ) and len(c) >= this.nInt and nChar <> 8 and nChar <> 127 
            if this.nDec > 0 and this.AutoDecimalPoint == true
               c = c + "."
            else
               ? chr(7)
               return false
            endif
         endif
      endif

      // Limit entry to allowed decimal places
      if "." $ c and len(c) > this.nDec and nChar <> 8 and nChar <> 127
         if at(".",c) == len(c) - this.nDec 
            ? chr(7)
            return false
         endif
      endif

      // Process Key entry
      if nChar == 127 // Ordinary delete key
         c = ""
      elseif nchar == 8 // Backspace
         c = left(c,(len(c))-1)
      else 
         c = c + chr(nChar)
      endif

      // Update control
      this.UpdateControl(c)

      return false


   /////////

   
   function UpdateControl(c)

      // Update Control Values
      this.HasChanged := true
      this.ExistingChars := c
      this.value := val(c)
      try
         this.datalink.parent.parent.modified := true
      catch (exception e)
         // No datalink
      endtry

      // Update Display
      this.CoverText.text := c

      return


   /////////


   function FireOnChange()
      if this.HasChanged == true
         try
            this.onChange()
         catch (exception e)
            // No onChange() code
         endtry
         this.HasChanged := false
      endif
      return
 

   /////////


   function Refresh()
      this.FireOnChange()
      try
         c = ltrim(str(this.datalink.value, this.nLength, this.nDec))
         if c = null
            throw e
         endif
         if len(this.picture) > 0
            if val(c) <> this.datalink.value
               c = replicate("*",this.nInt)
               if this.nDec > 0
                  c = c + "." + replicate("*",this.nDec)
               endif
               this.CoverText.text := c
            endif
         endif
         this.CoverText.text := c
      catch (exception e)
         // Datalink doesn't exist or null table entry
         this.CoverText.text := ""
      endtry
      this.Start = true
      return


   /////////


   function FormatDisplay()
      c = this.CoverText.text
      if this.nDec > 0
         if not "." $ c
            c = c + "." 
         endif
         c = c + replicate("0", this.nDec-(len(c)-at(".",c,1)))
      endif
      this.CoverText.text := c
      this.start = true
      return


   /////////


endclass

//eof

