/*

	FileName:		SysInfo.cc
   Author:			Gary White

	Version History:

   June 15, 1999	Original version

   June 21, 1999	Moved lines setting procedures to the init method
   					to ensure that they would be executed.

                  Changed the ShowAbout method to center the about
                  box on the form that called it.

   June 23, 1999	Modified the centering routine to avoid the about
   					box being partially off-screen.

   July 1, 1999	Modified the siGetDBEngine routine based on code
   					supplied by Vic McClung that corrects a problem
                  where multiple paths are shown in the registry.
                  Change to use GetVersionEx and return additional
                  Windows version info including the NT Service Pack
                  and the Windows build number.

   Jan 15, 2000	Fixed the operating system build number and cleaned
   					up the formatting of the operating system string.

   March, 2005    Bruce Beacham noted a problem with an include
                  statement not being where it needs to be ...
                  This has been fixed in this version.

	Description:	Generic object that surfaces a number of system 
   					descriptive properties.  Primarily designed for 
                  use in populating a Help->About box, or may be
                  called to display a generic about box with
                  system information.


   Dependencies:	SysInfo.h		(included with this cc)
   					Strucmem.dll	(from VdB samples directory)
   					Structure.prg	(from VdB samples directory)
                  Registry.prg	(from VdB samples directory)
                  StructApi.h		(from VdB include directory)
                  Windef.h			(from VdB include directory)
                  Winreg.h			(from VdB include directory)

	Note:	Structure.prg and Registry.prg must be included in your exe
   		file.  Strucmem.dll must be deployed with your application.


   Usage:			set procedure to sysinfo.cc additive
   					s = new sysInfo()
                  with ( s )
                     appName = 'My Super Application'
                     appVersion = '1.0'
                     Copyright = 'Copyright 1999, by Me'
                     UserName = 'My Customer'
                  endwith
                  s.ShowAbout()	// display an about box

                  Alternatively, you may instantiate a SysInfo object
                  and query its properties to populate your own about
                  dialog box.


	Properties:		AppName
   								Name of your app. You must set this manually
   					AppVersion
                  			Your app version. You must set this manually
                  AppCopyright
                  			Copyright statement. You must set this manually
                  AvailableMemory
                  			Free physical memory as reported by Windows API
                  ColorDepth
                  			Number of colors supported by the display
                  ComputerName
                  			The computer name as reported by the API
                  DbEngine
                  			BDE version and date
                  DdeName
                  			Application DDE Name
                  DefaultTableLevel
                  			Table level as set in the BDE Admin applet
                  LangDriver
                  			Language driver as set in yourapp.ini
                  LocalShare
                  			Value of Local Share as set in BDE Admin applet
                  NumProcessors
                  			Number of processors (significant only on NT)
                  OperatingSys
                  			Operating System (Win95, 98, or NT) and version
                  PageFileSize
                  			Size of swap file
                  PageFileUsage
                  			Percent of swap file usage
                  PhysicalMemory
                  			Amount of physical memory in the machine
                  Processor
                  			Type of processor
                  RunTimeEngine
                  			Version of VdB, or VdB runtime
                  ScreenRes
                  			Screen resolution in pixels


   Events:			None


   Methods:			Refresh			Refreshes information that may change
                  Release			Releases the object from memory
                  SetDdeName		Sets the app's DdeServiceName.  
                  					The old DdeServiceName is reset in
                                 the release() method.
                  ShowAbout		Shows a generic About box


*/

#include structapi.h
#include windef.h
#include winreg.h
#include sysinfo.h

class SysInfo

   #include sysinfo.h
   protect init, siGetComputerName, siGetDbEngine, siGetDDEServName, ;
   	siGetLangDriver, siGetLocalShare, siGetMemoryInfo, ;
      siGetNumProcessors, siGetOperatingSys, siGetRunTimeEngine, ;
      siGetVideoInfo, siGetTableLevel

   // These first three properties must be set manually by 
   // your code. The values inserted here are informational 
   // only and are intended to be overwritten with 
   // information about your application..
	this.AppName = 'System Information'
   this.AppVersion = 'v 1.1'
   this.AppCopyright = 'Copyright 1999-2005, Gary White'

   // these properties are intended to be set automatically by the class
   this.AvailableMemory = ''
   this.ComputerName = ''
   this.DbEngine = ''
   this.DdeName = this.siGetDDEServName()
   this.LangDriver = ldriver()
   this.LocalShare = ''
   this.NumProcessors = ''
   this.OperatingSys = ''
   this.PageFileSize = ''
   this.PageFileUsage = ''
   this.PhysicalMemory = ''
   this.Processor = ''
   this.RunTimeEngine = version(0)
   this.ScreenRes = ''
   this.ColorDepth = ''
   this.DefaultTableLevel = ''

   this.init()

   function Init
		local nNumberP, nProcessor, nPlatform, vMinor, vMajor, ;
      	cExtra, nBuild, v, s

      // Processor and operating system won't change during the
      // class life, so we'll just grab that information here.
      try

         set procedure to '&_dbwinhome.samples\structure.prg' additive
         set procedure to '&_dbwinhome.samples\registry.prg' additive

         // get operating system
         v=new OSVERSIONINFO()	// create a version info structure

         if GetVersionEx( v.value )	// fill the structure
         	nPlatform = v.getMember('dwPlatformId')
         	vMajor = int(v.getMember('dwMajorVersion'))
            vMinor = int(v.getMember('dwMinorVersion'))
            nBuild = int(bitand(v.getMember('dwBuildNumber'), 0xffff))
            cExtra = trim(v.getMember('szCSDVersion'))
            do case
            	case nPlatform == VER_PLATFORM_WIN32s	// Win 3.x should not happen
               	this.operatingSys = 'Windows'
               case nPlatform == VER_PLATFORM_WIN32_WINDOWS	// Win95, or Win98, or ?
               	do case 
                  	case vMajor == 4 and vMinor == 0
	                     this.operatingSys = 'Windows 95 '
                     case vMajor == 4 and vMinor == 10
                        this.operatingSys = 'Windows 98 '
                     otherwise  // ? unknown
                        this.operatingSys = 'Windows '
                  endcase
                  this.operatingSys += '(' + vMajor + '.' + ;
                  	vMinor + '.' + nBuild + ')'
		            if not empty( cExtra )
	   		         this.operatingSys += cExtra
                  endif
               case nPlatform == VER_PLATFORM_WIN32_NT	// WinNT
               	if vMajor == 5
                  	this.operatingSys = 'Windows 2000'
                  else
                  	this.operatingSys = 'Windows NT '
                  endif
		            this.operatingSys += vMajor + '.' + vMinor
		            this.operatingSys += ' (Build '+ int( nBuild )
		            if not empty( cExtra )
	   		         this.operatingSys += ': ' + cExtra
                  endif
						this.operatingSys += ')'
            endcase
                                    	
         else
             this.operatingSys = os()
         endif

      catch (exception e)
      	this.operatingSys := os()
      endtry

      try

      	// get processor info
         s=new SYSTEM_INFO()	// create the System Info structure
         GetSystemInfo( s.value )	// fill the structure

         if ( nPlatform == 2 )	// NT
            nProcessor = s.getMember("wProcessorLevel")
         else
            nProcessor = s.getMember("dwProcessorType")
         endif

         do case
            case ( nProcessor == 3 )
            	this.processor :=  "386"
            case ( nProcessor == 4 )
               this.processor :=  "486"
            case ( nProcessor == 5 OR nProcessor == 586)
               this.processor :=  "Pentium"
            case ( nProcessor == 6 )
               this.processor :=  "Pentium II or Pro"
            otherwise
               this.processor :=  ltrim( trim( str( nProcessor ) ) )
         endcase

         nNumberP   = s.getMember("dwNumberOfProcessors")
         this.numProcessors :=  '' + int(nNumberP)

         if this.numProcessors == 1
            this.processor += ' (single processor)'
         else
            this.processor += ' ('+this.numProcessors+' processors)'
         endif

      catch (exception e)
         this.processor := 'Undetermined'
         this.numProcessors := 'Undetermined'
      endtry

      release nNumberP, nProcessor, vMinor, vMajor, v, s

      this.ComputerName := this.siGetComputerName()
      this.siGetMemoryInfo()
      this.siGetVideoInfo()
      this.DbEngine = this.siGetDBEngine()
      this.LocalShare = this.siGetLocalShare()
      this.DefaultTableLevel = this.siGetTableLevel()

   	return

   function Refresh
      this.ComputerName := this.siGetComputerName()
      this.siGetMemoryInfo()
      this.siGetVideoInfo()
      return

   function Release
   	_app.ddeServiceName = iif(type('this.savedDdeName')=='C',this.savedDdeName,'VDB')
   	release object this
      release this
      return

   function SetDdeName
   	parameters cNewName
      local bRet
      bRet = true
   	if not empty( cNewName ) and type('cNewName') == 'C'
      	this.savedDdeName = _app.ddeServiceName
	   	_app.ddeServiceName = cNewName
      else
      	bRet := false
      endif
      this.ddeName := this.siGetDDEServName()
      return bRet

   function ShowAbout
   	// Shows an About box with system information
      // If called from an MDI application, it will center
      // the about box on _app.framewin.  If called from an
      // SDI application, it will center the about box on
      // the currently active form.  If an error is 
      // encountered the form is centered on the screen.
   	local f, l, t, r, b, h, w
      try
      	this.refresh()
         f=new AboutBoxForm()
         f.sysInfo = this
         f.mdi := false
         try
            if not _app.frameWin.visible
               // we're in an mdi app
               hParent = GetActiveWindow()
            else
               // otherwise find out who called us
               hParent = _app.framewin.hWnd
            endif
            // get position and size of our "parent"
            sWin=new rect()
            if GetWindowRect( hParent, sWin.value)
               l=sWin.getMember('left')
               t=sWin.getMember('top')
               r=sWin.getMember('right')
               b=sWin.getMember('bottom')
               h=b-t
               w=r-l
               // now center our about form on the parent
               f.left := l + (w/2 - f.width/2)
               // don't set left to less than zero
               f.left := max( f.left, 0 )
               // don't set left so that the right side is off screen
               f.left := min( getSystemMetrics(0)-f.width, f.left)
               f.top := t + (h/2 - f.height/2)
               // don't set top to less than zero
               f.top := max( f.top, 0 )
               // don't set top so that the bottom is off screen
               f.top := min( getSystemMetrics(1)-f.height, f.top)
            else
               f.autoCenter := true
            endif
         catch (exception e)
         	f.autoCenter := true
         endtry
         f.readModal()
         f.release()
      catch (exception e)
      	msgbox(e.message,'Error #'+e.code,16)
      endtry
      return

   function siGetComputerName
   	local cName
      cName = space(256)
      try
         if GetComputerName( cName, 256 )
            cName := trim(cName)
         else
         	cName := 'Undetermined'
         endif
      catch( exception e)
         cName := 'Undetermined'
      endtry
   	return cName

   function siGetDBEngine
   	// Thanks to David Peters for this code (slightly modified
      // to show the BDE date as well as the version).
      local n, iVersionSize, sBuffer, iStart, c, sReturnString, sBDEPath, aDLLs
      try
         n = 0
         sReturnString = new string()
         r = new registry(HKEY_LOCAL_MACHINE, BDE_REG_KEY, false)
         sBDEPath = r.queryValue("DLLPATH")
         if ( sBDEPath.right(1) = ";" )
            sBDEPath = sBDEPath.left(sBDEPath.length-1)
         endif
         
         // Thanks to Vic McClung for this fix
         // the registry may have more than one path in it, i.e.: 
         //    "C:\Progam Files\Common Files\Borland\BDE\;C:\IDAPI"
         // select the one with idapi32.dll
         aDLLs = tokenize(sBDEPath, ';')
         for i = 1 to aDLLs.size
            sBDEPath = aDLLs[i]+'\idapi32.dll'
            if file(sBDEPath)
               // if this is the one we're looking for
               exit
            endif
         next
         // don't need to add chr(0) to a cstring
         iVersionSize = GetFileVersionInfoSizeA(sBDEPath, n)
         sBuffer = new STRING(SPACE(iVersionSize)+1)
         GetFileVersionInfoA(sBDEPath, 0, iVersionSize, sBuffer)
         iStart = AT("FileVersion", sBuffer)
         if ( EMPTY(iStart) )
            sReturnString = "Error"
         else
            iStart += 12
            isNull = false
            do while not isNull
               c = sBuffer.CharAt(iStart)
               if ( ASC(c) = 0 or iStart > 1024 )
                  isNull = true
               else
                  sReturnString += c
               endif
               iStart++
            enddo
         endif
         sReturnString = stuff(version(.89),at(":",version(.89))+2,0,sReturnString+' (')+')'
      catch (exception e)
      	// if we fail to get the version, at least return the date
      	sReturnString := version(.89)
msgbox(e.message+chr(13)+'line '+e.lineNo+chr(13)+'file '+e.filename)
      endtry
	   return (sReturnString)

   function siGetDDEServName
   	return _app.ddeServiceName

   function siGetLangDriver
   	return ldriver()

   function siGetLocalShare
   	try
         reg = new Registry(HKEY_LOCAL_MACHINE, ;
                        BDE_REG_KEY + ;
                        "\Settings\SYSTEM\INIT" )
         bLShare = reg.queryValue("LOCAL SHARE")
         if ( reg.error == 0 )
            bLShare := ''+bLShare
         else 
            bLShare := "Error reading registry"
         endif
      catch (exception e)
      	bLShare = 'Undetermined'
      endtry
      release object reg
      release reg
   	return bLShare

   function siGetMemoryInfo
   	local m, nAvail, nTotal
      try
         m=new memoryStatus()
         GlobalMemoryStatus( m.value )
         this.PhysicalMemory := ltrim(transform(round(m.getMember('dwTotalPhys'),0)/1024, '999,999,999,999 kb'))
         this.availableMemory := ltrim(transform(round(m.getMember('dwAvailPhys'),0)/1024, '999,999,999,999 kb'))
         nTotal = m.getMember('dwTotalPageFile')
         nAvail = m.getMember('dwAvailPageFile')
         nUsage = round(100 - (100 * (nAvail / nTotal)),0)
         this.PageFileSize := ltrim(transform(round(m.getMember('dwTotalPageFile')/1024,0), '999,999,999,999 kb'))
         this.PageFileUsage := ltrim(transform(nUsage, '999%'))
      catch (exception e)
			msgbox(e.message+chr(13)+'Line: '+e.lineno,'Error #'+e.code,16)
      	this.PhysicalMemory := 'Undetermined'
      	this.AvailableMemory := 'Undetermined'
      	this.PageFileSize := 'Undetermined'
      	this.PageFileUsage := 'Undetermined'
      endtry
      try
      	release object m
      catch (exception e)
      endtry
      release m
   	return

   function siGetRunTimeEngine
   	return version(0)

   function siGetVideoInfo
   	local h, n
   	try
         this.screenRes := getsystemmetrics(0)+'x'+getsystemmetrics(1)
      catch (exception e)
      	this.screenRes := 'Undetermined'
      endtry
   	try
         // now get the display information
         h = CreateDC('DISPLAY', '','','')
         n = GetDeviceCaps( h, BITSPIXEL )
         do case
            case n > 16
               this.ColorDepth = 'True color'
            case n == 16
               this.ColorDepth = '16 bit (65536 colors)'
            case n == 8
               this.ColorDepth = '8 bit (256 color)'
            case n == 4
               this.ColorDepth = '4 bit (16 color)'
            case n == 2
               this.ColorDepth = 'Monochrome'
            otherwise
               this.ColorDepth = 'Undetermined'
         endcase
      catch (exception e)
         this.colorDepth := 'Undetermined'
      endtry
      release h, n
   	return

   function siGetTableLevel
   	local reg, cLevel, nLevel
      try
         reg = new Registry(HKEY_LOCAL_MACHINE, ;
                        BDE_REG_KEY + ;
                        "\Settings\DRIVERS\DBASE\TABLE CREATE" )
         nLevel = reg.queryValue("LEVEL")
         if ( reg.error == 0 )
            cLevel = 'Level ' + nLevel
         else 
            cLevel = 'Error reading registry'
         endif
      catch (exception e)
      	cLevel = 'Undetermined'
      endtry
      release object reg
   	return cLevel

endclass

class MEMORYSTATUS of structure
   super::addMember(TYPE_DWORD,'dwLength')        // sizeof(MEMORYSTATUS) 
   super::addMember(TYPE_DWORD,'dwMemoryLoad')    // percent of memory in use 
   super::addMember(TYPE_DWORD,'dwTotalPhys')     // bytes of physical memory 
   super::addMember(TYPE_DWORD,'dwAvailPhys')     // free physical memory bytes 
   super::addMember(TYPE_DWORD,'dwTotalPageFile') // bytes of paging file 
   super::addMember(TYPE_DWORD,'dwAvailPageFile') // free bytes of paging file 
   super::addMember(TYPE_DWORD,'dwTotalVirtual')  // user bytes of address space 
   super::addMember(TYPE_DWORD,'dwAvailVirtual')  // free user bytes 
endclass

class SYSTEM_INFO of Structure
   super::addMember( TYPE_WORD,   "wProcessor" )
   super::addMember( TYPE_WORD,   "wReserved" )
   super::addMember( TYPE_DWORD,  "dwPageSize" )
   super::addMember( TYPE_LPVOID, "lpMinimumApplicationAddress" )
   super::addMember( TYPE_LPVOID, "lpMaximumApplicationAddress" )
   super::addMember( TYPE_DWORD,  "dwActiveProcessorMask" )
   super::addMember( TYPE_DWORD,  "dwNumberOfProcessors" )
   super::addMember( TYPE_DWORD,  "dwProcessorType" )
   super::addMember( TYPE_DWORD,  "dwAllocationGranularity"  )
   super::addMember( TYPE_WORD,   "wProcessorLevel" )
   super::addMember( TYPE_WORD,   "wProcessorRevision" )  
endclass

class RECT of Structure
	super::addMember( TYPE_LONG, 'left' )
	super::addMember( TYPE_LONG, 'top' )
	super::addMember( TYPE_LONG, 'right' )
	super::addMember( TYPE_LONG, 'bottom' )
endclass

class OSVERSIONINFO of structure
   super::addMember(TYPE_DWORD,'dwOSVersionInfoSize')
   super::addMember(TYPE_DWORD,'dwMajorVersion')
   super::addMember(TYPE_DWORD,'dwMinorVersion')
   super::addMember(TYPE_DWORD,'dwBuildNumber')
   super::addMember(TYPE_DWORD,'dwPlatformId')
   super::addMember(TYPE_STRING,'szCSDVersion', 128)
   this.setMember('dwOSVersionInfoSize',this.length())
endclass

class AboutBoxForm of FORM
   set procedure to SYSINFO.CC additive
   with (this)
      open = class::OPEN
      readModal = class::READMODAL
      scaleFontBold = false
      metric = 6	// Pixels
      height = 315
      left = 159
      top = 91
      width = 511
      text = "About"
      topMost = true
      mdi = false
      smallTitle = true
      sizeable = false
      maximize = false
      minimize = false
   endwith


   this.TABBOX1 = new TABBOX(this)
   with (this.TABBOX1)
      onSelChange = class::TABBOX1_ONSELCHANGE
      id = 123
      height = 22
      left = 0
      top = 293
      width = 511
      fontSize = 8
      dataSource = 'ARRAY {"Program Info","System Info"}'
   endwith


   this.RECTANGLE1 = new RECTANGLE(this)
   with (this.RECTANGLE1)
      left = 7
      top = 9
      width = 496
      height = 226
      text = "Program Information"
      pageno = 0
   endwith


   this.TEXT1 = new LabelText(this)
   with (this.TEXT1)
      height = 22
      left = 23
      top = 31
      width = 128
      value = "Application name:"
      pageno = 1
   endwith


   this.TEXT2 = new LabelText(this)
   with (this.TEXT2)
      height = 22
      left = 23
      top = 53
      width = 128
      value = "Application version:"
      pageno = 1
   endwith


   this.TEXT3 = new LabelText(this)
   with (this.TEXT3)
      height = 22
      left = 23
      top = 75
      width = 128
      value = "Application copyright:"
      pageno = 1
   endwith


   this.TEXT4 = new LabelText(this)
   with (this.TEXT4)
      height = 22
      left = 23
      top = 97
      width = 128
      value = "Runtime engine:"
      pageno = 1
   endwith


   this.TEXT5 = new LabelText(this)
   with (this.TEXT5)
      height = 22
      left = 23
      top = 119
      width = 128
      value = "Language driver:"
      pageno = 1
   endwith


   this.TEXT6 = new LabelText(this)
   with (this.TEXT6)
      height = 22
      left = 23
      top = 141
      width = 128
      value = "Database engine:"
      pageno = 1
   endwith


   this.TEXT7 = new LabelText(this)
   with (this.TEXT7)
      height = 22
      left = 23
      top = 163
      width = 128
      value = "Default table level:"
      pageno = 1
   endwith


   this.TEXT8 = new LabelText(this)
   with (this.TEXT8)
      height = 17
      left = 23
      top = 185
      width = 128
      value = "Local share:"
      pageno = 1
   endwith


   this.TEXT9 = new LabelText(this)
   with (this.TEXT9)
      height = 22
      left = 23
      top = 207
      width = 128
      value = "DDE service name:"
      pageno = 1
   endwith


   this.APPNAME = new LabelText(this)
   with (this.APPNAME)
      height = 22
      left = 158
      top = 31
      width = 338
      value = ""
      pageno = 1
   endwith


   this.APPVERSION = new LabelText(this)
   with (this.APPVERSION)
      height = 22
      left = 158
      top = 53
      width = 338
      value = ""
      pageno = 1
   endwith


   this.APPCOPYRIGHT = new LabelText(this)
   with (this.APPCOPYRIGHT)
      height = 22
      left = 158
      top = 75
      width = 338
      value = ""
      pageno = 1
   endwith


   this.RUNTIMEENGINE = new LabelText(this)
   with (this.RUNTIMEENGINE)
      height = 22
      left = 158
      top = 97
      width = 338
      value = ""
      maxLength = 36
      pageno = 1
   endwith


   this.LANGDRIVER = new LabelText(this)
   with (this.LANGDRIVER)
      height = 22
      left = 158
      top = 119
      width = 338
      value = ""
      pageno = 1
   endwith


   this.DBENGINE = new LabelText(this)
   with (this.DBENGINE)
      height = 22
      left = 158
      top = 141
      width = 338
      value = ""
      maxLength = 34
      pageno = 1
   endwith


   this.DEFAULTTABLELEVEL = new LabelText(this)
   with (this.DEFAULTTABLELEVEL)
      height = 22
      left = 158
      top = 163
      width = 338
      value = ""
      pageno = 1
   endwith


   this.LOCALSHARE = new LabelText(this)
   with (this.LOCALSHARE)
      height = 22
      left = 158
      top = 185
      width = 338
      value = ""
      pageno = 1
   endwith


   this.DDENAME = new LabelText(this)
   with (this.DDENAME)
      height = 22
      left = 158
      top = 207
      width = 338
      value = ""
      pageno = 1
   endwith


   this.PUSHBUTTON1 = new PUSHBUTTON(this)
   with (this.PUSHBUTTON1)
      onClick = {;form.close()}
      height = 24
      left = 219
      top = 254
      width = 72
      text = "Ok"
      default = true
      pageno = 0
      value = false
   endwith


   this.TEXT30 = new LabelText(this)
   with (this.TEXT30)
      height = 22
      left = 23
      top = 30
      width = 167
      value = "Processor:"
      pageno = 2
   endwith


   this.TEXT22 = new LabelText(this)
   with (this.TEXT22)
      height = 22
      left = 23
      top = 51
      width = 167
      value = "Computer name:"
      pageno = 2
   endwith


   this.TEXT23 = new LabelText(this)
   with (this.TEXT23)
      height = 22
      left = 23
      top = 73
      width = 167
      value = "Operating system:"
      pageno = 2
   endwith


   this.TEXT24 = new LabelText(this)
   with (this.TEXT24)
      height = 22
      left = 23
      top = 95
      width = 167
      value = "Screen resolution:"
      pageno = 2
   endwith


   this.TEXT25 = new LabelText(this)
   with (this.TEXT25)
      height = 22
      left = 23
      top = 117
      width = 167
      value = "Screen color depth:"
      pageno = 2
   endwith


   this.TEXT26 = new LabelText(this)
   with (this.TEXT26)
      height = 22
      left = 23
      top = 139
      width = 167
      value = "Total physical memory:"
      pageno = 2
   endwith


   this.TEXT27 = new LabelText(this)
   with (this.TEXT27)
      height = 22
      left = 23
      top = 161
      width = 145 // 167
      value = "Available memory:"
      maxLength = 26
      pageno = 2
   endwith


   this.TEXT28 = new LabelText(this)
   with (this.TEXT28)
      height = 22
      left = 23
      top = 183
      width = 167
      value = "Page file size:"
      pageno = 2
   endwith


   this.TEXT29 = new LabelText(this)
   with (this.TEXT29)
      height = 22
      left = 23
      top = 206
      width = 167
      value = "Page file usage:"
      pageno = 2
   endwith


   this.PROCESSOR = new LabelText(this)
   with (this.PROCESSOR)
      height = 22
      left = 168
      top = 31
      width = 280
      value = ""
      pageno = 2
   endwith


   this.COMPUTERNAME = new LabelText(this)
   with (this.COMPUTERNAME)
      height = 22
      left = 168
      top = 51
      width = 280
      value = ""
      pageno = 2
   endwith


   this.OPERATINGSYS = new LabelText(this)
   with (this.OPERATINGSYS)
      height = 22
      left = 168
      top = 73
      width = 310
      value = ""
      pageno = 2
   endwith


   this.SCREENRES = new LabelText(this)
   with (this.SCREENRES)
      height = 22
      left = 168
      top = 95
      width = 280
      value = ""
      pageno = 2
   endwith


   this.COLORDEPTH = new LabelText(this)
   with (this.COLORDEPTH)
      height = 22
      left = 168
      top = 117
      width = 280
      value = ""
      pageno = 2
   endwith


   this.PHYSICALMEMORY = new LabelText(this)
   with (this.PHYSICALMEMORY)
      height = 22
      left = 168
      top = 139
      width = 280
      value = ""
      pageno = 2
   endwith


   this.AVAILABLEMEMORY = new LabelText(this)
   with (this.AVAILABLEMEMORY)
      height = 22
      left = 168
      top = 161
      width = 280
      value = ""
      pageno = 2
   endwith


   this.PAGEFILESIZE = new LabelText(this)
   with (this.PAGEFILESIZE)
      height = 22
      left = 168
      top = 183
      width = 280
      value = ""
      pageno = 2
   endwith


   this.PAGEFILEUSAGE = new LabelText(this)
   with (this.PAGEFILEUSAGE)
      height = 22
      left = 168
      top = 206
      width = 280
      value = ""
      pageno = 2
   endwith


   function open
      this.mdi=false
      this.init()
      return super::readModal()

   function readModal
      this.init()
      return super::readModal()

   function Init
      with (this)
         appCopyright.value = SysInfo.appCopyright
         appName.value = SysInfo.appName
         appVersion.value = SysInfo.appVersion
         availableMemory.value = SysInfo.availableMemory
         colorDepth.value = SysInfo.colorDepth
         computerName.value = SysInfo.computerName
         dbEngine.value = SysInfo.dbEngine
         ddeName.value = SysInfo.ddeName
         defaultTableLevel.value = SysInfo.defaultTableLevel
         langDriver.value = SysInfo.langDriver
         localShare.value = SysInfo.localShare
         operatingSys.value = SysInfo.operatingSys
         pageFileSize.value = SysInfo.pageFileSize
         pageFileUsage.value = SysInfo.pageFileUsage
         physicalMemory.value = SysInfo.physicalMemory
         processor.value = SysInfo.processor
         runtimeEngine.value = SysInfo.runtimeEngine
         screenRes.value = SysInfo.screenRes
      endwith
      return

   function TABBOX1_onSelChange
      form.pageNo=this.curSel
      if this.curSel==2
         form.rectangle1.text='System Information'
      else
         form.rectangle1.text='Program Information'
      endif
      return
endclass

class LabelText(oParent) of entryfield(oParent) custom
   with(this)
      colorNormal = 'BtnText/BtnFace'
      borderStyle = 3
      mousePointer = 1
      when = {||false}
   endwith
endclass

// ----------------------------------------------------------------------------
function tokenize(cString, cDelimiter)
   local aTokens; aTokens = new array()
   local nCntr; nCntr = 1
   local nPos
   if pcount() < 2
      cDelimiter = ' '
   endif
   do while .t.
      nPos = at(cDelimiter, cString)
      if nPos == 0
         nPos = len(cString)+1
         if nPos == 1
            exit
         endif
      endif
      aTokens.add(1)
      aTokens[nCntr] = left(cString, nPos-1)
      nCntr = nCntr+1
      if nPos > len(cString)
      	exit
      endif
      cString = substr(cString, nPos+1)
   enddo
   return aTokens

