/*
   -- TaskbarTray.cc    V1.4

   ------------------------------------------------------------------------------
   -- Description:

      This custom component is used to display an icon on Windows taskbar status
      area, that is named here "the Taskbar tray".

      Generally, icons in the Taskbar tray indicate the status of an operation,
      or can be sensitive to mouse events. For example, a printer icon in the 
      Taskbar tray shows that a print job is under way. 
      The Taskbar tray is at the right end of the taskbar (if the taskbar has a 
      horizontal orientation) or at the bottom (if the taskbar has a vertical 
      orientation), just near the clock (if selected).

      Icons in the Taskbar Tray often allow the user to obtain a tooltip
      information by moving the mouse over the icon, and perform some given
      actions by left/right clicking or double-clicking the icon.

      The TaskbarTray component will enable to perform all these actions, in
      an "event driven" way, as for any other Visual dBASE component. 

   -- Programmer:

         Nicolas Martin,          nicolas.martin@freesurf.fr

   -- Copyright Nicolas Martin, 1999

   -- History:

         10/11/1999	-- Original V1.0
         19-Nov-2006	--	Modified V1.1	by Andrew Shimmin
         					Added				getPopUpObjRef() method
        	11-Mar-2007	--	Modified	v1.2	by Andrew Shimmin
         			  		Fixed				The modifyIcon() method leaked 2 x GDI & 1 x User objects.
                                       This was caused by changing the icon and not releasing
                                       the previous icon memory. This occurred on each call
                                       to the modifyIcon() method when the icon was changed.
                        added				DestroyIcon() external
                        added				releaseIcon() method
                        modified			deleteIcon() method
                        modified			setupIconData() method

        	28-Jan-2008	--	Modified	v1.3	by Andrew Shimmin
         					Added				IamLoaded() method to support checking if the class is
                        					loaded by using the technique:

                                       	if type("TaskBarTray::IamLoaded") # "FP"
                                          	set procedure to :dUFLP:TaskBarTray.cc additive
                                          endif
                   		Tested			W2K,WXPP,Vista Ultimate using dBPlus b2059
 
          13-Feb-2008   modified	v1.4 by Sylvain Lavoie
                                   Correct the displayPopup method to make the popupmenu
                                   disappears when the user click outside of the popupmenu.

   -- designed for Visual dBase, version 7.01,
      tested under Win 95.

   -- dependancies:

            - Structure.prg         (_dbwinhome\samples)
            - StructAPI.prg         (_dbwinhome\samples)
            - structmem.dll         (_dbwinhome\samples)
            - StructAPI.h           (_dbwinhome\include)
            - Winuser.h             (_dbwinhome\include)
            - Winmisc.h             (_dbwinhome\include)
            - Vdbase.h              (_dbwinhome\include)

   -- How to use:
      -----------

      You can add this component to your component palette, or simply instantiate
      it in your form by programmation.

      Add the TaskbarTray component to your form, if you have it your component
      palette, drop it on the form. In the "events" page of the inspector, you
      have access to the component's events that will be triggered, as usual in
      Visual dBASE, by mouse interactions.

      The component will not be visible on the form at runtime.

      NOTE: Your form's metric need to be set to "pixels", otherwise, the component
      cannot work.

      Or you can add it programmatically, this can be done like this :

         set procedure to taskbarTray.cc additive
         local f
         f = new form()
         f.metric := 6    // pixels
         oTbTray = new TaskbarTray(f)
         f.open()

      After the component is instantiated, you can use the 2 properties 
      .IconFilename, .ToolTip to set the icon filename (an icon file having
      the standard Windows .ico format) and the tooltip text, and the property
      .AllowIconEvents to enable the mouse events when moved over the icon; 
      then call the addIcon() method to add the icon to the taskbar tray. 
      You can then remove it from the taskbar tray with deleteIcon(), or modify 
      it (.IconFilename, .Tooltip, and .AllowIconEvents) with modifyIcon() :

          oTbTray.IconFilename := _dbwinhome + "artwork\icon\chemical.ico"
          oTbTray.ToolTip := "My little icon"
          oTbTray.addIcon()

      Note: When the component is closed, the icon is automatically removed from 
            the taskbar tray.

      The code for the mouse events can be entered as for any other VdB's visual
      component, that is, either from the form designer by clicking the wrench tool,
      or directly programmatically. You have furthermore access to the flags, col 
      and row parameters provided with the mouse events. Note that setting the 
      .AllowIconEvents property to false, the calling ModifyIcon() will disable
      further mouse actions.

          oTbTray.onLeftDblClick = class::TaskbarTray_onLeftDblClick

      Adding a popup menu to be displayed near the icon, for instance, when the
      user right clicks the icon is also possible. For that, you need first to
      create a standard Visual dBASE popup menu, with, for instance, the popup
      menu designer. This will produce a .pop file, let's say MyMenu.pop, that
      contains a class definition, let's say MyMenuPOPUP.

      First, You attach the popup menu to the component using the attachPopup()
      method, supplying the 2 parameters "MyMenu.pop" and "MyMenuPOPUP".
      To display the menu, for instance in the onRightMouseUp event, you call
      the displayPopup() method.
      To detach the popup menu, use the detachPopup() method.

      Note : When the component is closed, its onClose event performs some
             clean up and remove the icon in the taskbar tray as well as 
             release the popup menu.


   -- Properties:
      ----------

      The 3 properties .IconFilename, .Tooltip, and .AllowIconEvents, when modified,
      must be followed by a call to AddIcon() or ModifyIcon() to take the change
      into account.

      .IconFilename: VdB string, default value is "".
                     The name of the icon file that is to be displayed in the 
                     taskbar tray. The format of this file must be the standard 
                     Windows .ico format. To create such icons, you can use,
                     for instance, the following applications :

                       - EasyIcons, shareware, to directly create icons and save 
                         them in the .ico format. 
                         http://www.easyapps.com/

                       - IrfanView, freeware, that lets you convert many image 
                         file formats into Windows .ico files. You can create
                         your icons in .bmp format with MS paint, Photoshop, 
                         Paint shop pro..., then convert them with this tool 
                         into a .ico file.
                         http://stud1.tuwien.ac.at/~e9227474/

      .ToolTip:   VdB String, default value is "". the tooltip to be displayed 
                   when the mouse cursor is dragged over the icon in the taskbar 
                  tray. If this string is empty, no tooltip is dispayed.

      .AllowIconEvents: VdB boolean, default value is true.
                        if true, the mouse interactions on the icon will be 
                        transformed into corresponding component's mouse events. 
                        If false, no mouse event will be generated.

      .DisplayErrors: VdB boolean, default value is true.
                      If true, internal error messages will be displayed if errors
                      occur. 
                      If false, no error messages are displayed. 
                      In any case, you can test the return code of each method 
                      to know whether an error occured.


   -- Methods:
      -------

      addIcon( )
         This method will add the desired icon with its tooltip text, supplied 
         in the properties .IconFilename and .ToolTip, to the taskbar tray. It 
         will allow also the mouse events, according to the state of the 
         .AllowIconEvents flag.

         return : true if OK, false if an error occured.

      modifyIcon( )
         This method changes the current icon, tooltip text of the icon and 
         mouse interactions previously set using addIcon(), to reflect the 
         values currently held by the properties .IconFilename, .ToolTip and
         .AllowIconEvents.

         return : true if OK, false if an error occured.

      deleteIcon( )
         This method will remove the icon from the taskbar tray.

         return : true if OK, false if an error occured.

      attachPopup( <cPopupMenuFilename>, <cPopupMenuClassname> )
         In order to display easily a popup menu near the icon in the taskbar
         tray, for instance when the user clicks on it, this method will load
         the <cPopupMenuFilename> popup menu in order to display it later, using
         the displayPopup method. The popup menu must be a VdB popup menu, that
         is inherited from the class Popup. This type of popup menus can be
         created with VdB's popup menu designer.

         <cPopupMenuFilename> : VdB string, the name of the file that contains
                                the popup menu class definition

         <cPopupMenuClassname> : VdB string, the classname of the popup menu
                                 to load. This name appears in the 
                                 <cPopupMenuFilename> file, in the class ..
                                 statement :
                                 class <cPopupMenuClassname>(...) of POPUP(...)

         return : true if OK, false if an error occured.

      displayPopup( [<bAtCursor>] )
         After attaching the popup menu with the attachPopup method, this method
         will display it. Two options are available for the popup position, depending
         on the bAtCursor flag.

         <bAtCursor> : optional VdB boolean, if not supplied, defaults to true.

                       - if true, the popup will appear sticked the mouse 
                         cursor. If you call this method in an mouse event when 
                         the cursor is over the icon, then the menu is displayed 
                         near the icon. But a call to this method if the cursor 
                         is anywhere on the screen will pop-up the menu at the 
                         cursor's position, not necessarily near the icon.

                       - if false, the popup will be displayed at a position
                         independant from the mouse cursor, and corresponding
                         to the center of the taskbar tray rectangle, where the
                         icons are displayed.

         return : true if OK, false if an error occured.

      detachPopup( )
         You can release a previously attached popup menu by a call to this method.

         return : true if OK, false if an error occured.

   -- Events:
      -------
      As this component is designed to handle mouse interactions with the icon
      added to the taskbar tray, the following events generated by the component 
      can be used under VdB, the same way as for the paintbox component :

          onMouseMove
          onLeftDblClick
          onLeftMouseDown
          onLeftMouseUp
          onMiddleDblClick
          onMiddleMouseDown
          onMiddleMouseUp
          onRightDblClick
          onRightMouseDown
          onRightMouseUp

      You can, for instance, in the form designer, select the component and 
      in the inspector's event tab, select an event and enter your code for it.

   -- Run in VdB IDE and Deployment information:
      -----------------------------------------

      1/ You need to have the samples directory installed as part of your
         VdB 7.01 installation.
      2/ To build an executable that includes Taskbartray, you need to:
         - add the following files to your project .prj description file:
            TaskbarTray.cc
            structure.prg      (Both are in the samples directory; leave them there
            structAPI.prg       simply reference them in the .prj file)
         - place a copy of the following .dll in the executable directory
            (you need to add it to your deployed application files)
            structmem.dll        (it is in the samples directory)


   ------------------------------------------------------------------------------
*/

// Necessary include files for API prototypes
#include <Vdbase.h>
#include <Winuser.h>
#include <StructAPI.h>
#include <Winmisc.h>

#define PNOTIFYICONDATA LPSTRUCTURE
#define METRIC_PIXELS    6
#define TASKTRAYERROR   "Error in the TaskbarTray component. :-("
#define TASKTRAYUID      123

// The TaskTray class is inherited from the paintBox class
class TaskbarTray( f ) of paintBox( f ) custom

   // used for structure API stuff
   set procedure to '&_dbwinhome.samples\structure.prg' additive

   // Reference some externs
   extern BOOL        Shell_NotifyIcon(DWORD, PNOTIFYICONDATA) shell32 ;
                      from "Shell_NotifyIconA"
   extern HANDLE      LoadImage(HINSTANCE, LPCSTR, UINT, CINT, ;
                      CINT, UINT) user32 from "LoadImageA"
   extern BOOL			 DestroyIcon(HANDLE) user32										//	added 11-Mar-2007 by Andrew Shimmin
   extern LRESULT     SendMessage(HANDLE, UINT, WPARAM, LPARAM) user32 ;
                      from "SendMessageA"
   extern BOOL        GetCursorPos(LPPOINT) user32
   extern SHORT       GetAsyncKeyState(CINT) user32
   extern HANDLE      FindWindowEx(HANDLE, HANDLE, LPCSTR, LPCSTR) user32 ;
                      from "FindWindowExA"
   extern BOOL        GetWindowRect(HANDLE, LPRECT) user32
   extern BOOL        SetForegroundWindow(HANDLE) user32
  
   // Set default properties and visual look when displayed in the form designer.
   with ( this )
      visible := false
      enabled := false
      height := 1
      width := 4
      tabStop := false
      borderStyle := 6   // Drop Shadow
      colorNormal := "0x80ffff/0x80ffff"
      printable := false
   endwith
   
   // Create the structures for the API 32 calls
   this.NOTIFYICONDATA = new _NOTIFYICONDATA( )
   this.TbTLpPoint = new tagPOINT( )
   this.TbTLpRect = new _RECT( )

   // Initialize some structure values used when the icon is setup.
   // Note especially that the uCallbackMessage is configured to send a WM_CHAR
   // message. When the mouse will interact with the icon, this message will be
   // sent by Windows and received by the component. The latter will consider it 
   // and process it as if it was a standard WM_CHAR message, and thus will 
   // generate a VdB onChar event. By creating in the component a VdB event 
   // handler hooked to the onChar event, we can be informed of and process 
   // the mouse interactions to the icon.
   this.NOTIFYICONDATA.setMember( "uID", TASKTRAYUID )
   this.NOTIFYICONDATA.setMember( "uCallbackMessage", WM_CHAR )

   // Setup the default component's properties values
   this.IconFilename = ""
   this.ToolTip = ""
   this.AllowIconEvents = true
   this.DisplayErrors = true

   //	Setup current icon properties

   this.nCurrentHIcon = 0																		//	current icon handle
   this.cCurrentIconFileName = ""															//	current icon filename

   // Setup the onChar event handler that will process the messages sent by 
   // the mouse interactions with the icon.
   this.onChar := this.TaskTray_onChar

   // The onClose event is used to perform some clean up when the component is
   // closed.
   this.onClose := this.TaskTray_onClose

   // These protections will have for effect to show only the useful stuff when
   // editing the component in the form designer. 
   protect  transparent, mousePointer, statusMessage, borderStyle, colorNormal, ;
            printable, visible, enabled, tabStop, pageno, ;
            TaskTray_onChar, TaskTray_onClose, onClose, onChar, ;
            onFormSize, onkeyDown, onKeyUp, onPaint, onLostFocus, onGotFocus, ;
            onDesignOpen, onOpen, setFocus, move, ;
            NOTIFYICONDATA, TbTLpPoint, TbTLpRect, SetupIconData, signalError, ;
            TbTPopupHolderForm


//--------------------------  Component Methods  -------------------------------


   // _------------------------------- IamLoaded -------------------------------
   // This method will add the capability for a cjeck to see if the class is
   //	loaded.

   function IamLoaded()
   	return

   // -------------------------------- addIcon ---------------------------------
   // This method will add the icon to the taskbar tray, using the .IconFilename
   // and .ToolTip properties.
   function addIcon( )
      // Check whether the form has it's metric property to Pixels.
      if this.form.metric <> METRIC_PIXELS
         this.signalError( "TaskTray requires the container form to" + ;
                           chr( 13 ) + "have its metric set to : Pixels (6)" )
         return false
      endif

      // Try to load the icon.
      if not this.SetupIconData( )
         return false
      endif

      // Display the icon
      return Shell_NotifyIcon( NIM_ADD, this.NOTIFYICONDATA.value )

   // ------------------------------- deleteIcon -------------------------------
   // This method will delete the icon on the taskbar tray, and unload the popup
   // menu, if any.
   function deleteIcon
   	this.releaseIcon()																		//	release icon resource
      this.detachPopup( )
      return Shell_NotifyIcon( NIM_DELETE, this.NOTIFYICONDATA.value )

   // ------------------------------- modifyIcon -------------------------------
   // This method will modify the current displayed icon, with the new .IconFilename
   // and .ToolTip properties.
   function modifyIcon
      // Try to load the new icon.
      if not this.SetupIconData( )
         return false
      endif

      // Display the icon
      return Shell_NotifyIcon( NIM_MODIFY, this.NOTIFYICONDATA.value )

	//	-------------------------------- releaseIcon -----------------------------
   //	This method releases an old icon resource if needed.
   function	releaseIcon

      if	this.nCurrentHIcon <> 0																//	current icon there?
         if	DestroyIcon(this.nCurrentHIcon)     										//	release current icon	ok?
         	this.nCurrentHIcon := 0															//	reset current icon properties
            this.cCurrentIconFilename = ""
         else
            this.signalError("Tasktray could not release icon : "+this.cCurrentIconFilename)
            return false
         endif	
      endif

      return true
   // ------------------------------- setupIconData ----------------------------
   // This method loads the icon parameters and initializes the icon structure.
   function setupIconData

      local nHwnd,nIconDelete

      //	check for a new icon

      if	.not.(this.IconFilename == this.cCurrentIconFilename)						//	new icon filename?	

         // Try to load the icon from its filename

         this.nHIcon = LoadImage( 0, this.IconFilename, IMAGE_ICON, 0, 0, ;
                                  LR_DEFAULTCOLOR + LR_DEFAULTSIZE + LR_LOADFROMFILE )
         if this.nHIcon == 0
            this.signalError( "Tasktray could not load icon : " + this.IconFilename )
            return false
         else

            //	release current icon if needed

            if	.not.this.releaseIcon()														//	current icon release error?
               return false
            endif

            //	save new icon as current icon

            this.nCurrentHIcon = this.nHIcon												//	save handle
            this.cCurrentIconFilename = this.IconFilename							//	save filename

         endif

      endif

      // initialize the icon structure

      nHwnd = this.hwnd
      this.NOTIFYICONDATA.setMember( "hWnd", nHwnd )
      this.NOTIFYICONDATA.setMember( "uFlags", NIF_ICON + NIF_TIP + NIF_MESSAGE )
      this.NOTIFYICONDATA.setMember( "hIcon" , this.nHIcon )
      this.NOTIFYICONDATA.setMember( "szTip", this.ToolTip )
      this.NOTIFYICONDATA.setMember( "uCallbackMessage" , iif( this.AllowIconEvents, WM_CHAR, WM_KEYUP ))

      return true

   // ------------------------------- attachPopup  -----------------------------
   // This method loads the supplied VdB popup menu, given its filename and
   // classname.
   function attachPopup( cPopupFilename, cPopupClassname )
      local exTbTray
      private cTemp

      // This creates an owned form, invisible, that will only be used to attach
      // the popup menu. By later setting the position of this form on the screen,
      // we can choose where the popup menu will be displayed, for instance, near
      // our taskbar tray icon.
      this.TbTPopupHolderForm = new TbTPopupHolderForm( )
      this.TbTPopupHolderForm.parent = this
      this.TbTPopupHolderForm.form = this.form

      // Load the popup menu filename
      try
         cTemp = cPopupFilename
         set procedure to &cTemp. additive
      catch( Exception exTbTray )
         this.signalError( "Could not load popup menu file : " + cPopupFilename ) 
         return false
      endtry

      // load the popup menu class, and attach it to the TbTPopupHolderForm form
      // just created.
      try
         cTemp = "this.TbTPopupHolderForm.popupMenu := new " + ;
                  cPopupClassname + "( this.TbTPopupHolderForm, 'TbTPopupHolderForm' )"
         &cTemp.
      catch( Exception exTbTray )
         this.signalError( "Could not load popup menu class : " + cPopupClassname ) 
         return false
      endtry

      // Open the form, but it will be invisible.
      this.TbTPopupHolderForm.open( )
      return true

   // ------------------------------- getPopup  -----------------------------
   // This method returns the VdB popup menu object reference.
   function getPopupObjRef( )
      return this.TbTPopupHolderForm.popupMenu

   // ------------------------------- detachPopup  -----------------------------
   // This method unloads the popup menu, if previously added.
   function detachPopup
      local bLoaded

      // Was there any popup attached ?
      bLoaded = ( type( "this.TbTPopupHolderForm" ) == 'O' )
      if bLoaded                       // Yes, so release.
         this.TbTPopupHolderForm.release( )
         this.TbTPopupHolderForm = Null
      endif
      return bLoaded

   // ------------------------------- displayPopup  -----------------------------
   // This method displays the popup menu, if previously added. Two possibilities
   // are given, either the menu appears at the mouse cursor location, or else
   // at a default position given by the center of the taskbar tray rectangle.
   function displayPopup( bAtCursor )
      local x, y, retCode, bAtCursorSel, nHwnd
      bAtCursorSel = iif( argcount( ) > 0, bAtCursor, true )   // Default is true

       // Popup attached ?
      if type( "this.TbTPopupHolderForm" ) == 'O'
         if bAtCursorSel       // case sticked to cursor
            // get the mouse cursor's position
            retCode = GetCursorPos( this.TbTLpPoint.value )
            x = this.TbTLpPoint.getMember( "x" )
            y = this.TbTLpPoint.getMember( "y" )
         else                  // case tray center
            // First, we need to retreive the handle to the taskbar tray. First
            // we start by the Windows explorer taskbar :
              nHwnd = FindWindowEx( 0, Null, "Shell_TrayWnd", Null )
            // Now, let's get a little bit deeper in the taskbar to get the tray
            // handle
              nHwnd = FindWindowEx( nHwnd, Null, "TrayNotifyWnd", Null )
            // With the taskbar tray handle, we retreive its coordinates
            retCode = GetWindowRect( nHwnd, this.TbTLpRect.value )
            x = (this.TbTLpRect.getMember( "left" ) + this.TbTLpRect.getMember( "right" ))/2
            y = (this.TbTLpRect.getMember( "top" ) + this.TbTLpRect.getMember( "bottom" ))/2
         endif

         // Set the TbTPopupHolderForm form's position to the desired location
         this.TbTPopupHolderForm.left := x
         this.TbTPopupHolderForm.top := y

         // The popup is at the good position. Now open it.
         SetForegroundWindow(this.hwnd)
         this.TbTPopupHolderForm.popupMenu.open( )
      else
         this.signalError( "No Popup menu attached, unable to open it!" ) 
         return false
      endif
      return retCode
 
   // This method is executed when the component is closed, it deletes the icon
   // and cleans up the TbTPopupHolderForm form, if any.
   function TaskTray_onClose
      return this.deleteIcon( )
      

   // This method handles the WM_CHAR windows message sent by the Tasktray's 
   // icon. The WM_CHAR message received after an interaction on the icon will 
   // not contain the usual information of such a message, but rather :
   //  - lParam contains the identifier of the mouse message that the system 
   //    generated as a result of the mouse interaction with the icon
   //  - wParam contains the identifier uID of the taskbar icon in which the mouse 
   //    event occurred
   // After being processed by VdB's message loop, the parameters supplied in the
   // onChar event will contain :
   //  - nChar contains wParam, that is the taskbar icon identifier,
   //  - nRepCnt contains LOW(lParam), that is the mouse message identifier.
   //  - nFlags contains 0
   function TaskTray_onChar(nChar, nRepCnt, nFlags)
      local e1, x, y, nMouseAndKbdFlags

      // Check if the received message is relevant
      if nChar <> this.NOTIFYICONDATA.getMember( "uID")
         this.signalError( "Tasktray received an unknown Icon identifier : " + nChar )
         return
      endif

      // We need to retreive the mouse position and the the flags state before
      // issuing the sendmessage that will be processed by VdB's message loop 
      // to generate the correponding on... VdB mouse event
      // First, get the mouse position :
      GetCursorPos( this.TbTLpPoint.value )
      x = this.TbTLpPoint.getMember( "x" )
      y = this.TbTLpPoint.getMember( "y" )

      // and then, the mouse buttons and keyboard shift/ctrl state :
      nMouseAndKbdFlags = 0
      if GetAsyncKeyState( VK_LBUTTON ) < 0
         nMouseAndKbdFlags += 1
      endif
      if GetAsyncKeyState( VK_RBUTTON ) < 0
         nMouseAndKbdFlags += 2
      endif
      if GetAsyncKeyState( VK_SHIFT ) < 0
         nMouseAndKbdFlags += 4
      endif
      if GetAsyncKeyState( VK_CONTROL ) < 0
         nMouseAndKbdFlags += 8
      endif
      if GetAsyncKeyState( VK_MBUTTON ) < 0
         nMouseAndKbdFlags += 16
      endif

      // Now that we have collected all the parameters, we can generate the 
      // WM_ message that corresponds to the mouse interaction on the icon.
      // This message will be received by VdB and transformed into the
      // equivalent mouse event for the component.
      sendMessage( this.hwnd, nRepCnt, nMouseAndKbdFlags, MAKELONG(x, y) )

      return

   // This function displays the error message if the .DisplayErrors property
   // is asserted.
   function signalError( cMessage )
      if this.DisplayErrors
         AlertMessage( cMessage, TASKTRAYERROR )
      endif
      return

endclass

// This form class is used only to attach the popup menu that is to be displayed.
// It stays always invisible, and is moved to the position at which the popup is
// to be displayed.
class TbTPopupHolderForm of form
   with ( this )
      mdi := false
      visible := false
      metric := METRIC_PIXELS
      text := "TaskbarTrayPopupHolder"
      OnLostFocus = class::FormOnLostFocus
   endwith
   
   Procedure FormOnLostFocus
      form.refresh()
      return 
      
endclass

// _NOTIFYICONDATA structure for shell_NotifyIcon
class _NOTIFYICONDATA of structure
   this.addMember( TYPE_DWORD , "cbSize" ) 
   this.addMember( TYPE_DWORD , "hWnd" ) 
   this.addMember( TYPE_DWORD , "uID" ) 
   this.addMember( TYPE_DWORD , "uFlags" )
   this.addMember( TYPE_DWORD , "uCallbackMessage" )
   this.addMember( TYPE_DWORD , "hIcon" )
   this.addMember( TYPE_STRING , "szTip" , 64 )
   this.setMember( "cbSize", this.length( ) ) 
endclass

// tagPOINT structure for GetCursorPos
class tagPOINT of structure // pt  
   this.addMember( TYPE_DWORD , "x" ) 
   this.addMember( TYPE_DWORD , "y" ) 
endclass

// _RECT structure for GetWindowRect
class _RECT of structure
   this.addMember( TYPE_DWORD , "left" )
   this.addMember( TYPE_DWORD , "top" )
   this.addMember( TYPE_DWORD , "right" )
   this.addMember( TYPE_DWORD , "bottom" )
endclass

