// -- File: APPFLAG.CC
class APPFLAG

/*
   ----------------------------------------------------------------
   Description:

      This class was written to test a unique system-wide global 
      flag for an application that can be used to determine if the
      application is already running.

      The following is by Jim Sare, posted on the VdBASE 
      Newsgroups:
      "For a brief understanding, Win32 (95 and NT) runs each 
      32-bit application as a separate process. Module handles 
      are _not_ shared across processes. This is a Good Thing. 
      <g> It means each process (application) is more effectively 
      isolated from other processes (applications). However, it 
      means that each time you start the _same_ application, 
      Win32 treats each instance of the app as a separate 
      process. Therefore, GetModuleUsage is useless, and in 
      fact is unsupported in Win32."

   Programmers: 

      Jim Sare, Mark Berry (MCB Systems), Andrew Shimmin

   History: 

      1/20/1998  -- Jim's original code (only modifications for 
                    dUFLP are the header ... and some formatting )
      12/10/1998 -- MCB modified to run under NT, which enforces 
                    security for semaphores. Original code left 
                    in as comments. Bumped to version 1.1.
      10-Nov-2007 - Andrew Shimmin - v1.2 - modified to create a global mutex
                    rather than a semaphore. This allows integration with the
                    INNO install system's application mutex for both install
                    and uninstall.
      10-Nov-2007 - Andrew Shimmin - v1.2.1 - modified to allow a kernel
                    namespace type mutex to be specified in support of terminal
                    services.
      12-Nov-2007 - Andrew Shimmin - v1.2.2 - added the MutexClose() method to
                    allow independent operation of the original code using a
                    semaphore and the new mutex code.
                    Fixed bug in original semaphore code that prohibited the
                    semaphore from being reused without re-instantiating the
                    AppFlag object.  
      26-Nov-2007	  Andrew Shimmin - v1.2.3 - added MutexReset() after testing
      				  using Windows XP fast user switching for Global\ mutex behaviour
                    across multiple user sessions.
      20-Feb-2008	  Andrew Shimmin - v1.2.4 - removed the EXTERN statement for the
      				  WTSGetActiveConsoleSessionId() API as it was not used and only
                    supported by WindowsXP and above. 
      21-Feb-2008	  Andrew Shimmin - v1.2.5 - added mutex context management during
      				  mutexCreate() method.

   SEMAPHORE Usage: 

      Place the following  in the beginning of the application's
      startup code:

      close all
      clear all
      release all

      set procedure to appflag.cc additive
         or
      set procedure to :dUFLP:appflag.cc additive
      _app.Flag = new AppFlag()
      if not _app.Flag.open( "MyUniqueFlagName" )
         if "RUNTIME" $ upper(version(0))
            quit
         endif
         RETURN
      endif
      // place the rest of your startup code here.

      // DO NOT FORGET:
      // In the SHUTDOWN code, add:
      //   _app.Flag.Close()
      // otherwise you will not be able to re-start the
      // software until you shutdown and restart your machine.

   MUTEX Usage: 

   	See the url http://msdn2.microsoft.com/en-us/library/ms682411.aspx for an
      explanation of kernel namespace types.

      set procedure to :dUFLP:appflag.cc additive

      _app.oMutex = new appFlag()

      if	_app.oMutex.MutexCreate("YourUniqueMutexName")								//	create a kernel local namespace mutex?

      	//	process items dependent on the mutex

         _app.oMutex.MutexClose()															//	close mutex

      else

      	if	"runtime"$lower(version(0))
         	quit
         else
         	//	do whatever you need to do
         endif

      endif

   ---------------------------------------------------------------------- 

*/

	#define SYNCHRONIZE              				(0x00100000)
	#define STANDARD_RIGHTS_REQUIRED 				(0x000F0000)
	#define SEMAPHORE_ALL_ACCESS 						(bitor(bitor(STANDARD_RIGHTS_REQUIRED,SYNCHRONIZE),0x3))

   this.Version = "0.1.2.5"

   this.SemaphoreName = ""                                                    //	semaphore properties
   this.Handle = 0

   this.cMutexName = ""																			//	mutex properties
   this.cMutexNamespaceType = "Local/"														//	default to the local kernel namespace
   this.nMutexHandle = 0
   this.cMutexStatus = ""
   this.nMutexSystemError = 0
   this.cMutexSystemError = ""

   class::Externals()																			//	define externals

   protect Version,SemaphoreName,Handle,CreateFlag

	function Externals()

      if type("OpenSemaphore") # "FP"
         extern CLONG OpenSemaphore(CLONG,CLOGICAL,CSTRING) Kernel32 from "OpenSemaphoreA"
      endif
      if type("CreateSemaphore") # "FP"
           extern CLONG CreateSemaphore(CPTR,CLONG,CLONG,CSTRING) Kernel32 from "CreateSemaphoreA"
      endif
      if type("ReleaseSemaphore") # "FP"
         extern CLONG ReleaseSemaphore(CHANDLE,CLONG,CPTR CLONG) Kernel32
      endif
      if type("CloseHandle") # "FP"
         extern CLOGICAL CloseHandle(CLONG) Kernel32
      endif
      if type("OpenMutexA") # "FP"
         extern CHANDLE OpenMutexA(CLONG,CLOGICAL,CSTRING) Kernel32
      endif
      if type("CreateMutexA") # "FP"
         extern CHANDLE CreateMutexA(CPTR,CLOGICAL,CSTRING) Kernel32
      endif
      if type("GetSystemMetrics") # "FP"
         extern CINT GetSystemMetrics(CINT) user32 from "GetSystemMetrics"
      endif
      if type("GetLastError") # "FP"
         extern CLONG GetLastError(CVOID) Kernel32
      endif
		if	type("FormatMessageA") # "FP"
      	extern CLONG FormatMessageA(CLONG,CLONG,CLONG,CLONG,CSTRING,CLONG,CLONG) Kernel32
      endif

   	return

   function GetVersion()
      return this.Version

   function Open(cName)

      local nHandle

      if len(ltrim(rtrim(cName))) = 0
         return false                      												// name must have a character value
      endif
      if len(this.SemaphoreName) > 0 or this.Handle # 0
         return false                      												// already exists
      endif

      this.SemaphoreName := ltrim(rtrim(cName))
      nHandle = OpenSemaphore(SEMAPHORE_ALL_ACCESS,false,this.SemaphoreName)

      if nHandle # 0                                                          // flag already exists?
         CloseHandle(nHandle)                                                 //	close OpenSemaphore handle
         this.Handle := 0                                                     //	als - original bug left the old handle intact
         this.SemaphoreName := ""                                             //	als - original bug left the old semaphore name intact
         return false                      												
      endif

   	return this.CreateFlag()        														// true on success, false on failure

   function Close()

      Local 	lReturn

      lReturn = true																				//	assume success

      Private	nRelease,lpPreviousCount

      nRelease = 0
      lpPreviousCount = 0

      if	this.Handle # 0																		//	semaphore?	
            lReturn = CloseHandle(this.Handle)                                //	close handle
            this.Handle := 0                                                  //	reset semaphore properties
            this.SemaphoreName := ""
      endif

   	return lReturn																				// true on success, false on failure

   function CreateFlag()

      if len(this.SemaphoreName) > 0
         this.Handle := CreateSemaphore(null,1,1,this.SemaphoreName)
         if	this.handle # 0																	//	semaphore created?
         	return true
         else
            this.Handle := 0																	//	reset semaphore properties
            this.SemaphoreName := ""
         endif
      endif

   	return false    																			// if here, semaphore name not specified

	function MutexCreate(cMutexName,cNamespaceType)
   //
   //	Function to create a mutex.
   //
   //	Input:
   //
   //		cMutexName			= mutex name (no backslashes allowed)
   //		cNamespaceType		= optional type of kernel namespace mutex (global or local, default is local)
   //
   //	Output:
   //
   //		true/false			= success/failure
   //		this.cMutexStatus	= result status string
   //
   //	Notes:
   //
   //		It appears the GetLastError() API does not work reliably in dBase and
   //		therefore the subtlties of the CreateMutexA() and OpenMutexA() API calls
   //		unfortunately cannot be used. This is the same for the semaphore API
   //		calls.
   //
   //		To alleviate this issue the mutex is opened first and if the open does
   //		NOT fail then the mutex already exists. If the open fails the mutex is
   //		created. If the create fails it is assumed there is a Global\ mutex or
   //		a Local\ semaphore of the same name.
   //
   //		My testing has revealed mutex and semaphore names need to be unique as
   //		an attempt to create a mutex of the same name as a semaphore may fail.
   //
   //		If the calling thread has already created a mutex of the same name the
   //		function will return TRUE when attempting to create the mutex again.
   //
   //		The dynamics among Global\ and Local\ mutexes is complex and if you wish
   //		to use both scopes in your application you need to test thoroughly your
   //		mutex strategy. The AppFlagTest.exe executable is provided for this
   //		purpose.
   //
   //		See the url http://msdn2.microsoft.com/en-us/library/ms682411.aspx for
   //		an explanation of kernel namespace types.
   //

   	Local		lNoError,lReturn

      lNoError = true																			//	assume no errors
      lReturn = true																				//	assume success

      Private	oContext,cMutexNameSpace,dwDesiredAccess,bInheritHandle,;
      			lpMutexAttributes,bInitialOwner,lpName
      	
      //	initialise

      oContext = new ContextObject()                                          //	create mutex context object
      oContext.lContextRestore = true														//	indicate to restore context

		#define SM_REMOTESESSION      				(0x00001000)
		#define MUTEX_ALL_ACCESS       				(0x001F0001)

*     dwDesiredAccess = STANDARD_RIGHTS_REQUIRED+SYNCHRONIZE						//	openMutexA arguments
*     dwDesiredAccess = null																	//	openMutexA arguments
      dwDesiredAccess = MUTEX_ALL_ACCESS													//	openMutexA arguments
      bInheritHandle = false

      lpMutexAttributes = null																//	createMutexA arguments
*     bInitialOwner = true
      bInitialOwner = false
      lpName = "MutexName"

      //	save current mutex context

      oContext.cMutexName = this.cMutexName												//	save mutex properties
      oContext.cMutexNamespaceType = this.cMutexNamespaceType					
      oContext.nMutexHandle = this.nMutexHandle
      oContext.cMutexStatus = this.cMutexStatus
      oContext.nMutexSystemError = this.nMutexSystemError
      oContext.cMutexSystemError = this.cMutexSystemError

      //	set kernel namespace type

      do	case
      	case	argcount() = 0
         		this.cMutexStatus := "Mutex name not provided"
               lNoError = false
               lReturn = false
      	case	argcount() = 1
               cMutexNamespace = "Local\"													//	default to a local kernel namespace type
      	otherwise
               cMutexNamespace = lower(ltrim(rtrim(argvector(2))))				//	set user kernel namespace type
         		do	case
               	case	cMutexNamespace == "local"
                        cMutexNamespace = "Local\"
               	case	cMutexNamespace == "global"
                        cMutexNamespace = "Global\"
                  otherwise
                  		cMutexNamespace = "Local\"                 				//	default namespace type
               endcase
      endcase

      //	check mutex name legality

		if	lNoError

         if	len(ltrim(rtrim(cMutexName))) <> 0											//	must supply a name	
            if	("\"$ltrim(rtrim(cMutexName)))											//	no backslashes
               this.cMutexStatus := "Mutex name contains a backslash"         //	set status
               lNoError = false																//	set error
               lReturn = false                                                //	return status state
            else
               if	(cMutexNamespace+ltrim(rtrim(cMutexName)) == oContext.cMutexName).and.(oContext.nMutexHandle # 0)	//	same mutex?
                  this.cMutexStatus := "Mutex already created"            		//	set status
                  lNoError = false															//	set error
                  lReturn = true                                             	//	return status state
               else
                  class::MutexReset()														//	reset mutex
                  this.cMutexNamespaceType := cMutexNamespace                    		//	set mutex namespace
                  this.cMutexName := this.cMutexNamespaceType+ltrim(rtrim(cMutexName))	//	set mutex name
               endif
            endif
         else
            this.cMutexStatus := "Mutex name not provided"                		//	set status
            lNoError = false																	//	set error
            lReturn = false                                                	//	return status state
         endif

      endif

      //	create mutex

      if	lNoError.and.;
      	(.not.empty(this.cMutexName)).and.;
         this.nMutexHandle = 0																//	no errors and no mutex existing?

         //	check if mutex already exists
               	
         this.nMutexHandle := OpenMutexA(dwDesiredAccess,;
                                         bInheritHandle,;
                                         this.cMutexName)							//	open mutex

*        ?program()+" - OpenMutexA handle = "+ltrim(str(this.nMutexHandle))

         class::MutexSystemError()															//	get system error

         if	this.nMutexHandle # 0															//	mutex already exist?

            if	class::MutexClose()															//	close mutex?
               this.cMutexStatus := "Mutex already exists"            			//	set status
            else
      			msgbox("Error closing mutex from openMutexA().",program(),48)
            endif
            lReturn = false                                             		//	return status state

         else

            //	create mutex

            this.nMutexHandle := CreateMutexA(lpMutexAttributes,;
                                              bInitialOwner,;
                                              this.cMutexName)						//	create mutex

*        	?program()+" - CreateMutexA handle = "+ltrim(str(this.nMutexHandle))

            class::MutexSystemError()														//	get system error

            if	this.nMutexHandle # 0                             					//	ok?
            	oContext.lContextRestore := false										//	do not restore mutex context
               this.cMutexStatus := "Mutex created"									//	set status
               lReturn = true																	//	return status state
            else
               class::MutexReset()															//	reset
               this.cMutexStatus := "Mutex already exists - cannot create mutex"	//	set status
               lReturn = false																//	return status state
            endif

         endif

      endif

      //	finish

      if	oContext.lContextRestore															//	restore context?

         this.cMutexName := oContext.cMutexName											//	restore mutex properties
         this.cMutexNamespaceType := oContext.cMutexNamespaceType					
         this.nMutexHandle := oContext.nMutexHandle

      endif

      oContext.close()																			//	release mutex context object

   	return lReturn

	function MutexReset()

      this.nMutexHandle := 0
      this.cMutexName := ""
      this.cMutexStatus := "Mutex reset"

      return

	function MutexClose()

		Local	lReturn

      lReturn = true																				//	assume success

      if this.nMutexHandle # 0																//	mutex?
         if	CloseHandle(this.nMutexHandle)												//	close?
            class::MutexReset()                                               //	reset mutex
         else
            class::MutexSystemError()
            this.cMutexStatus := "Mutex close error"
            lReturn = false
         endif
      else
      	this.cMutexStatus := "Mutex does noes exist"
         lReturn = false
      endif

   	return lReturn

	function MutexSystemError()

      Private	lpSource,dwMessageId,dwLanguageId,nSize,cMessage,va_list,TChars

		//	initialise

		#define FORMAT_MESSAGE_FROM_SYSTEM      	(0x00001000)

      lpSource = 0																				//	FormatMessageA arguments
      dwMessageId = GetLastError()
      dwLanguageId = 0
      nSize = 255
      this.cMutexSystemError := space(nSize)
      lpBuffer = this.cMutexSystemError
      va_list = null
      TChars = 0

      //	set mutex system error

      TChars = FormatMessageA(FORMAT_MESSAGE_FROM_SYSTEM,;
                              lpSource,;
                              dwMessageId,;
                              dwLanguageId,;
                              lpBuffer,;
                              nSize,;
                              va_list)

      this.nMutexSystemError := dwMessageId
      this.cMutexSystemError := trim(lpBuffer)

      //	finish

      if	this.nMutexSystemError > 0															//	system error?
      	msgbox(this.cMutexSystemError,program()+"- SystemErrorId = "+ltrim(str(this.nMutexSystemError,48)))
      endif

   	return
       
ENDCLASS
// -- End of Class: AppFlag

//	contextObject class

class ContextObject of Object custom

   this.className = "CONTEXTOBJECT"
   this.baseClassName = "OBJECT"

   function close()
      this.release()
      return

   function release()
      try
         release this
      catch	(exception e)
      endtry
      return true

endclass

/* 
   ----------------------------------------------------------------
   End of File: APPFLAG.CC
   ----------------------------------------------------------------
*/
