   /*
   --------------------------------------------------------
   Filename.....: colorDlg.cc
   ClassName....: colorDlg
   Purpose......: ChooseColor dialog.
   Programmer...: Rick Miller -  www.treturn.com/dbase
   Date.........: November 07, 2003
   Notes........: See Notes.
   Written for..: dB2K 0.1 and newer.
   Rev. History.: See Revisions.
   Dependencies.: VStruct32Ex.cc or VStruct.cc, VStruct.h.
   Usage........: set procedure to colorDlg.cc additive
   HelpFile.....: None.
   Example:.....: See Example.
   --------------------------------------------------------
   Example:
   --------------------------------------------------------
      oRef  =  new colorDlg()
      // assign property values as desired.
      // to assign an initial color:
      oRef.set("Color", any dbase color)
      // to change a custom color for 1 - 16:
      oRef.set("color1", <int> a window's color)
      // to display the dialog:
      oRef.show()
      // make selections.
      oRef.inspect()
      oRef.release()
   --------------------------------------------------------
   Notes:
   --------------------------------------------------------
      This file and it's dependencies
         are closed when calling release.
      Error messages do not take place in runtime.
      This class has built-in property value conversion
         for the different Color properties.
      Conversion of all Color properties takes place when
         either of the 2 cases below take place:
         1) a set("Color property name") is made and the
            return value is true.
            ie:
               oRef.set("htmlColor", "408080")
         2) a programmatic Color property assignment is made
            and a call to show() are made.
            ie:
               oRef.htmlColor := "408080"
               oRef.show()
   --------------------------------------------------------
   Properties:          type:       default:
   --------------------------------------------------------
      Color             <char>      "black"
         a dBASE compatible color.
      colorStyle        <int>          0
         0 = Java, 1 = dBASE, 2 = System, 3 = Hex.
         preferred style of Color property.
      fullOpen          <bool>      true
         open with custom selection.
      hexColor          <char>      "000000"
         hexColor is a (window's color in hex (bgr)).
      htmlColor         <char>      "000000"
         htmlColor is a (window's color in hex (rgb).
      htmlColorSafe     <char>      "000000"
         nearest color safe htmlColor.
      iniFile           <char>      _app.exe + ".ini"
      javaColor         <char>      "black"
         format is either "#408080"
         or a java color name "teal"
      messages          <bool>      true
         enable/disable error messages.
      ownerHandle       <handle>    activeWindow.
      preventFullOpen   <bool>      false
         allow custom selection ?
      showHelp          <bool>      false
         display the help button ?
      winColor          <int>          0
         window's color.
   --------------------------------------------------------
   Methods:
   --------------------------------------------------------
      get(<char> property name)
         get a property or member value.
      getColor([<char> dBASE color])  display dialog.
         return "" or this.color.
      inspect()   display object in the inspector.
      loadIni()   adds custom colors from iniFile.
      release()   release and close procedure files.
      saveIni()   writes custom colors to iniFile.
      set(<char> property name, <pValu> new value)
         change a property or member.
      show()      display dialog.
   --------------------------------------------------------
   Revisions:
   --------------------------------------------------------
      February 10, 2005
         1) renamed all external functions.
      November 15, 2004
         1) added getColor method.
            displays dialog and returns either
            an empty string or this.color.
         2) modified code to insure the integrity of
            custom colors.
         3) added "messages" property to enable and disable
            error messages.
         4) default folder for the iniFile property
            using set("iniFile", "filename") without
            any pathing defaults the folder to:
               a) _dbwinhome folder in runtime.
               b) set("directory") folder in the dBASE IDE.
         5) added better error handling when Vstruct file
            is not available during instantiation.
      October 04, 2004
         1) changed to a dynamic lookup for system colors.
      October 02, 2004
         1) bug fix for x_IsDbaseColor.
         2) improved integrity in the set method.
      October 01, 2004
         1) added property colorStyle <int> 0-3.
            used for determining the preferred style
            of the Color property.
            ie: Java name, dBASE name, System name, Hex.
      September 30, 2004
         1) added property javaColor.
            javaColor is formatted as:
            either:
               a) a java color name ie: "teal"
            or:
               b) a color value ie: "#408080"
         2) updated all methods for support
            of the javaColor property.
      September 29, 2004
         1) added code to handle an assignment to "Color"
            that may have a "/" in the string.
            the string before the "/" is used.
      September 28, 2004
         Thanks to Rich Mueller (AutoTraker)
            for code samples.
         I Apologize for any inconvenience caused by the
            changes below.
         The changes were made in an effort for the control
            to be more friendly to the dBASE developer.
         Thought it would be best to make the changes now.
         1) added property "Color" supporting a dBASE color.
            a dBASE color.
         2) changed property name from "rgbColor" to "winColor".
         3) added property "iniFile".
            ini file to manage custom colors.
         4) added method loadIni.
            a loadIni() call is made in the constructor.
         5) added method saveIni.
            a saveIni() call is made in the release method.
         6) added compatibility to :dUFLP:Vstruct.prg.
         7) changed "isvalid_rgb" macro to "isvalid_win".
         8) adjusted code for these changes.
      September 27, 2004
         1) added property htmlColorSafe.
            closest html colorSafe color to the selection
            (thanks to Michael Nuwer).
         2) added code to insure that when a color is
            programmatically assigned that the color item
            is in the dialog for selection.
            when the color is not already in the dialog,
            it is placed in the custom color palette at
            position 16.
      September 26, 2004
         1) added get, set, x_CustomColor methods.
            these methods provide the oppurtunity to
            add and retrieve custom colors.
         2) changed code for compatibility with VStruct.cc.
      September 25, 2004
         1) updated code in x_UpdateColor function
            to deal with a programmatic change to
            the hexColor, htmlColor, or winColor.
            property value.
         2) reviewed and updated comments.
   */
// VStruct constants.
#include VStruct.h

// ChooseColor Dialog Flag constants.
#define  CC_RGBINIT                 0x00000001
#define  CC_FULLOPEN                0x00000002
#define  CC_PREVENTFULLOPEN         0x00000004
#define  CC_SHOWHELP                0x00000008
#define  CC_ENABLEHOOK              0x00000010
#define  CC_ENABLETEMPLATE          0x00000020
#define  CC_ENABLETEMPLATEHANDLE    0x00000040
#define  CC_SOLIDCOLOR              0x00000080
#define  CC_ANYCOLOR                0x00000100

// global alloc constants.
#define  GMEM_MOVEABLE              0x00000002
#define  GMEM_ZEROINIT              0x00000040

// system color name constants used by dBASE.
// provided by Rich Mueller.
#define  COLOR_SCROLLBAR            0     // The color of a scroll bar's gray area.
#define  COLOR_BACKGROUND           1     // The desktop color. 
#define  COLOR_ACTIVECAPTION        2     // The active window title bar color. Windows 98, 2000: The color of the left side of the active window title bar gradient, if the gradient effect is used.
#define  COLOR_INACTIVECAPTION      3     // The color of an inactive window's caption. Windows 98, 2000: The color of the right side of an inactive window's title bar gradient, if the gradient effect is used.
#define  COLOR_MENU                 4     // The background color of menus.
#define  COLOR_WINDOW               5     // The background color of a window.
#define  COLOR_WINDOWFRAME          6     // The color of a window frame.
#define  COLOR_MENUTEXT             7     // The color of menu text.
#define  COLOR_WINDOWTEXT           8     // The color of text in a window.
#define  COLOR_CAPTIONTEXT          9     // The color of window caption text, size boxes, and scroll bar arrow boxes.
#define  COLOR_ACTIVEBORDER         10    // The active window border color.
#define  COLOR_INACTIVEBORDER       11    // The color of an inactive window's border.
#define  COLOR_APPWORKSPACE         12    // The background color of multiple document interface (MDI) windows.
#define  COLOR_HIGHLIGHT            13    // The color used to highlight selected items.
#define  COLOR_HIGHLIGHTTEXT        14    // The color used for the text of highlighted items.
#define  COLOR_BTNFACE              15    // The color of a button's face.
#define  COLOR_BTNSHADOW            16    // The color of a button's shadow area.
#define  COLOR_GRAYTEXT             17    // The color for disabled (grayed-out) text.
#define  COLOR_BTNTEXT              18    // The text color for pushbuttons.
#define  COLOR_INACTIVECAPTIONTEXT  19    // The color of an inactive window's caption text.
#define  COLOR_BTNHIGHLIGHT         20    // The color of a button's highlight area.
#define  COLOR_INFOTEXT             23    // The text color for tooltip controls.
#define  COLOR_INFOBK               24    // The background color for tooltop controls.

// programmer defined macros.
#define  ghex_color(n)              right("000000" + ItoH(n), 6)
#define  ghtml_color(s)             s.substring(4, 6) +;
                                    s.substring(2, 4) +;
                                    s.substring(0, 2)
#define  isvalid_win(n)             n > -1 and n < 16777216
#define  zeros(n)                   replicate(chr(0), n)

// programmer defined constants.
#define  COLORDLG_DEFAULTCOLOR      "black"
#define  COLORDLG_PUBLISHERINFO     "Rick Miller"
#define  COLORDLG_VERSIONINFO       "5.0210"
#define  ERROR_COLORDLG_1(p,s)      "2nd parameter " + p +;
                                    " was a type '" + s +;
                                    "'" + chr(13) + "expecting "
#define  ERROR_COLORDLG_3           "'C' a character type."
#define  ERROR_COLORDLG_4           "'N' a numeric type."
#define  ERROR_COLORDLG_5           "'L' a logical type."
#define  ERROR_COLORDLG_6(p)        "2nd parameter " + p +;
                                    " was invalid."
#define  ERROR_COLORDLG_7(s)        "for " + s  + " is inValid."
#define  ERROR_COLORDLG_8(s)        "2nd parameter " + s +;
                                    " folder is not available."
#define  ERROR_COLORDLG_9(s)        "1st parameter " + s +;
                                    " is out of range" +;
                                    chr(13) +;
                                    "expecting color1 to color16."
#define  ERROR_DEFAULT_SET          "Set request "
//---------------------------------------------------------------//
//---------------------------------------------------------------//
class colorDlg
   this.className =  "colorDlg"

   this.debug     =  false
   this.messages  =  true
   this.versioninfo  =  COLORDLG_VERSIONINFO
   this.publishinfo  =  COLORDLG_PUBLISHERINFO
   this.~runtime  =  iif("run" $ lower(version(0)), true, false)
   this.~procfile =  class::x_GetProcfile()

   if not empty(this.~procFile)
      class::x_Protect()
      class::x_InitExterns()
      class::x_Initialize()
   else
      class::release()
   endif

   //------------------------------------------------------------//
   //       end of constructor   -  methods below
   //------------------------------------------------------------//
   //------------------------------------------------------------//
   // return a property or member value.
   function get(cMember)
      Local cKey, pRet
      cKey  =  lower("" + cMember)
      pRet  =  null
      if cKey == "color"
         pRet  := (this.Color)
      elseif cKey == "colorstyle"
         pRet  := int(this.colorStyle)
      elseif cKey == "fullopen"
         pRet  := iif(this.fullopen, true, false)
      elseif cKey == "hexcolor"
         pRet  := (this.hexColor)
      elseif cKey == "htmlcolor"
         pRet  := (this.htmlColor)
      elseif cKey == "htmlcolorsafe"
         pRet  := (this.htmlColorSafe)
      elseif cKey == "javacolor"
         pRet  := (this.javaColor)
      elseif cKey == "inifile"
         pRet  := (this.iniFile)
      elseif cKey == "messages"
         pRet  := iif(this.messages, true, false)
      elseif cKey == "preventfullopen"
         pRet  := iif(this.preventFullOpen, true, false)
      elseif cKey == "showhelp"
         pRet  := iif(this.showhelp, true, false)
      elseif cKey == "wincolor"
         pRet  := int(this.winColor)
      elseif cKey.left(5) == "color"
         Local nColr
         nColr =  val(cKey.substring(5, cKey.length))
         if nColr > 0 and nColr < 17
            pRet  := this.custColors.getMem( ;
               "color" + str(nColr, 2, 0, "0"))
         endif
      endif
      if pRet == null
         class::x_ErrorMessage( ;
            "Get request for " + cMember + " is inValid")
      endif
   return pRet

   //------------------------------------------------------------//
   // method for those wishing to see a character value returned.
   function getColor(cColr)
      if type("argVector(1)") == "C"
         class::set("color", cColr)
      endif
   return iif(class::show(), "" + this.color, "")

   //------------------------------------------------------------//
   function inspect
      inspect(this)
   return

   //------------------------------------------------------------//
   // load the custom color palette from an ini file entry.
   function loadIni
      Local cColr, i, nColr, nPos1, nPos2
      // get the string of custom colors use.
      cColr =  class::x_ReadIni("CustomColors", "Colors")
      if not empty(cColr)
         // parse the string and
         // add the custom colors into the dialog.
         nPos1 =  0
         for i = 1 to 16
            nPos2 =  cColr.indexof(",", nPos1)
            if nPos2 == -1
               // last piece of the string.
               nPos2 := cColr.length
            endif
            nColr =  val(cColr.substring(nPos1, nPos2))
            if isvalid_win(nColr)
               this.custColors.setMem("color" + str(i, 2, 0, "0"), nColr)
            endif
            if nPos2 == cColr.length
               // assign i to exit the loop.
               i  := 20
            endif
            nPos1 := nPos2 + 1
         endfor
      endif
   return

   //------------------------------------------------------------//
   // this file and it's dependencies
   // are closed when calling here.
   function release
      class::saveIni()
      if type("this.chooseColor") == "O"
         release object this.chooseColor
         this.chooseColor  := null
      endif
      if type("this.custColors") == "O"
         release object this.custColors
         this.custColors   := null
      endif
      if type("this.dlgColors") == "A"
         release object this.dlgColors
         this.dlgColors := null
      endif
      if type("this.dbaseColors") == "O"
         release object this.dbaseColors
         this.dbaseColors  := null
      endif
      if type("this.javaColors") == "O"
         release object this.javaColors
         this.javaColors   := null
      endif
      if type("this.sysColors") == "O"
         release object this.sysColors
         this.sysColors   := null
      endif
      if type("this.sysNames") == "O"
         release object this.sysNames
         this.sysNames  := null
      endif
      class::x_CloseProcedure(this.~procfile)
      try
         release object this
      catch(exception e)
      endtry
      class::x_CloseProcedure(program(1))
   return

   //------------------------------------------------------------//
   // save the custom color palette to an ini file entry.
   function saveIni
      Local cColr
      cColr =  ""
      // the ini entry is a single string.
      // so put the palette entries into 1 string.
      for i = 1 to 16
         cColr += this.custColors.getMem( ;
                  "color" + str(i, 2, 0, "0"))
         if i < 16
            cColr += ","
         endif
      endfor
      class::x_WriteIni("CustomColors", "Colors", cColr)
   return

   //------------------------------------------------------------//
   // return true/false (success/fail)
   // set a property or member value.
   // the cMes? are used for a developer aid.
   function set(cMember, pValu)
      Local bRet, bUpd, cKey, cGet, cMes1, cMes2, cType
      bRet  =  false
      bUpd  =  false
      cKey  =  lower("" + cMember)
      cType =  type("argVector(2)")
      cMes1 =  ERROR_COLORDLG_1(pValu, cType)
      cMes2 =  ""
      if cKey == "color"
         if cType == "C"
            // evaluate proposed assignment value.
            cGet  =  class::x_IsDbaseColor(pValu.leftTrim().rightTrim())
            bUpd  := not empty(cGet)
            if bUpd
               this.Color  := cGet
            else
               cMes1 := ERROR_COLORDLG_6(pValu)
            endif
         else
            cMes2 := ERROR_COLORDLG_3
         endif
      elseif cKey == "colorstyle"
         if cType == "N"
            this.colorStyle   := int(mod(pValu, 4))
            bUpd  := true
         else
            cMes2 := ERROR_COLORDLG_4
         endif
      elseif cKey == "fullopen"
         if cType == "L"
            this.fullopen  := iif(pValu, true, false)
            bRet  := true
         else
            cMes2 := ERROR_COLORDLG_5
         endif
      elseif cKey == "hexcolor"
         if cType == "C"
            // evaluate proposed assignment value.
            cGet  =  class::x_IsHtmlColor(pValu)
            bUpd  := not empty(cGet)
            if bUpd
               // convert to a 6 char string.
               this.hexColor  := ghex_color(HtoI(cGet))
            else
               cMes1 := ERROR_COLORDLG_6(pValu)
            endif
         else
            cMes2 := ERROR_COLORDLG_3
         endif
      elseif cKey == "htmlcolor"
         if cType == "C"
            // evaluate proposed assignment value.
            cGet  =  class::x_IsHtmlColor(pValu)
            bUpd  := not empty(cGet)
            if bUpd
               // convert to a 6 char string.
               this.htmlColor  := ghex_color(HtoI(cGet))
            else
               cMes1 := ERROR_COLORDLG_6(pValu)
            endif
         else
            cMes2 := ERROR_COLORDLG_3
         endif
      elseif cKey == "javacolor"
         if cType == "C"
            // evaluate proposed assignment value.
            cGet  =  class::x_IsJavaColor(pValu)
            bUpd  := not empty(cGet)
            if bUpd
               if cGet.left(1) == "#"
                  cGet  := class::x_UpdateJavaColor(HtoI( ;
                     ghtml_color(ghex_color(HtoI( ;
                     cGet.substring(1, cGet.length))))))
               endif
               this.javaColor  := cGet
            else
               cMes1 := ERROR_COLORDLG_6(pValu)
            endif
         else
            cMes2 := ERROR_COLORDLG_3
         endif
      elseif cKey == "wincolor"
         if cType == "N"
            // evaluate proposed assignment value.
            bUpd  := isvalid_win(pValu)
            if bUpd
               this.winColor  := int(pValu)
            else
               cMes1 := ERROR_COLORDLG_6(pValu)
            endif
         else
            cMes2 := ERROR_COLORDLG_4
         endif
      elseif cKey.left(5) == "color"
         Local nColr
         nColr =  val(cKey.substring(5, cKey.length))
         // evaluate proposed value.
         if nColr > 0 and nColr < 17 and cType == "N"
            // call to try and make assignment.
            bRet  := class::x_CustomColor(nColr, pValu)
         else
            if cType == "N"
               cMes1 := ERROR_COLORDLG_9(cMember)
            else
               cMes2 := ERROR_COLORDLG_4
            endif
         endif
      elseif cKey == "inifile"
         if cType == "C"
            Local cFold, cIni, nPos
            cIni  =  "" + pValu
            nPos  =  cIni.lastindexof("\")
            if nPos == -1
               if this.~runtime
                  cFold =  _dbwinhome.left(_dbwinhome.length - 1)
               else
                  cFold =  set("directory")
               endif
               cIni  := cFold + "\" + pValu
            else
               cFold =  cIni.left(nPos)
            endif
            bRet  := not empty(funique(cFold + "\????????"))
            if bRet
               this.iniFile   := "" + cIni
            else
               cMes1 := ERROR_COLORDLG_8(cFold)
            endif
         else
            cMes2 := ERROR_COLORDLG_3
         endif
      elseif cKey == "messages"
         if cType == "L"
            this.messages  := iif(pValu, true, false)
            bRet  := true
         else
            cMes2 := ERROR_COLORDLG_5
         endif
      elseif cKey == "preventfullopen"
         if cType == "L"
            this.preventFullopen := iif(pValu, true, false)
            bRet  := true
         else
            cMes2 := ERROR_COLORDLG_5
         endif
      elseif cKey == "showhelp"
         if cType == "L"
            this.showhelp  := iif(pValu, true, false)
            bRet  := true
         else
            cMes2 := ERROR_COLORDLG_5
         endif
      else
         cMes1 := ERROR_COLORDLG_7(cMember)
      endif
      if bUpd
         // update all Color property values.
         class::x_UpdateColor()
         bRet  := true
      else
         if not bRet
            class::x_ErrorMessage(ERROR_DEFAULT_SET + cMes1 + cMes2)
         endif
      endif
   return iif(bRet, true, false)

   //------------------------------------------------------------//
   // display the chooseColor dialog.
   function show
      class::x_CheckDialogColors()
      Local bRet, hMem, nSize, oColr, oCust, pMem

      // create temporary objects to work with.
      oCust =  COLORDLG_Structures::custColorsStructure()
      oColr =  COLORDLG_Structures::chooseColorStructure()

      // update the temporary objects using the results
      // from the last call or this initial values.
      oCust.string   =  (this.custColors.string)
      oColr.string   =  (this.chooseColor.string)

      // allocate global memory for a custColors structure.
      nSize =  oCust.getSize()
      hMem  =  RMM_GlobalAlloc( ;
         GMEM_MOVEABLE + GMEM_ZEROINIT, nSize)
      // increment lock count and get an address.
      pMem  =  RMM_GlobalLock(hMem)
      // copy the temporary custColors structure
      // to global memory at address pMem.
      RMM_CopyMemoryTo(pMem, oCust, nSize)

      // update any changes from property values to a
      // few members of the temporary chooseColor structure.
      // set the hwndOwner member.
      oColr.setMem('hwndOwner', class::x_UpdateOwner())
      // set the initial color member.
      oColr.setMem('rgbResult', class::x_UpdateColor())
      // set the lpCustColors pointer.
      oColr.setMem('lpCustColors', pMem)
      // set the Flags member.
      class::x_UpdateFlags()
      oColr.setMem('Flags', class::x_UpdateFlags())

      release object oCust
      oCust := null

      // open the dialog.
      bRet  =  RMM_ChooseColor(oColr)
      if bRet
         // user made a selection.
         // get the size (may have changed).
         nSize := RMM_GlobalSize(hMem)
         // create a new string to copy to.
         oCust := zeros(nSize)
         // copy nSize information at address pMem to oCust.
         RMM_CopyMemoryFrom(oCust, pMem, nSize)
         // create object property for re-use and alternate use.
         this.custColors   =  COLORDLG_Structures::custColorsStructure()
         // place the information into a property.
         this.custColors.string  =  oCust.string
         // create object property for re-use and alternate use.
         this.chooseColor  =  COLORDLG_Structures::chooseColorStructure()
         // place the information into a property.
         this.chooseColor.string =  oColr.string
         // call to update the property values.
         class::x_Update()
      endif

      // decrement the lock count.
      RMM_GlobalUnlock(hMem)
      // free the global memory.
      RMM_GlobalFree(hMem)
      // release the string objects.
      release object oColr
      oColr := null
      oCust := null
   return iif(bRet, true, false)

   //------------------------------------------------------------//
   // insure current rgbResult is a selection in the dialog.
   function x_CheckDialogColors
      Local nRgb
      nRgb  =  int(this.rgbResult)
      if this.dlgColors.scan(nRgb) == 0
         // rgbResult is a non-standard color,
         // see if rgbResult is in the custom colors.
         Local i
         for i = 1 to 16
            if this.custColors.getMem( ;
               "color" + str(i, 2, 0, "0")) == nRgb
               // rgbResult is in custom colors.
               i  := 25
            endif
         endfor
         if i < 20
            // rgbResult is not in the custom colors.
            // put rgbResult into custom color 16.
            this.custColors.setMem("color16", nRgb)
         endif
      endif
   return

   //------------------------------------------------------------//
   // try and close a procedure file.
   function x_CloseProcedure(cProc)
      try
         close procedure (cProc)
      catch(exception e)
      endtry
   return

   //------------------------------------------------------------//
   // change a custom color member.
   function x_CustomColor(nColr, nRgb)
      Local bRet, cMemb, nRgb
      cMemb =  "color" + str(nColr, 2, 0, "0")
      nMemb =  this.custColors.getMem(cMemb)
      // evaluate proposed member change.
      // is the proposed assignment value valid.
      // is this.rgbResult already using the current member.
      bRet  =  isvalid_win(nRgb) and ;
               not this.rgbResult == nMemb
      if bRet
         // replace a current custom color.
         this.custColors.setMem(cMemb , nRgb)
      endif
   return iif(bRet, true, false)

   //------------------------------------------------------------//
   function x_ErrorMessage(cBody)
      if lower("" + this.messages) == "true"
         msgbox(cBody, "  colorDlg Message", 64)
      endif
   return

   //------------------------------------------------------------//
   // return the VStruct file that will be used.
   function x_GetProcfile
      Local cRet
      cRet  =  ""
      try
         set procedure to VStruct.cc additive
         cRet  := "VStruct.cc"
      catch(exception e)
         try
            set procedure to VStruct32Ex.cc additive
            cRet  := "VStruct32Ex.cc"
         catch(exception e)
            try
               set procedure to VStruct.prg additive
               cRet  := "VStruct.prg"
            catch(exception e)
               try
                  set procedure to :dUFLP:VStruct.prg additive
                  cRet  := ":dUFLP:VStruct.prg"
               catch(exception e)
                  class::x_ErrorMessage("VStruct files are missing.")
               endtry
            endtry
         endtry
      endtry
   return cRet

   //------------------------------------------------------------//
   // return the closest html color safe font color item.
   function x_HtmlColorSafeItem(nColr)
      Local nMod
      nMod  =  mod(nColr, 51)
      nColr += iif(nMod > 25, (51 - nMod), (-nMod))
   return right("00" + ItoH(nColr), 2)

   //------------------------------------------------------------//
   // initialize external functions.
   function x_InitExterns
      //---------------------------------------------------------//
      //    dialog functions.
      //---------------------------------------------------------//
      if not type("RMM_ChooseColor") == "FP"
         extern CLOGICAL RMM_ChooseColor(CPTR) ;
            comdlg32 from "ChooseColorA"
      endif
      if not type("RMM_CommDlgExtendedError") == "FP"
         extern CLONG RMM_CommDlgExtendedError(CVOID) ;
            comdlg32 from "CommDlgExtendedError"
      endif
      if not type("RMM_CopyMemoryFrom") == "FP"
         extern CVOID RMM_CopyMemoryFrom(CPTR, CPTR, CULONG) ;
            kernel32 from "RtlMoveMemory"
      endif
      if not type("RMM_CopyMemoryTo") == "FP"
         extern CVOID RMM_CopyMemoryTo(CLONG, CPTR, CULONG) ;
            kernel32 from "RtlMoveMemory"
      endif
      if not type("RMM_GetActiveWindow") == "FP"
         extern CHANDLE RMM_GetActiveWindow(CVOID) ;
            user32 from "GetActiveWindow"
      endif
      if not type("RMM_GlobalAlloc") == "FP"
         extern CHANDLE RMM_GlobalAlloc(CUINT, CULONG) ;
            kernel32 from "GlobalAlloc"
      endif
      if not type("RMM_GlobalFree") == "FP"
         extern CHANDLE RMM_GlobalFree(CHANDLE) ;
            kernel32 from "GlobalFree"
      endif
      if not type("RMM_GlobalLock") == "FP"
         extern CLONG RMM_GlobalLock(CHANDLE) ;
            kernel32 from "GlobalLock"
      endif
      if not type("RMM_GlobalSize") == "FP"
         extern CULONG RMM_GlobalSize(CHANDLE) ;
            kernel32 from 'GlobalSize'
      endif
      if not type("RMM_GlobalUnLock") == "FP"
         extern CLOGICAL RMM_GlobalUnLock(CHANDLE) ;
            kernel32 from "GlobalUnlock"
      endif
      //---------------------------------------------------------//
      //    ini functions.
      //---------------------------------------------------------//
      if not type("RMM_GetIniString") == "FP"
         extern CULONG RMM_GetIniString( ;
            CSTRING, CSTRING, CSTRING, CSTRING, CULONG, CSTRING) ;
            kernel32 from "GetPrivateProfileStringA"
      endif
      if not type("RMM_WriteIniString") == "FP"
         extern CLOGICAL RMM_WriteIniString( ;
            CSTRING, CSTRING, CSTRING, CSTRING) ;
            kernel32 from "WritePrivateProfileStringA"
      endif
      //---------------------------------------------------------//
      //    misc. functions.
      //---------------------------------------------------------//
      if not type("RMM_GetSysColor") == "FP"
         extern CULONG RMM_GetSysColor(CINT) user32 from "GetSysColor"
      endif
      if not type("RMM_GetWindowRect") == "FP"
         extern CLOGICAL RMM_GetWindowRect(CHANDLE, CPTR) ;
            user32 from "GetWindowRect"
      endif
   return

   //------------------------------------------------------------//
   function x_Initialize
      this.fullOpen        =  true
      this.rgbResult       =  0
      this.Color           =  COLORDLG_DEFAULTCOLOR
      this.colorStyle      =  0
      this.hexColor        =  "000000"
      this.htmlColor       =  "000000"
      this.htmlColorSafe   =  "000000"
      this.iniFile         =  _app.exename.left( ;
                              _app.exename.lastindexof(".")) + ".ini"
      this.javaColor       =  "black"
      this.winColor        =  0
      this.preventFullOpen =  false
      this.ownerHandle     =  0
      this.showHelp        =  false
      this.chooseColor     =  COLORDLG_Structures::chooseColorStructure()
      this.custColors      =  COLORDLG_Structures::custColorsStructure()
      class::x_LoadCustColors()
      class::x_LoadColors()
      class::loadIni()
   return

   //------------------------------------------------------------//
   // called from the set method.
   // used to decide whether the parameter is a valid dBASE Color.
   // check the keys of the lookup arrays for a match.
   // sequence is determined by the colorStyle property.
   function x_IsDbaseColor(cValu)
      Local cRet, cKey, cOrd, i, nChck
      cKey  =  cValu.leftTrim().rightTrim().toLowerCase()
      if not cKey.indexof("/") == -1
         // support a passed value like colorNormal.
         // ie: "foregrond/background"
         cKey  := cKey.left(cKey.indexof("/"))
      endif
      cRet  =  ""
      nChck =  int(mod(val("" + this.colorStyle), 4))
      if nChck == 0
         cOrd  =  "0,1,2,3"
      elseif nChck == 1
         cOrd  =  "1,0,2,3"
      elseif nChck == 2
         cOrd  =  "2,0,1,3"
      elseif nChck == 3
         cOrd  =  "3,0,1,2"
      endif
      for i = 0 to 3
         nChck := int(val(cOrd.substring(i * 2, i * 2 + 1)))
         if nChck == 3
            if cKey.length > 2 and cKey.left(2) == "0x"
               // return "0x" + 6 character string in uppercase.
               cRet  := "0x" + ghex_color(HtoI( ;
                  cValu.substring(2, cValu.length)))
            endif
         elseif nChck == 2
            if this.sysColors.iskey(cKey)
               cRet  := this.sysNames[cKey]
            endif
         elseif nChck == 1
            if this.dbaseColors.iskey(cKey)
               cRet  := cKey
            endif
         elseif nChck == 0
            if this.javaColors.iskey(cKey)
               cRet  := cKey
            endif
         endif
         if not empty(cRet)
            i  := 8
         endif
      endfor
   return cRet

   //------------------------------------------------------------//
   // called from the set method.
   // used to decide whether the parameter is a valid html Color.
   function x_IsHtmlColor(cValu)
      Local cRet, cKey
      cRet  =  ""
      cKey  =  cValu.leftTrim().rightTrim().toUpperCase()
      if cKey.length > 0
         Local i, nCnt
         nCnt  =  cKey.length
         for i = 0 to nCnt - 1
            if not cKey.charAt(i) $ "01234567890ABCDEF"
               // invalid character.
               i  := nCnt + 10
            endif
         endfor
         if i < nCnt + 10
            cRet  := cKey
         endif
      endif
   return cRet

   //------------------------------------------------------------//
   // called from the set and x_UpdateColor method.
   // used to decide whether the parameter is a valid java Color.
   function x_IsJavaColor(cValu)
      Local cKey, cRet
      cRet  =  ""
      cKey  =  cValu.leftTrim().rightTrim().toLowerCase()
      if cKey.left(1) == "#"
         if cKey.length > 1
            // call for a valid character check,
            // with the leading # sign stripped off.
            cKey  := class::x_IsHtmlColor( ;
               cKey.substring(1, cKey.length))
            if not empty(cKey)
               // return "#" + 6 character string in uppercase.
               cRet  := "#" + ghex_color(HtoI(cKey))
            endif
         endif
      else
         if this.javaColors.iskey(cKey)
            // return a java color in lowercase.
            cRet  := cKey
         endif
      endif
   return cRet

   //------------------------------------------------------------//
   // create a lookup array with the old dBASE color names.
   function x_LoadColors
      this.dbaseColors   =  new assocArray()
      this.dbaseColors["b"]   =   8388608    // navy
      this.dbaseColors["b+"]  =  16711680    // blue
      this.dbaseColors["b*"]  =  16711680    // blue
      this.dbaseColors["+b"]  =  16711680    // blue
      this.dbaseColors["*b"]  =  16711680    // blue
      this.dbaseColors["g"]   =     32768    // dark green
      this.dbaseColors["g+"]  =     65280    // lime
      this.dbaseColors["g*"]  =     65280    // lime
      this.dbaseColors["+g"]  =     65280    // lime
      this.dbaseColors["*g"]  =     65280    // lime
      this.dbaseColors["n"]   =         0    // black
      this.dbaseColors["n+"]  =   8421504    // gray
      this.dbaseColors["n*"]  =   8421504    // gray
      this.dbaseColors["+n"]  =   8421504    // gray
      this.dbaseColors["*n"]  =   8421504    // gray
      this.dbaseColors["r"]   =       128    // maroon
      this.dbaseColors["r+"]  =       255    // red
      this.dbaseColors["r*"]  =       255    // red
      this.dbaseColors["+r"]  =       255    // red
      this.dbaseColors["*r"]  =       255    // red
      this.dbaseColors["w"]   =  12632256    // silver
      this.dbaseColors["w+"]  =  16777215    // white
      this.dbaseColors["w*"]  =  16777215    // white
      this.dbaseColors["+w"]  =  16777215    // white
      this.dbaseColors["*w"]  =  16777215    // white
      this.dbaseColors["bg"]  =   8421376    // teal
      this.dbaseColors["bg+"] =  16776960    // cyan
      this.dbaseColors["bg*"] =  16776960    // cyan
      this.dbaseColors["+bg"] =  16776960    // cyan
      this.dbaseColors["b+g"] =  16776960    // cyan
      this.dbaseColors["*bg"] =  16776960    // cyan
      this.dbaseColors["b*g"] =  16776960    // cyan
      this.dbaseColors["br"]  =   8388736    // purple
      this.dbaseColors["br+"] =  16711935    // magenta
      this.dbaseColors["br*"] =  16711935    // magenta
      this.dbaseColors["+br"] =  16711935    // magenta
      this.dbaseColors["*br"] =  16711935    // magenta
      this.dbaseColors["b+r"] =  16711935    // magenta
      this.dbaseColors["b*r"] =  16711935    // magenta
      this.dbaseColors["gb"]  =   8421376    // teal
      this.dbaseColors["gb+"] =  16776960    // cyan
      this.dbaseColors["gb*"] =  16776960    // cyan
      this.dbaseColors["+gb"] =  16776960    // cyan
      this.dbaseColors["*gb"] =  16776960    // cyan
      this.dbaseColors["g+b"] =  16776960    // cyan
      this.dbaseColors["g*b"] =  16776960    // cyan
      this.dbaseColors["gr"]  =     32896    // olive
      this.dbaseColors["gr+"] =     65535    // yellow
      this.dbaseColors["gr*"] =     65535    // yellow
      this.dbaseColors["+gr"] =     65535    // yellow
      this.dbaseColors["*gr"] =     65535    // yellow
      this.dbaseColors["g+r"] =     65535    // yellow
      this.dbaseColors["g*r"] =     65535    // yellow
      this.dbaseColors["rb"]  =   8388736    // purple
      this.dbaseColors["rb+"] =  16711935    // magenta
      this.dbaseColors["rb*"] =  16711935    // magenta
      this.dbaseColors["+rb"] =  16711935    // magenta
      this.dbaseColors["*rb"] =  16711935    // magenta
      this.dbaseColors["r+b"] =  16711935    // magenta
      this.dbaseColors["r*b"] =  16711935    // magenta
      this.dbaseColors["rg"]  =     32896    // olive
      this.dbaseColors["rg+"] =     65535    // yellow
      this.dbaseColors["rg*"] =     65535    // yellow
      this.dbaseColors["+rg"] =     65535    // yellow
      this.dbaseColors["*rg"] =     65535    // yellow
      this.dbaseColors["r+g"] =     65535    // yellow
      this.dbaseColors["r*g"] =     65535    // yellow
      //---------------------------------------------------------//
      // create a lookup array with the standard dialog colors.
      // the dialog uses the color (4194368) 2 times.
      this.dlgColors =   new array(47, 1)
      this.dlgColors[ 1]   =        64
      this.dlgColors[ 2]   =       128
      this.dlgColors[ 3]   =       255
      this.dlgColors[ 4]   =     16384
      this.dlgColors[ 5]   =     16512
      this.dlgColors[ 6]   =     32768
      this.dlgColors[ 7]   =     32896
      this.dlgColors[ 8]   =     33023
      this.dlgColors[ 9]   =     65280
      this.dlgColors[10]   =     65408
      this.dlgColors[11]   =     65535
      this.dlgColors[12]   =   4194368
      this.dlgColors[13]   =   4194304
      this.dlgColors[14]   =   4194432
      this.dlgColors[15]   =   4210688
      this.dlgColors[16]   =   4210816
      this.dlgColors[17]   =   4227200
      this.dlgColors[18]   =   4227072
      this.dlgColors[19]   =   4227327
      this.dlgColors[20]   =   4259584
      this.dlgColors[31]   =   8388608
      this.dlgColors[21]   =   8388672
      this.dlgColors[22]   =   8388736
      this.dlgColors[32]   =   8388863
      this.dlgColors[23]   =   8404992
      this.dlgColors[24]   =   8421376
      this.dlgColors[25]   =   8421440
      this.dlgColors[26]   =   8421504
      this.dlgColors[27]   =   8421631
      this.dlgColors[28]   =   8453888
      this.dlgColors[29]   =   8454016
      this.dlgColors[30]   =   8454143
      this.dlgColors[33]   =  10485760
      this.dlgColors[34]   =  12615680
      this.dlgColors[35]   =  12615808
      this.dlgColors[36]   =  12615935
      this.dlgColors[37]   =  12632256
      this.dlgColors[38]   =  16711680
      this.dlgColors[39]   =  16711808
      this.dlgColors[40]   =  16711935
      this.dlgColors[41]   =  16744448
      this.dlgColors[42]   =  16744576
      this.dlgColors[43]   =  16744703
      this.dlgColors[44]   =  16776960
      this.dlgColors[45]   =  16777088
      this.dlgColors[46]   =  16777215
      this.dlgColors[47]   =         0
      //---------------------------------------------------------//
      // create a lookup array with the standard java colors.
      this.javaColors   =  new assocArray()
      this.javaColors["aliceblue"]           =  16775408
      this.javaColors["antiquewhite"]        =  14150650
      this.javaColors["aqua"]                =  16776960
      this.javaColors["aquamarine"]          =  13959039
      this.javaColors["azure"]               =  16777200
      this.javaColors["beige"]               =  14480885
      this.javaColors["bisque"]              =  12903679
      this.javaColors["black"]               =         0
      this.javaColors["blanchedalmond"]      =  13495295
      this.javaColors["blue"]                =  16711680
      this.javaColors["blueviolet"]          =  14822282
      this.javaColors["brown"]               =   2763429
      this.javaColors["burlywood"]           =   8894686
      this.javaColors["cadetblue"]           =  10526303
      this.javaColors["chartreuse"]          =     65407
      this.javaColors["chocolate"]           =   1993170
      this.javaColors["coral"]               =   5275647
      this.javaColors["cornflowerblue"]      =  15570276
      this.javaColors["cornsilk"]            =  14481663
      this.javaColors["crimson"]             =   3937500
      this.javaColors["cyan"]                =  16776960
      this.javaColors["darkblue"]            =   9109504
      this.javaColors["darkcyan"]            =   9145088
      this.javaColors["darkgoldenrod"]       =    755384
      this.javaColors["darkgray"]            =  11119017
      this.javaColors["darkgreen"]           =     25600
      this.javaColors["darkkhaki"]           =   7059389
      this.javaColors["darkmagenta"]         =   9109643
      this.javaColors["darkolivegreen"]      =   3107669
      this.javaColors["darkorange"]          =     36095
      this.javaColors["darkorchid"]          =  13382297
      this.javaColors["darkred"]             =       139
      this.javaColors["darksalmon"]          =   8034025
      this.javaColors["darkseagreen"]        =   9419919
      this.javaColors["darkslateblue"]       =   9125192
      this.javaColors["darkslategray"]       =   5197615
      this.javaColors["darkturquoise"]       =  13749760
      this.javaColors["darkviolet"]          =  13828244
      this.javaColors["deeppink"]            =   9639167
      this.javaColors["deepskyblue"]         =  16760576
      this.javaColors["dimgray"]             =   6908265
      this.javaColors["dodgerblue"]          =  16748574
      this.javaColors["firebrick"]           =   2237106
      this.javaColors["floralwhite"]         =  15792895
      this.javaColors["forestgreen"]         =   2263842
      this.javaColors["fuchsia"]             =  16711935
      this.javaColors["gainsboro"]           =  14474460
      this.javaColors["ghostwhite"]          =  16775416
      this.javaColors["gold"]                =     55295
      this.javaColors["goldenrod"]           =   2139610
      this.javaColors["gray"]                =   8421504
      this.javaColors["green"]               =     32768
      this.javaColors["greenyellow"]         =   3145645
      this.javaColors["honeydew"]            =  15794160
      this.javaColors["hotpink"]             =  11823615
      this.javaColors["indianred"]           =   6053069
      this.javaColors["indigo"]              =   8519755
      this.javaColors["ivory"]               =  15794175
      this.javaColors["khaki"]               =   9234160
      this.javaColors["lavender"]            =  16443110
      this.javaColors["lavenderblush"]       =  16118015
      this.javaColors["lawngreen"]           =     64636
      this.javaColors["lemonchiffon"]        =  13499135
      this.javaColors["lightblue"]           =  15128749
      this.javaColors["lightcoral"]          =   8421616
      this.javaColors["lightcyan"]           =  16777184
      this.javaColors["lightgoldenrodyellow"]=  13826810
      this.javaColors["lightgreen"]          =   9498256
      this.javaColors["lightgrey"]           =  13882323
      this.javaColors["lightpink"]           =  12695295
      this.javaColors["lightsalmon"]         =   8036607
      this.javaColors["lightseagreen"]       =  11186720
      this.javaColors["lightskyblue"]        =  16436871
      this.javaColors["lightslategray"]      =  10061943
      this.javaColors["lightsteelblue"]      =  14599344
      this.javaColors["lightyellow"]         =  14745599
      this.javaColors["lime"]                =     65280
      this.javaColors["limegreen"]           =   3329330
      this.javaColors["linen"]               =  15134970
      this.javaColors["magenta"]             =  16711935
      this.javaColors["maroon"]              =       128
      this.javaColors["mediumaquamarine"]    =  11193702
      this.javaColors["mediumblue"]          =  13434880
      this.javaColors["mediumorchid"]        =  13850042
      this.javaColors["mediumpurple"]        =  14381203
      this.javaColors["mediumseagreen"]      =   7451452
      this.javaColors["mediumslateblue"]     =  15624315
      this.javaColors["mediumspringgreen"]   =  10156544
      this.javaColors["mediumturquoise"]     =  13422920
      this.javaColors["mediumvioletred"]     =   8721863
      this.javaColors["midnightblue"]        =   7346457
      this.javaColors["mintcream"]           =  16449525
      this.javaColors["mistyrose"]           =  14804223
      this.javaColors["moccasin"]            =  11920639
      this.javaColors["navajowhite"]         =  11394815
      this.javaColors["navy"]                =   8388608
      this.javaColors["oldlace"]             =  15136253
      this.javaColors["olive"]               =     32896
      this.javaColors["olivedrab"]           =   2330219
      this.javaColors["orange"]              =     42495
      this.javaColors["orangered"]           =     17919
      this.javaColors["orchid"]              =  14053594
      this.javaColors["palegoldenrod"]       =  11200750
      this.javaColors["palegreen"]           =  10025880
      this.javaColors["paleturquoise"]       =  15658671
      this.javaColors["palevioletred"]       =   9662683
      this.javaColors["papayawhip"]          =  14020607
      this.javaColors["peachpuff"]           =  12180223
      this.javaColors["peru"]                =   4163021
      this.javaColors["pink"]                =  13353215
      this.javaColors["plum"]                =  14524637
      this.javaColors["powderblue"]          =  15130800
      this.javaColors["purple"]              =   8388736
      this.javaColors["red"]                 =       255
      this.javaColors["rosybrown"]           =   9408444
      this.javaColors["royalblue"]           =  14772545
      this.javaColors["saddlebrown"]         =   1262987
      this.javaColors["salmon"]              =   7504122
      this.javaColors["sandybrown"]          =   6333684
      this.javaColors["seagreen"]            =   5737262
      this.javaColors["seashell"]            =  15660543
      this.javaColors["sienna"]              =   2970272
      this.javaColors["silver"]              =  12632256
      this.javaColors["skyblue"]             =  15453831
      this.javaColors["slateblue"]           =  13458026
      this.javaColors["slategray"]           =   9470064
      this.javaColors["snow"]                =  16448255
      this.javaColors["springgreen"]         =   8388352
      this.javaColors["steelblue"]           =  11829830
      this.javaColors["tan"]                 =   9221330
      this.javaColors["teal"]                =   8421376
      this.javaColors["thistle"]             =  14204888
      this.javaColors["tomato"]              =   4678655
      this.javaColors["turquoise"]           =  13688896
      this.javaColors["violet"]              =  15631086
      this.javaColors["wheat"]               =  11788021
      this.javaColors["white"]               =  16777215
      this.javaColors["whitesmoke"]          =  16119285
      this.javaColors["yellow"]              =     65535
      this.javaColors["yellowgreen"]         =   3329434
      //---------------------------------------------------------//
      // create a lookup array with the system colors.
      // change for dynamic color change.
      this.sysColors =  new assocArray()
      this.sysColors["activeborder"]   =  COLOR_ACTIVEBORDER
      this.sysColors["activecaption"]  =  COLOR_ACTIVECAPTION
      this.sysColors["appworkspace"]   =  COLOR_APPWORKSPACE
      this.sysColors["background"]     =  COLOR_BACKGROUND
      this.sysColors["btnface"]        =  COLOR_BTNFACE
      this.sysColors["btnhighlight"]   =  COLOR_BTNHIGHLIGHT
      this.sysColors["btnshadow"]      =  COLOR_BTNSHADOW
      this.sysColors["btntext"]        =  COLOR_BTNTEXT
      this.sysColors["captiontext"]    =  COLOR_CAPTIONTEXT
      this.sysColors["graytext"]       =  COLOR_GRAYTEXT
      this.sysColors["highlight"]      =  COLOR_HIGHLIGHT
      this.sysColors["highlighttext"]  =  COLOR_HIGHLIGHTTEXT
      this.sysColors["inactiveborder"] =  COLOR_INACTIVEBORDER
      this.sysColors["inactivecaption"]      =  COLOR_INACTIVECAPTION
      this.sysColors["inactivecaptiontext"]  =  COLOR_INACTIVECAPTIONTEXT
      this.sysColors["infobk"]         =  COLOR_INFOBK
      this.sysColors["infotext"]       =  COLOR_INFOTEXT
      this.sysColors["menu"]           =  COLOR_MENU
      this.sysColors["menutext"]       =  COLOR_MENUTEXT
      this.sysColors["scrollbar"]      =  COLOR_SCROLLBAR
      this.sysColors["window"]         =  COLOR_WINDOW
      this.sysColors["windowframe"]    =  COLOR_WINDOWFRAME
      this.sysColors["windowtext"]     =  COLOR_WINDOWTEXT
      //---------------------------------------------------------//
      // create a lookup array with the system color names.
      this.sysNames  =  new assocArray()
      this.sysNames["activeborder"]    =  "ActiveBorder"
      this.sysNames["activecaption"]   =  "ActiveCaption"
      this.sysNames["appworkspace"]    =  "AppWorkspace"
      this.sysNames["background"]      =  "Background"
      this.sysNames["btnface"]         =  "BtnFace"
      this.sysNames["btnhighlight"]    =  "BtnHighlight"
      this.sysNames["btnshadow"]       =  "BtnShadow"
      this.sysNames["btntext"]         =  "BtnText"
      this.sysNames["captiontext"]     =  "CaptionText"
      this.sysNames["graytext"]        =  "GrayText"
      this.sysNames["highlight"]       =  "Highlight"
      this.sysNames["highlighttext"]   =  "HighlightText"
      this.sysNames["inactiveborder"]  =  "InactiveBorder"
      this.sysNames["inactivecaption"]       =  "InactiveCaption"
      this.sysNames["inactivecaptiontext"]   =  "InactiveCaptionText"
      this.sysNames["infobk"]          =  "InfoBk"
      this.sysNames["infotext"]        =  "InfoText"
      this.sysNames["menu"]            =  "Menu"
      this.sysNames["menutext"]        =  "MenuText"
      this.sysNames["scrollbar"]       =  "Scrollbar"
      this.sysNames["window"]          =  "Window"
      this.sysNames["windowframe"]     =  "WindowFrame"
      this.sysNames["windowtext"]      =  "WindowText"
   return

   //------------------------------------------------------------//
   // assign the dBASE default custom colors.
   function x_LoadCustColors
      Local i, nColr
      // create the dBASE default (all White).
      nColr =  16777215
      for i = 1 to 16
         this.custColors.setMem("color" + str(i, 2, 0, "0"), nColr)
      endfor
   return

   //------------------------------------------------------------//
   function x_Protect
      // protected methods.
      protect  x_CheckDialogColors, x_CloseProcedure,;
               x_CustomColor, x_ErrorMessage, x_GetProcfile,;
               x_HtmlColorSafeItem, x_InitExterns, x_Initialize,;
               x_IsDbaseColor, x_IsHtmlColor, x_IsJavaColor,;
               x_LoadColors, x_LoadCustColors, x_Protect,;
               x_ReadIni, x_Update, x_UpdateColor,;
               x_UpdatedBASEColor, x_UpdateFlags,;
               x_UpdateHtmlColorSafe, x_UpdateJavaColor,;
               x_UpdateOwner, x_WriteIni

      // protected properties.
      protect  chooseColor, custColors, dbaseColors, dlgColors,;
               javaColors, rgbResult, sysColors, sysNames,;
               ~runtime, ~procfile
   return

   //------------------------------------------------------------//
   // read a value from an ini file.
   function x_ReadIni(cSect, cItem)
      Local cBuff, nLen
      cBuff =  space(254)
      nLen  =  RMM_GetIniString( ;
               cSect, cItem, "", cBuff, cBuff.length, this.iniFile)
   return cBuff.left(nLen)

   //------------------------------------------------------------//
   // update the different Color property values
   // after user has selected OK.
   function x_Update
      Local nRgb
      // retrieve the user selection.
      nRgb  =  int(this.chooseColor.getMem('rgbResult'))
      // update protected property.
      this.rgbResult := int(nRgb)
      // update surfaced property (same as protected).
      this.winColor  := int(nRgb)
      // update surfaced hexColor value.
      this.hexColor  := ghex_color(nRgb)
      // update surfaced htmlColor value.
      this.htmlColor := ghtml_color(ghex_color(nRgb))
      // update surfaced htmlColorSafe value.
      this.htmlColorSafe   := ;
         class::x_UpdateHtmlColorSafe((this.htmlColor))
      // update surfaced Color value.
      this.Color  := class::x_UpdatedBASEColor(nRgb)
      // update surfaced javaColor value.
      this.javaColor := class::x_UpdateJavaColor(nRgb)
   return

   //------------------------------------------------------------//
   // update Color property values before display of dialog
   // in the event that direct programmatic changes were made
   // to one of the Color property values.
   // priority sequence when checking for a change is:
   //    winColor, htmlColor, hexColor, Color, javaColor.
   function x_UpdateColor
      Local bChng, nRgb
      bChng =  true
      nRgb  =  int(this.rgbResult)
      if not nRgb == this.winColor
         // replace nRgb with this.winColor.
         nRgb  := int(this.winColor)
      elseif not this.htmlColor == ghtml_color(ghex_color(nRgb))
         // convert this.htmlColor to a winColor.
         nRgb  := int(HtoI(ghtml_color(this.htmlColor)))
      elseif not this.hexColor == ghex_color(nRgb)
         // convert this.hexColor to a winColor.
         nRgb  := int(HtoI(this.hexColor))
      elseif not this.color == class::x_UpdatedBASEColor(nRgb)
         // make sure any direct assignment made is valid.
         Local cGet
         cGet  =  class::x_IsDbaseColor((this.Color))
         if empty(cGet)
            cGet  := COLORDLG_DEFAULTCOLOR
         endif
         this.Color  := cGet
         // convert this.Color to a winColor.
         if this.Color.left(2) == "0x"
            nRgb  := int(HtoI(this.Color.right(6)))
         else
            // must be in 1 of the arrays.
            Local cBase
            cBase =  this.Color.toLowerCase()
            nRgb  := ;
               iif(this.sysColors.iskey(cBase),;
               int(RMM_GetSysColor(this.sysColors[cBase])),;
               iif(this.javaColors.iskey(cBase),;
               this.javaColors[cBase],;
               this.dbaseColors[cBase]))
         endif
      elseif not this.javaColor == class::x_UpdateJavaColor(nRgb)
         Local cGet
         cGet  =  class::x_IsJavaColor((this.javaColor))
         if empty(cGet)
            cGet  := "black"
         endif
         this.javaColor := cGet
         nRgb  := iif(cGet.left(1) == "#",;
            int(HtoI(ghtml_color(cGet.substring(1, cGet.length)))),;
            this.javaColors[cGet])
      else
         // no change has been made.
         bChng := false
      endif
      if bChng
         // a property value was changed.
         // update all property values with a winColor value.
         this.winColor  := int(nRgb)
         this.hexColor  := ghex_color(nRgb)
         this.htmlColor := ghtml_color(ghex_color(nRgb))
         this.rgbResult := int(nRgb)
         this.htmlColorSafe   := ;
            class::x_UpdateHtmlColorSafe((this.htmlColor))
         this.Color  := class::x_UpdatedBASEColor(nRgb)
         this.javaColor := class::x_UpdateJavaColor(nRgb)
      endif
   return int(this.rgbResult)

   //------------------------------------------------------------//
   // check the elements of the lookup arrays for a match.
   // sequence is determined by the colorStyle property.
   function x_UpdatedBASEColor(nRgb)
      Local cKey, cOrd, cRet, i, nCnt, nChck
      // set the default return value.
      cRet  =  "0x" + ghex_color(nRgb)

      nChck =  int(mod(val("" + this.colorStyle), 4))
      if nChck == 0
         cOrd  =  "0,1,2,3"
      elseif nChck == 1
         cOrd  =  "1,0,2,3"
      elseif nChck == 2
         cOrd  =  "2,0,1,3"
      elseif nChck == 3
         cOrd  =  "3,0,1,2"
      endif
      for i = 0 to 3
         nChck := int(val(cOrd.substring(i * 2, i * 2 + 1)))
         if nChck == 3
            // use the default.
            nCnt  =  250
         elseif nChck == 2
            // check in sysColors.
            cKey  =  this.sysColors.firstkey
            for nCnt = 1 to this.sysColors.count()
               // change to a dynamic check. 10/02/2004.
               // if this.sysColors[cKey] == nRgb
               if int(RMM_GetSysColor(this.sysColors[cKey])) == nRgb
                  nCnt  := 250
                  cRet  := this.sysNames[cKey]
               else
                  cKey  := this.sysColors.nextkey(cKey)
               endif
            endfor
         elseif nChck == 1
            // check in dbaseColors.
            cKey  =  this.dbaseColors.firstkey
            for nCnt = 1 to this.dbaseColors.count()
               if this.dbaseColors[cKey] == nRgb
                  nCnt  := 250
                  cRet  := cKey
               else
                  cKey  := this.dbaseColors.nextkey(cKey)
               endif
            endfor
         elseif nChck == 0
            // check in javaColors.
            cKey  =  this.javaColors.firstkey
            for nCnt = 1 to this.javaColors.count()
               if this.javaColors[cKey] == nRgb
                  nCnt  := 250
                  cRet  := cKey
               else
                  cKey  := this.javaColors.nextkey(cKey)
               endif
            endfor
         endif
         if nCnt > 200
            i  := 8
         endif
      endfor
   return cRet

   //------------------------------------------------------------//
   // return a value for the Flags member of the dialog.
   function x_UpdateFlags
      nRet  =  CC_RGBINIT
      if this.preventFullOpen
         nRet  += CC_PREVENTFULLOPEN
      else
         if this.fullOpen
            nRet  += CC_FULLOPEN
         endif
      endif
      if this.showHelp
         nRet  += CC_SHOWHELP
      endif
   return nRet

   //------------------------------------------------------------//
   // return the closest html color safe font color.
   function x_UpdateHtmlColorSafe(cHtml)
   return ;
      class::x_HtmlColorSafeItem(HtoI(cHtml.substring(0, 2))) +;
      class::x_HtmlColorSafeItem(HtoI(cHtml.substring(2, 4))) +;
      class::x_HtmlColorSafeItem(HtoI(cHtml.substring(4, 6)))

   //------------------------------------------------------------//
   function x_UpdateJavaColor(nRgb)
      Local cKey, cRet, i
      cKey  =  this.javaColors.firstkey
      for i = 1 to this.javaColors.count()
         if this.javaColors[cKey] == nRgb
            cRet  =  cKey
            i  := 250
         else
            cKey  := this.javaColors.nextkey(cKey)
         endif
      endfor
      if i < 250
         cRet  =  "#" + ghtml_color(ghex_color(nRgb))
      endif
   return cRet

   //------------------------------------------------------------//
   // return the hwnd for the dialog to display in.
   function x_UpdateOwner
      // not functioning properly Yet, use the activeWindow call.
      // Local oRect
      // oRect =  COLORDLG_Structures::rectStructure()
      // if not RMM_GetWindowRect(int(this.ownerHandle), oRect)
         this.ownerHandle  := int(RMM_GetActiveWIndow())
      // endif
      // release object oRect
      // oRect := null
   return int(this.ownerHandle)

   //------------------------------------------------------------//
   // write value to ini file.
   function x_WriteIni(cSect, cItem, cValu)
   return   RMM_WriteIniString( ;
            cSect, cItem, cValu, this.iniFile)
endclass
//---------------------------------------------------------------//
//          end of colorDlg
//---------------------------------------------------------------//
//---------------------------------------------------------------//
//---------------------------------------------------------------//
class COLORDLG_Structures
//---------------------------------------------------------------//
   //------------------------------------------------------------//
   //       end of constructor   -  methods below
   //------------------------------------------------------------//
   function chooseColorStructure
      Local sStruct
      if not VSTRUCT_ISDEF("CHOOSECOLOR")
         VSTRUCT_DEF('CHOOSECOLOR',;
         'lStructSize',       TYPE_DWORD,;
         'hwndOwner',         TYPE_CHANDLE,;
         'hInstance',         TYPE_CHANDLE,;
         'rgbResult',         TYPE_DWORD,;
         'lpCustColors',      TYPE_POINTER,;
         'Flags',             TYPE_DWORD,;
         'lCustData',         TYPE_POINTER,;
         'lpfnHook',          TYPE_POINTER,;
         'lpTemplateName',    TYPE_POINTER)
      endif
      sStruct  =  new VStruct('CHOOSECOLOR')
      sStruct.setMem('lStructSize', sStruct.getSize())
   return sStruct

   //------------------------------------------------------------//
   function custColorsStructure
      Local sStruct
      if not VSTRUCT_ISDEF("CUSTCOLORS")
         VSTRUCT_DEF('CUSTCOLORS',;
         'color01',           TYPE_DWORD,;
         'color02',           TYPE_DWORD,;
         'color03',           TYPE_DWORD,;
         'color04',           TYPE_DWORD,;
         'color05',           TYPE_DWORD,;
         'color06',           TYPE_DWORD,;
         'color07',           TYPE_DWORD,;
         'color08',           TYPE_DWORD,;
         'color09',           TYPE_DWORD,;
         'color10',           TYPE_DWORD,;
         'color11',           TYPE_DWORD,;
         'color12',           TYPE_DWORD,;
         'color13',           TYPE_DWORD,;
         'color14',           TYPE_DWORD,;
         'color15',           TYPE_DWORD,;
         'color16',           TYPE_DWORD)
      endif
      sStruct  =  new VStruct('CUSTCOLORS')
   return sStruct

   //------------------------------------------------------------//
   function rectStructure
      Local sStruct
      if not VSTRUCT_ISDEF("RECT")
         VSTRUCT_DEF('RECT',;
         'left',              TYPE_LONG,;
         'top',               TYPE_LONG,;
         'right',             TYPE_LONG,;
         'bottom',            TYPE_LONG)
      endif
      sStruct  =  new VStruct('RECT')
   return sStruct
endclass
//---------------------------------------------------------------//
//          end of COLORDLG_Structures
//---------------------------------------------------------------//