/* 
   ----------------------------------------------------------------
    Program...: DBLOAT.H
    Programmer: Ken Mayer -- based on code by a lot of contributors
    Date......: 02/15/1997
                12/11/1997 -- VdBASE 7 -- stripped out the xBase
                              inline functions
                01/26/1998 -- Jose Luis Perez fixed a couple of
                              the date routines to handle set epoch
                              ...
                Feb. 25, 1999 -- Added some definitions for
                              use with datetime calculations ...
                August, 2000 -- added PadL() and PadR()
    Notes.....: This file contains a series of "inline" functions 
                or other routines. These are defined with the 
                #define statement, and are fairly simple. In some 
                cases, you may find more involved, or perhaps 
                "expanded" (code -- for readability) versions of 
                the same functions elsewhere in the dUFLP library. 
                The comments are shortened as well, to keep the 
                size of this file small.
   ----------------------------------------------------------------
*/

/* 
   ----------------------------------------------------------------
   Misc. Definitions:
   ----------------------------------------------------------------
*/
// -- CR = Carriage Return:
#define CR chr(13)

// ----------------------------------------------------------------
// -- LF = Line Feed:
#define LF chr(10)

// ----------------------------------------------------------------
// -- CRLF = Carriage Return/Line Feed:
#define CRLF chr(13)+chr(10)

/*
   ----------------------------------------------------------------
   DateTime definitions
   dt = datetime()
   ? dt            // current value of dt
   ? dt+1          // add a day
   ? dt+DTSECOND   // add a second
   ? dt+DTMINUTE   // add a minute
   ? dt+DTHOUR     // add an hour

   d1 = CtoDT( "01/01/2000 10:00:00" )
   d2 = CtoDT( "01/01/2000 10:01:01" ) // 1 minute, 1 second later
                                       // 61 seconds difference
   ? int( ( d2 - d1 ) * 86400 )        // should return 61 seconds
   ----------------------------------------------------------------
*/
/*
   ----------------------------------------------------------------
   -- DTSecond = used with datetime calculations to give precise
      values. This assumes the datetime function or a timestamp
      field in a dBF7 table:
*/
#define DTSECOND 1/86400

/*
   ----------------------------------------------------------------
   -- DTMinute = used with datetime calculations to give precise
      values. This assumes the datetime function or a timestamp
      field in a dBF7 table:
*/
#define DTMINUTE ( DTSECOND ) * 60

/*
   ----------------------------------------------------------------
   -- DTHour = used with datetime calculations to give precise
      values. This assumes the datetime function or a timestamp
      field in a dBF7 table:
*/
#define DTHOUR ( DTMINUTE ) * 60

/*
   ----------------------------------------------------------------
    date routines
   ----------------------------------------------------------------
*/

/*
    ----------------------------------------------------------------
    Age -- Jay Parsons, Martin Leon and others -- returns numeric
           value, age (years) based on current date
    Usage:   age( dDate )
    Returns: numeric
*/
#ifndef AGE
   #define AGE(dDate) ( floor( ( val( dtos( date() ) ) ;
                      - val( dtos( dDate ) ) ) / 10000 ) )
#endif

/*
    ------------------------------------------------------------------
    Age2 -- Ken Chan, Martin Leon, and others -- returns numeric
            value, age (years) between two dates
    Usage:    age2( dDate1, dDate2 )
    Returns:  numeric
*/
#ifndef AGE2
   #define AGE2(dDate1,dDate2) ( floor( ( val( dtos(dDate2) ) ;
                                - val( dtos(dDate1) ) ) / 10000 ) )
#endif

/*
    ------------------------------------------------------------------
    DateText -- Miriam Liskin created original version of this 
                years ago, "lifted" by Ken M., and turned into 
                inline by Jay Parsons
    Usage:  DateText( dDate )
    Returns: character - date in month day, year format:
                         i.e.,  February 15, 1997
*/
#ifndef DATETEXT
   #define DATETEXT( dDate ) ( cmonth( dDate ) +" "+;
                             ltrim(str(day( dDate ),2))+", "+;
                             str(year( dDate ),4) )
#endif

/*
    ------------------------------------------------------------------
    DateText2 -- Miriam Liskin created original version of this 
                 years ago, "lifted" by Ken M., and turned into 
                 inline by Jay Parsons (This one requires 
                 DateText above) 
    Usage:   DateText2( dDate )
    Returns: character - date in character day, month day, year 
                         format:
                         i.e.,  Saturday, February 15, 1997
*/
#ifndef DATETEXT2
   #define DATETEXT2(dDate) ( cdow(dDate ) +  ", " + DATETEXT(dDate ))
#endif

/*
    ------------------------------------------------------------------
    DateText3 -- Miriam Liskin created original version of this 
                 years ago, "lifted" by Ken M., and turned into 
                 inline by Jay Parsons -- this is Doug Seine's 
                 version -- he removed the day of month bits
    Usage:   DateText3( dDate )
    Returns: character - date in month day, year format:
                         i.e.,  February 1997
*/
#ifndef DATETEXT3
    #define DATETEXT3(dDate) ( cmonth(dDate)+", "+str(year(dDate),4) )
#endif

/*
    ------------------------------------------------------------------
    Date2Jul -- Jay Parsons -- returns # of days since January
                1, 4713 B.C. (Astronomer's "Julian" date)
                Do not use for dBASE dates before 1582, or without 
                considering time zones if using for astronomy.
    Usage:   Date2Jul( dDate )
    Returns: numeric - # of days ...
*/
#ifndef DAT2JUL
   #define DAT2JUL(dDate) ( 1721426 + dDate - ctod( "001/001/001" ) )
   // previous definition assumed set epoch at 1900
#endif

/*
    ------------------------------------------------------------------
    Jul2Date -- Jay Parsons -- Converts Julian # of days to dBASE 
                Date (Opposite of Date2Jul)
    Usage:   Jul2Date( nDays )
    Returns: Date
*/
#ifndef JUL2DAT
   #define JUL2DAT(nDays) ( ctod( "001/001/001" ) + (nDays) - 1721426)
#endif

/*
    ------------------------------------------------------------------
    FDoM -- Ken Chan -- this returns the first day of the month for
                        the month of a specific date.
    Usage:   FDoM( dDate )
    Returns: Date
*/
#ifndef FDOM
   #define FDOM( dDate )  ( ( dDate ) - day( dDate ) + 1 )
#endif

/*
    ------------------------------------------------------------------
    FRSTNxtMth -- Todd Barry -- this returns the first day of the 
                  next month for the month of a specific date.
    Usage:   FRSTNxtMth( dDate )
    Returns: Date
*/
#ifndef FRSTNXTMTH
   #define FRSTNXTMTH(dDate)  ( FDOM( ( dDate ) - day( dDate ) + 45 ) )
#endif

/*
    ------------------------------------------------------------------
    NDoW -- Jay Parsons -- returns numeric day of week for text
            version.
    Usage:   NDoW( cDay )
    Returns: Numeric
*/
#ifndef NDOW
   #define NDOW(cDay)   ( at( upper( left( ltrim( cDay ), 3 ) ), ;
                        "   SUN MON TUE WED THU FRI SAT" ) / 4 )
#endif

/*
    ------------------------------------------------------------------
    Quarter -- Bowen Moursund -- returns quarter of year a date 
               falls in
    Usage:   Quarter( dDate )
    Returns: numeric
*/
#ifndef QUARTER
   #define QUARTER(dDate) ( iif( type( "dDate" )="D",;
                            ceiling( month( dDate ) /3 ),0 ) )
#endif

/*
   ------------------------------------------------------------------
    string routines
   ------------------------------------------------------------------
*/

/*
   ------------------------------------------------------------------
   AllTrim -- lots of sources ... trims on both sides of string.
   Usage:   alltrim( cString )
   Returns: character
*/
#ifndef ALLTRIM
   #define ALLTRIM( cString )   ltrim(rtrim( cString ))
#endif

/*
   ------------------------------------------------------------------
   LastWord -- Martin Leon - returns last word of character string
               (assumes spaces between words)
   Usage:   LastWord( cString )
   Returns: character
*/
#ifndef LASTWORD
   #define LASTWORD(cString) ( substr( (cString), ;
                             rat( " ", trim(cString) ) + 1 ) )
#endif

/*
   ------------------------------------------------------------------
   PadR -- Paul van House -- pads string (on right) to a length as
           defined, with character defined.
   Usage: PadR(<cString>,<nLength>,<cFill>)
   Returns: returns string of nlength padded on the right with 
           fill character. all three parameters must be passed.
           Padr("123',4,"0") returns "1230"
*/
#ifndef PADR
   #define PADR(cSTRING,nLEN,cPADCHAR) ;
           Transform(Ltrim(Rtrim(cSTRING)) + ;
           Replicate(cPADCHAR,nLEN - Len(Ltrim(Rtrim(cSTRING)))) ,;
           Replicate('X',nLEN))
#endif

/*
   ------------------------------------------------------------------
   PadL -- Paul van House -- pads string (on left) to a length
           as defined, with character defined.
   Usage: PadL(<cString>,<nLength>,<cFill>)
   Returns: string of nlength padded on the left with 
            fill character. all three parameters must be passed.
            Padr("123',4,"0") returns "0123"
*/
#ifdef PADL
   #define PADL(cSTRING,nLEN,cPADCHAR) ;
           Transform(Replicate(cPADCHAR,nLEN - Len(Ltrim(Rtrim(cSTRING)))) + ;
           ltrim(rtrim(cSTRING)),replicate('X',nLEN))
#endif


/*
   ------------------------------------------------------------------
   ToString -- Peter Rorlick -- converts any data type to a 
   string. (Turned into an inline function by KM)
*/
#ifndef ToString
   #define ToString( xValue ) new String( ""+xValue )
#endif

/*
   ------------------------------------------------------------------
   Misc routines:
   ------------------------------------------------------------------
*/

/*
   ------------------------------------------------------------------
   Beep -- Created by Bowen Moursund, simply sounds the 'bell' 
           n times.
   Usage:   beep(<n>)
   Returns: null
*/
#ifndef BEEP
   #define beep(n) ?? replicate( chr(7), n)
#endif

/*
   ------------------------------------------------------------------
   RandNum -- Ken Mayer -- Random number between 1 and number 
              supplied. If used often, suggest one time per run 
              of an application call to RAND( -1 ) to seed 
              random number table.
   Usage:   RandNum( nMax )
   Returns: numeric
*/
#ifndef RandNum 
    #define RandNum( nMax ) ( int( rand() * ( nMax ) ) + 1 ) 
#endif 

/*
   ------------------------------------------------------------------
   IsDir() -- Jim Sare/Romain Strieff -- works on network
   drives and local drives ...
   Usage: IsDir( "some path" )
   Returns: Logical
*/
#ifndef IsDir
    #define isDir(cDir) (not empty(funique(cDir+"\isdir???.???")))
#endif

/* 
   ----------------------------------------------------------------
    End of File: dBLOAT.H
   ----------------------------------------------------------------
*/