/*
   ----------------------------------------------------------------------------
   Programmer..: Matthew James BLACK
   Date........: 1/09/06
   Notes.......: This is a custom class for an encryption object using the
                 Rijndael algorithm (which is the basis of the AES
                 Encryption Standard).
   Written for.: dBASEPlus, v2.6.1
   Rev. History: None
   Calls.......: None
   Called by...: N/A
   Usage.......: oRef = new dulux_encrypt()
                 oRef.get_key(<cKey>[,<nKeySize>[,<nBlockSize>]]) 
                 oRef.encrypt(<cPlainText>)
                 oRef.decrypt(<cCypherText>)
   Example.....: oRef.get_key("abcdefghijklmnop",128,128)
                 s1 := oRef.encrypt("Mary had a little lamb.")
                 s2 := oRef.decrypt(s1)
                 len(s1) == 64
                 len(s2) == 23
   Returns.....: oRef.encrypt(cPlainText) returns cCypherText
                 oRef.decrypt(cCypherText) returns cPlainText
   Parameters..: oRef - a Reference to the dulux_encrypt object
                 cCypherText - a Character String containing Hexadecimal 
                        values representing the text to be 
                        decrypted.  cCypherText has a length equal 
                        to double the length of cPlainText padded out
                        to the next block-boundary.  A block-boundary 
                        is a multiple of (nBlockSize/8) ie 16, 20, 
                        24, 28, or 32.  See the Example Section above 
                        for an example.
                 cKey - a Character String containing the encryption/decryption
                        Key
                 cPlainText - a Character String containing the text to be
                        encrypted
                 nBlockSize - the size of the Block in bits.  Must be a
                        multiple of 32 and between 128 and 256.  Defaults to 
                        128. 
                 nKeySize - the size of the Key in bits.  Must be a multiple of
                        32 and between 128 and 256.  Defaults to 128.

   NOTE: The string returned by the ENCRYPT() method is hexadecimal, which
   means that the size of the string is going to be double the size
   of the original string you encrypted, because each character is now
   represented by a hexadecimal "digit" ... if you wish to store
   the value returned in a table, make sure the field is twice the size
   of the actual data ...
   --------------------------------------------------------------------------

   Programmer Notes:
	The Rijndael Algorithm is the algorithm choosen to become the replacement
for the DES Algorithm, the standard encryption/decryption algorithm used
throughout the world.  Most encryption/decryption used today, over the Internet,
by Bank ATMs, by Smart Cards, or whatever, use AES, which is the standard
implementation of Rijndael.  Full details of the Rijndael Algorithm/AES can be
found at http://csrc.nist.gov/cryptotoolkit/aes/rijndael/.

	This implementation of Rijndael has been written to provide the same level
of security and encryption to dBL applications as can be found in other
languages such as Java and C/C++.  It has been written to use (by default) the
minimum settings recommended by the Rijndael authors, but is able to be used
with greater Key-length and larger Block-size if desired (to provide even
greater security).  For most applications, the default settings will be
sufficient (cryptographly) for the next several years, and it is recommended
that, unless the application requires greater cryptographic security, the
default settings be used.

	I encourage others experienced in security and cryptographic techniques to
examine the code for mistakes and to let me know if you find any.  I can be
contacted at:

		matthew@peregrineit.net

	However, changeing code by those not experienced in these areas is highly
discouraged, as, in general, making a small change is cryptographic code can
have devasting effects on the resulting encryption.

	This code is probably not the most efficient implementation of Rijndael
possible in dBL, but it is as efficient as I can make it.  Again, if you feel
you can improve on this (without compromising the cryptography) please let me
know.

	A good encryption/decryption algorithm (or cypher), according to
cryptographic experts, is one that relies only of having the Key and the
Plaintext (the unencrypted text) kept secret.  In fact, it is advisable that
the actual cypher be made public knowledge so that other experts can verify that
the cypher is in fact secure.  Rijndael (and hence this implementation of
Rijndael) follows this practice.  In otherwords, don't be worried about others
knowing how your encryption works (ie by using this Custom Class) as the
underlying cypher has been proved to be secure (see the Rijndael documentation
for what this means).

	The best way to use this custom class is to initalise it once with the same
Key (say, from the _app object) and then leave it in memory.  Although this is
a small security risk, this allows various initialisation routines (run from
the class's Init() method and the get_key() method) to be performed once, thus
improving execution speed.  Of course, if you change Keys then you need to run
get_key() again.

	To fully understand some of the Properties and Methods of the dulux_encrypt
Class, you need to have an understanding of the Rijndael Algorithm.  To attempt
to provide that understanding here would be folly on my part, so I have instead
simply provided the necessary comments to allow an experienced cryptographer to
check my work, and a general dBase developer to use the Class without having to
become an experienced cryptographer.

	(Incedently, I hold a Master's Degree in IT with one of my Majors being
Security.  Although I do not claim to be a cryptographic expert, I have studied
and worked in cryptography, computer forensics, and IT auditing, and so feel
comfortable in producing this Custom Class for use by my fellow dBase
Developers.)

Properties:
	aExpandKey
		A 1-dimensional array whose elements are the hexadecimal representation
	of the Key and its Expansion.

	aInvSBox
   	A 2-dimensional array containing the Inverse Substitution Box 

	aSBox
   	A 2-dimensional array containing the Substitution Box

	aState
		A 2-dimensional array containing the intermediate state of the
	plaintext/cyphertext.

	cKey
		A string, padded with Spaces, which contains the Key and which has a
	length of Nk*4

	Nb
		The number of Columns of the Block.  Equal to Block size divided by 32.

	Nk
		The number of Columns of the Key.  Equal to Key size divided by 32.

	Nr
		The number of Rounds that the cypher is required to go through.
	Equal to 6 plus the maximum of Nb and Nk.

   aRcon
   	A 1-dimensional array containg the Rcon values

Methods:
	addroundkey()
		Used to add (XOR) the Round Key with the intermediate text.
		Calls: None

	bytesub()
		Used to substitue each byte with another via the aSBox array.
		Calls: None

	create_aInvSBox()
   	Used to populate the aInvSBox array.  The values come from the Federal
      Information Processing Standards Publication 197.
		Calls: None

   create_aRcon()
   	Used to populate the aRcon array.  The values come from the Federal
      Information Processing Standards Publication 197.
      Calls: None

	create_aSBox()
   	Used to populate the aSBox array.  The values come from the Federal
      Information Processing Standards Publication 197.
      Calls: None

	cypher_final_round()
		Used to perform each transformation required for the final encryption
      round.
		Calls:	bytesub()
					shiftrow()
					addroundkey()

	cypher_round()
		Used to perform each transformation required for the encryption in each
      round except the last.
		Calls:	bytesub()
					shiftrow()
					mixcol()
					addroundkey()

	decrypt()
		Used to decrypt the Cyphertext into the Plaintext.
		oRef.get_key() MUST be performed at least once before decrypt() is called.
		Calls: 	decypher_round()
					decypher_final_round()
					load_cypher_aState()
					unload_plain_aState()

	decypher_final_round()
		Used to perform each transformation required for the final decryption
      round.
		Calls:	invshiftrow()
					invbytesub()
					addroundkey()

	decypher_round()
		Used to perform each transformation required for the decryption in each
      round except the last.
		Calls:	invshiftrow()
					invbytesub()
					invaddroundkey()
					invmixcol()

	encrypt()
		Used to encrypt the Plaintext into the Cyphertext.
		oRef.get_key() MUST be performed at least once before encrypt() is called.
		Calls: 	cypher_round()
					cypher_final_round()
					load_plain_aState()
					unload_cypher_aState()

	get_key()
		Used to obtain the Key and expand it to get the Round Keys.
		Calls: key_expand()

	init()
   	Used to perform various initalisation routines.
		Calls: 	create_aSBox()
   	   		create_aInvSBox()

	invaddroundkey()
		Used to add (XOR) the Round Key with the intermediate text.
		Calls: None

	invbytesub()
		Used to (inverse)-substitue each byte with another via the aInvSBox array.
		Calls: None

	invmixcol()
		Used to (inverse)-mix the bytes within each column.
		Calls: None

	invshiftrow()
		Used to (inverse)-shift the bytes within each row.
		Calls: None

	key_expand()
		Used to expand the Key into the various Round Keys required for each
		Round of the cypher.
		Calls:	subWord()

	load_cypher_aState()
		Used to load the Cyphertext into the aState array.
		Calls: None

	load_plain_aState()
		Used to load the PlainText into the aState array.
		Calls: None

	mixcol()
		Used to mix the bytes within each column.
		Calls: None

	shiftrow()
		Used to shift the bytes within each row.
		Calls: None

	subWord()
		Used to apply the S-Box (aSBox) to each character in a WordKey in the
		Key_expand() function.
		Calls: None

	unload_cypher_aState()
		Used to retrive the Cyphertext from the aState array.
		Calls: None

	unload_plain_aState()
		Used to retrive the PlainText from the aState array.
		Calls: None
   ----------------------------------------------------------------------------
*/

class dulux_encrypt of object
   // The only unprotected elements are: getkey(), encrypt(), and decrypt()
   protect aExpandKey, aInvSBox, aSBox, aRCon, aState 
   protect cKey, Nb, Nk, Nr, aSBox, aRcon
   protect addroundkey, bytesub, create_aRcon, create_aSBox
   protect cypher_final_round, cypher_round, decypher_final_round, decypher_round
   protect init, invbytesub, invmixcol, invshiftrow, key_expand
   protect load_cypher_astate, load_plain_astate, mixcol, shiftrow, subWord
   protect unload_cypher_aState, unload_plain_aState

	this.Nb = 0
	this.Nk = 0
	this.Nr = 0
	this.aSBox = new array(16,16)
	this.aInvSBox = new array(16,16)
   this.aRcon = new array(29)

	class::init()

	function addroundkey(nRoundNumber)
		local i,j
		for j=1 to this.Nb
			for i=1 to 4
				this.aState[i,j] := itoh(bitxor(htoi(this.aState[i,j]),htoi(substr(this.aExpandKey[(nRoundNumber-1)*this.Nb+j],(i-1)*2+1,2))),2)
	 		next
		next
	return

	function bytesub()
		local i,j
		for j = 1 to this.Nb
			for i = 1 to 4
				this.aState[i,j] := this.aSBox[htoi(substr(this.aState[i,j],1,1))+1,htoi(substr(this.aState[i,j],2,1))+1]
			next
		next
	return

	function create_aInvSBox()
   		this.aInvSBox[1,1] := "52"
	   	this.aInvSBox[1,2] := "09"
   		this.aInvSBox[1,3] := "6a"
	   	this.aInvSBox[1,4] := "d5"
   		this.aInvSBox[1,5] := "30"
	   	this.aInvSBox[1,6] := "36"
   		this.aInvSBox[1,7] := "a5"
	   	this.aInvSBox[1,8] := "38"
   		this.aInvSBox[1,9] := "bf"
	   	this.aInvSBox[1,10] := "40"
   		this.aInvSBox[1,11] := "a3"
	   	this.aInvSBox[1,12] := "9e"
   		this.aInvSBox[1,13] := "81"
	   	this.aInvSBox[1,14] := "f3"
   		this.aInvSBox[1,15] := "d7"
	   	this.aInvSBox[1,16] := "fb"
   		this.aInvSBox[2,1] := "7c"
	   	this.aInvSBox[2,2] := "e3"
   		this.aInvSBox[2,3] := "39"
	   	this.aInvSBox[2,4] := "82"
   		this.aInvSBox[2,5] := "9b"
	   	this.aInvSBox[2,6] := "2f"
   		this.aInvSBox[2,7] := "ff"
	   	this.aInvSBox[2,8] := "87"
   		this.aInvSBox[2,9] := "34"
	   	this.aInvSBox[2,10] := "8e"
   		this.aInvSBox[2,11] := "43"
	   	this.aInvSBox[2,12] := "44"
   		this.aInvSBox[2,13] := "c4"
	   	this.aInvSBox[2,14] := "de"
   		this.aInvSBox[2,15] := "e9"
	   	this.aInvSBox[2,16] := "cb"
   		this.aInvSBox[3,1] := "54"
	   	this.aInvSBox[3,2] := "7b"
   		this.aInvSBox[3,3] := "94"
	   	this.aInvSBox[3,4] := "32"
   		this.aInvSBox[3,5] := "a6"
	   	this.aInvSBox[3,6] := "c2"
   		this.aInvSBox[3,7] := "23"
	   	this.aInvSBox[3,8] := "3d"
   		this.aInvSBox[3,9] := "ee"
	   	this.aInvSBox[3,10] := "4c"
   		this.aInvSBox[3,11] := "95"
	   	this.aInvSBox[3,12] := "0b"
   		this.aInvSBox[3,13] := "42"
	   	this.aInvSBox[3,14] := "fa"
   		this.aInvSBox[3,15] := "c3"
	   	this.aInvSBox[3,16] := "4e"
   		this.aInvSBox[4,1] := "08"
	   	this.aInvSBox[4,2] := "2e"
   		this.aInvSBox[4,3] := "a1"
	   	this.aInvSBox[4,4] := "66"
   		this.aInvSBox[4,5] := "28"
	   	this.aInvSBox[4,6] := "d9"
   		this.aInvSBox[4,7] := "24"
	   	this.aInvSBox[4,8] := "b2"
			this.aInvSBox[4,9] := "76"
   		this.aInvSBox[4,10] := "5b"
	   	this.aInvSBox[4,11] := "a2"
   		this.aInvSBox[4,12] := "49"
	   	this.aInvSBox[4,13] := "6d"
	   	this.aInvSBox[4,14] := "8b"
   		this.aInvSBox[4,15] := "d1"
   		this.aInvSBox[4,16] := "25"
   		this.aInvSBox[5,1] := "72"
	   	this.aInvSBox[5,2] := "f8"
	   	this.aInvSBox[5,3] := "f6"
   		this.aInvSBox[5,4] := "64"
	   	this.aInvSBox[5,5] := "86"
   		this.aInvSBox[5,6] := "68"
	   	this.aInvSBox[5,7] := "98"
   		this.aInvSBox[5,8] := "16"
	   	this.aInvSBox[5,9] := "d4"
   		this.aInvSBox[5,10] := "a4"
	   	this.aInvSBox[5,11] := "5c"
   		this.aInvSBox[5,12] := "cc"
	   	this.aInvSBox[5,13] := "5d"
   		this.aInvSBox[5,14] := "65"
	   	this.aInvSBox[5,15] := "b6"
   		this.aInvSBox[5,16] := "92"
	   	this.aInvSBox[6,1] := "6c"
   		this.aInvSBox[6,2] := "70"
	   	this.aInvSBox[6,3] := "48"
   		this.aInvSBox[6,4] := "50"
	   	this.aInvSBox[6,5] := "fd"
   		this.aInvSBox[6,6] := "ed"
	   	this.aInvSBox[6,7] := "b9"
   		this.aInvSBox[6,8] := "da"
	   	this.aInvSBox[6,9] := "5e"
   		this.aInvSBox[6,10] := "15"
	   	this.aInvSBox[6,11] := "46"
   		this.aInvSBox[6,12] := "57"
	   	this.aInvSBox[6,13] := "a7"
   		this.aInvSBox[6,14] := "8d"
	   	this.aInvSBox[6,15] := "9d"
   		this.aInvSBox[6,16] := "84"
	   	this.aInvSBox[7,1] := "90"
   		this.aInvSBox[7,2] := "d8"
	   	this.aInvSBox[7,3] := "ab"
   		this.aInvSBox[7,4] := "00"
	   	this.aInvSBox[7,5] := "8c"
   		this.aInvSBox[7,6] := "bc"
	   	this.aInvSBox[7,7] := "d3"
   		this.aInvSBox[7,8] := "0a"
	   	this.aInvSBox[7,9] := "f7"
   		this.aInvSBox[7,10] := "e4"
	   	this.aInvSBox[7,11] := "58"
	   	this.aInvSBox[7,12] := "05"
   		this.aInvSBox[7,13] := "b8"
	   	this.aInvSBox[7,14] := "b3"
   		this.aInvSBox[7,15] := "45"
	   	this.aInvSBox[7,16] := "06"
   		this.aInvSBox[8,1] := "d0"
	   	this.aInvSBox[8,2] := "2c"
   		this.aInvSBox[8,3] := "1e"
	   	this.aInvSBox[8,4] := "8f"
	   	this.aInvSBox[8,5] := "ca"
   		this.aInvSBox[8,6] := "3f"
   		this.aInvSBox[8,7] := "0f"
	   	this.aInvSBox[8,8] := "02"
   		this.aInvSBox[8,9] := "c1"
	   	this.aInvSBox[8,10] := "af"
	   	this.aInvSBox[8,11] := "bd"
   		this.aInvSBox[8,12] := "03"
   		this.aInvSBox[8,13] := "01"
	   	this.aInvSBox[8,14] := "13"
   		this.aInvSBox[8,15] := "8a"
	   	this.aInvSBox[8,16] := "6b"
   		this.aInvSBox[9,1] := "3a"
	   	this.aInvSBox[9,2] := "91"
   		this.aInvSBox[9,3] := "11"
	   	this.aInvSBox[9,4] := "41"
   		this.aInvSBox[9,5] := "4f"
	   	this.aInvSBox[9,6] := "67"
   		this.aInvSBox[9,7] := "dc"
	   	this.aInvSBox[9,8] := "ea"
   		this.aInvSBox[9,9] := "97"
	   	this.aInvSBox[9,10] := "f2"
   		this.aInvSBox[9,11] := "cf"
	   	this.aInvSBox[9,12] := "ce"
   		this.aInvSBox[9,13] := "f0"
	   	this.aInvSBox[9,14] := "b4"
   		this.aInvSBox[9,15] := "e6"
	   	this.aInvSBox[9,16] := "73"
   		this.aInvSBox[10,1] := "96"
	   	this.aInvSBox[10,2] := "ac"
   		this.aInvSBox[10,3] := "74"
	   	this.aInvSBox[10,4] := "22"
   		this.aInvSBox[10,5] := "e7"
	   	this.aInvSBox[10,6] := "ad"
   		this.aInvSBox[10,7] := "35"
	   	this.aInvSBox[10,8] := "85"
	   	this.aInvSBox[10,9] := "e2"
   		this.aInvSBox[10,10] := "f9"
   		this.aInvSBox[10,11] := "37"
	   	this.aInvSBox[10,12] := "e8"
   		this.aInvSBox[10,13] := "1c"
	   	this.aInvSBox[10,14] := "75"
   		this.aInvSBox[10,15] := "df"
	   	this.aInvSBox[10,16] := "6e"
	   	this.aInvSBox[11,1] := "47"
   		this.aInvSBox[11,2] := "f1"
   		this.aInvSBox[11,3] := "1a"
	   	this.aInvSBox[11,4] := "71"
   		this.aInvSBox[11,5] := "1d"
	   	this.aInvSBox[11,6] := "29"
   		this.aInvSBox[11,7] := "c5"
	   	this.aInvSBox[11,8] := "89"
	   	this.aInvSBox[11,9] := "6f"
   		this.aInvSBox[11,10] := "b7"
   		this.aInvSBox[11,11] := "62"
	   	this.aInvSBox[11,12] := "0e"
   		this.aInvSBox[11,13] := "aa"
	   	this.aInvSBox[11,14] := "18"
   		this.aInvSBox[11,15] := "be"
	   	this.aInvSBox[11,16] := "1b"
   		this.aInvSBox[12,1] := "fc"
	   	this.aInvSBox[12,2] := "56"
   		this.aInvSBox[12,3] := "3e"
	   	this.aInvSBox[12,4] := "4b"
   		this.aInvSBox[12,5] := "c6"
   		this.aInvSBox[12,6] := "d2"
	   	this.aInvSBox[12,7] := "79"
	   	this.aInvSBox[12,8] := "20"
   		this.aInvSBox[12,9] := "9a"
	   	this.aInvSBox[12,10] := "db"
   		this.aInvSBox[12,11] := "c0"
	   	this.aInvSBox[12,12] := "fe"
   		this.aInvSBox[12,13] := "78"
	   	this.aInvSBox[12,14] := "cd"
   		this.aInvSBox[12,15] := "5a"
	   	this.aInvSBox[12,16] := "f4"
   		this.aInvSBox[13,1] := "1f"
	   	this.aInvSBox[13,2] := "dd"
   		this.aInvSBox[13,3] := "a8"
	   	this.aInvSBox[13,4] := "33"
   		this.aInvSBox[13,5] := "88"
	   	this.aInvSBox[13,6] := "07"
   		this.aInvSBox[13,7] := "c7"
	   	this.aInvSBox[13,8] := "31"
   		this.aInvSBox[13,9] := "b1"
	   	this.aInvSBox[13,10] := "12"
   		this.aInvSBox[13,11] := "10"
	   	this.aInvSBox[13,12] := "59"
   		this.aInvSBox[13,13] := "27"
   		this.aInvSBox[13,14] := "80"
	   	this.aInvSBox[13,15] := "ec"
	   	this.aInvSBox[13,16] := "5f"
   		this.aInvSBox[14,1] := "60"
   		this.aInvSBox[14,2] := "51"
	   	this.aInvSBox[14,3] := "7f"
	   	this.aInvSBox[14,4] := "a9"
   		this.aInvSBox[14,5] := "19"
	   	this.aInvSBox[14,6] := "b5"
   		this.aInvSBox[14,7] := "4a"
	   	this.aInvSBox[14,8] := "0d"
   		this.aInvSBox[14,9] := "2d"
	   	this.aInvSBox[14,10] := "e5"
   		this.aInvSBox[14,11] := "7a"
	   	this.aInvSBox[14,12] := "9f"
   		this.aInvSBox[14,13] := "93"
	   	this.aInvSBox[14,14] := "c9"
   		this.aInvSBox[14,15] := "9c"
	   	this.aInvSBox[14,16] := "ef"
   		this.aInvSBox[15,1] := "a0"
	   	this.aInvSBox[15,2] := "e0"
   		this.aInvSBox[15,3] := "3b"
	   	this.aInvSBox[15,4] := "4d"
   		this.aInvSBox[15,5] := "ae"
	   	this.aInvSBox[15,6] := "2a"
   		this.aInvSBox[15,7] := "f5"
	   	this.aInvSBox[15,8] := "b0"
   		this.aInvSBox[15,9] := "c8"
	   	this.aInvSBox[15,10] := "eb"
   		this.aInvSBox[15,11] := "bb"
	   	this.aInvSBox[15,12] := "3c"
   		this.aInvSBox[15,13] := "83"
	   	this.aInvSBox[15,14] := "53"
   		this.aInvSBox[15,15] := "99"
	   	this.aInvSBox[15,16] := "61"
   		this.aInvSBox[16,1] := "17"
	   	this.aInvSBox[16,2] := "2b"
   		this.aInvSBox[16,3] := "04"
	   	this.aInvSBox[16,4] := "7e"
   		this.aInvSBox[16,5] := "ba"
	   	this.aInvSBox[16,6] := "77"
   		this.aInvSBox[16,7] := "d6"
	   	this.aInvSBox[16,8] := "26"
   		this.aInvSBox[16,9] := "e1"
	   	this.aInvSBox[16,10] := "69"
   		this.aInvSBox[16,11] := "14"
	   	this.aInvSBox[16,12] := "63"
   		this.aInvSBox[16,13] := "55"
	   	this.aInvSBox[16,14] := "21"
   		this.aInvSBox[16,15] := "0c"
	   	this.aInvSBox[16,16] := "7d"
	return

	function create_aRcon()
   		this.aRcon[1] := "01000000"
   		this.aRcon[2] := "02000000"
   		this.aRcon[3] := "04000000"
   		this.aRcon[4] := "08000000"
   		this.aRcon[5] := "10000000"
   		this.aRcon[6] := "20000000"
   		this.aRcon[7] := "40000000"
   		this.aRcon[8] := "80000000"
   		this.aRcon[9] := "1b000000"
   		this.aRcon[10] := "36000000"
   		this.aRcon[11] := "6c000000"
   		this.aRcon[12] := "d8000000"
   		this.aRcon[13] := "43000000"
   		this.aRcon[14] := "86000000"
   		this.aRcon[15] := "0f000000"
   		this.aRcon[16] := "1e000000"
   		this.aRcon[17] := "3c000000"
   		this.aRcon[18] := "78000000"
   		this.aRcon[19] := "f0000000"
   		this.aRcon[20] := "c5000000"
   		this.aRcon[21] := "6f000000"
   		this.aRcon[22] := "de000000"
   		this.aRcon[23] := "a1000000"
   		this.aRcon[24] := "27000000"
   		this.aRcon[25] := "4e000000"
   		this.aRcon[26] := "9c000000"
   		this.aRcon[27] := "1d000000"
   		this.aRcon[28] := "3a000000"
   		this.aRcon[29] := "74000000"
	return

	function create_aSBox()
   		this.aSBox[1,1] := "63"
			this.aSBox[1,2] := "7c"
	   	this.aSBox[1,3] := "77"
   		this.aSBox[1,4] := "7b"
	   	this.aSBox[1,5] := "f2"
   		this.aSBox[1,6] := "6b"
	   	this.aSBox[1,7] := "6f"
	   	this.aSBox[1,8] := "c5"
	   	this.aSBox[1,9] := "30"
	   	this.aSBox[1,10] := "01"
	   	this.aSBox[1,11] := "67"
	   	this.aSBox[1,12] := "2b"
	   	this.aSBox[1,13] := "fe"
	   	this.aSBox[1,14] := "d7"
	   	this.aSBox[1,15] := "ab"
	   	this.aSBox[1,16] := "76"
	   	this.aSBox[2,1] := "ca"
	   	this.aSBox[2,2] := "82"
	   	this.aSBox[2,3] := "c9"
	   	this.aSBox[2,4] := "7d"
	   	this.aSBox[2,5] := "fa"
	   	this.aSBox[2,6] := "59"
	   	this.aSBox[2,7] := "47"
	   	this.aSBox[2,8] := "f0"
	   	this.aSBox[2,9] := "ad"
	   	this.aSBox[2,10] := "d4"
	   	this.aSBox[2,11] := "a2"
	   	this.aSBox[2,12] := "af"
	   	this.aSBox[2,13] := "9c"
	   	this.aSBox[2,14] := "a4"
			this.aSBox[2,15] := "72"
			this.aSBox[2,16] := "c0"
	   	this.aSBox[3,1] := "b7"
   		this.aSBox[3,2] := "fd"
	   	this.aSBox[3,3] := "93"
   		this.aSBox[3,4] := "26"
	   	this.aSBox[3,5] := "36"
   		this.aSBox[3,6] := "3f"
			this.aSBox[3,7] := "f7"
	   	this.aSBox[3,8] := "cc"
			this.aSBox[3,9] := "34"
	   	this.aSBox[3,10] := "a5"
   		this.aSBox[3,11] := "e5"
	   	this.aSBox[3,12] := "f1"
   		this.aSBox[3,13] := "71"
	   	this.aSBox[3,14] := "d8"
   		this.aSBox[3,15] := "31"
	   	this.aSBox[3,16] := "15"
   		this.aSBox[4,1] := "04"
	   	this.aSBox[4,2] := "c7"
   		this.aSBox[4,3] := "23"
	   	this.aSBox[4,4] := "c3"
   		this.aSBox[4,5] := "18"
	   	this.aSBox[4,6] := "96"
   		this.aSBox[4,7] := "05"
	   	this.aSBox[4,8] := "9a"
   		this.aSBox[4,9] := "07"
	   	this.aSBox[4,10] := "12"
   		this.aSBox[4,11] := "80"
	   	this.aSBox[4,12] := "e2"
   		this.aSBox[4,13] := "eb"
	   	this.aSBox[4,14] := "27"
   		this.aSBox[4,15] := "b2"
	   	this.aSBox[4,16] := "75"
   		this.aSBox[5,1] := "09"
	   	this.aSBox[5,2] := "83"
   		this.aSBox[5,3] := "2c"
	   	this.aSBox[5,4] := "1a"
   		this.aSBox[5,5] := "1b"
	   	this.aSBox[5,6] := "6e"
   		this.aSBox[5,7] := "5a"
	   	this.aSBox[5,8] := "a0"
   		this.aSBox[5,9] := "52"
	   	this.aSBox[5,10] := "3b"
   		this.aSBox[5,11] := "d6"
	   	this.aSBox[5,12] := "b3"
   		this.aSBox[5,13] := "29"
	   	this.aSBox[5,14] := "e3"
   		this.aSBox[5,15] := "2f"
	   	this.aSBox[5,16] := "84"
   		this.aSBox[6,1] := "53"
	   	this.aSBox[6,2] := "d1"
			this.aSBox[6,3] := "00"
	   	this.aSBox[6,4] := "ed"
   		this.aSBox[6,5] := "20"
	   	this.aSBox[6,6] := "fc"
   		this.aSBox[6,7] := "b1"
	   	this.aSBox[6,8] := "5b"
   		this.aSBox[6,9] := "6a"
	   	this.aSBox[6,10] := "cb"
   		this.aSBox[6,11] := "be"
	   	this.aSBox[6,12] := "39"
   		this.aSBox[6,13] := "4a"
	   	this.aSBox[6,14] := "4c"
   		this.aSBox[6,15] := "58"
	   	this.aSBox[6,16] := "cf"
   		this.aSBox[7,1] := "d0"
	   	this.aSBox[7,2] := "ef"
   		this.aSBox[7,3] := "aa"
	   	this.aSBox[7,4] := "fb"
   		this.aSBox[7,5] := "43"
	   	this.aSBox[7,6] := "4d"
   		this.aSBox[7,7] := "33"
	   	this.aSBox[7,8] := "85"
   		this.aSBox[7,9] := "45"
	   	this.aSBox[7,10] := "f9"
   		this.aSBox[7,11] := "02"
	   	this.aSBox[7,12] := "7f"
   		this.aSBox[7,13] := "50"
	   	this.aSBox[7,14] := "3c"
   		this.aSBox[7,15] := "9f"
	   	this.aSBox[7,16] := "a8"
   		this.aSBox[8,1] := "51"
	   	this.aSBox[8,2] := "a3"
   		this.aSBox[8,3] := "40"
	   	this.aSBox[8,4] := "8f"
   		this.aSBox[8,5] := "92"
	   	this.aSBox[8,6] := "9d"
   		this.aSBox[8,7] := "38"
	   	this.aSBox[8,8] := "f5"
   		this.aSBox[8,9] := "bc"
	   	this.aSBox[8,10] := "b6"
			this.aSBox[8,11] := "da"
	   	this.aSBox[8,12] := "21"
   		this.aSBox[8,13] := "10"
	   	this.aSBox[8,14] := "ff"
   		this.aSBox[8,15] := "f3"
	   	this.aSBox[8,16] := "d2"
   		this.aSBox[9,1] := "cd"
	   	this.aSBox[9,2] := "0c"
   		this.aSBox[9,3] := "13"
	   	this.aSBox[9,4] := "ec"
   		this.aSBox[9,5] := "5f"
	   	this.aSBox[9,6] := "97"
   		this.aSBox[9,7] := "44"
	   	this.aSBox[9,8] := "17"
   		this.aSBox[9,9] := "c4"
	   	this.aSBox[9,10] := "a7"
   		this.aSBox[9,11] := "7e"
	   	this.aSBox[9,12] := "3d"
   		this.aSBox[9,13] := "64"
	   	this.aSBox[9,14] := "5d"
   		this.aSBox[9,15] := "19"
	   	this.aSBox[9,16] := "73"
   		this.aSBox[10,1] := "60"
	   	this.aSBox[10,2] := "81"
   		this.aSBox[10,3] := "4f"
	   	this.aSBox[10,4] := "dc"
   		this.aSBox[10,5] := "22"
	   	this.aSBox[10,6] := "2a"
   		this.aSBox[10,7] := "90"
	   	this.aSBox[10,8] := "88"
	   	this.aSBox[10,9] := "46"
   		this.aSBox[10,10] := "ee"
	   	this.aSBox[10,11] := "b8"
   		this.aSBox[10,12] := "14"
	   	this.aSBox[10,13] := "de"
   		this.aSBox[10,14] := "5e"
	   	this.aSBox[10,15] := "0b"
			this.aSBox[10,16] := "db"
   		this.aSBox[11,1] := "e0"
	   	this.aSBox[11,2] := "32"
   		this.aSBox[11,3] := "3a"
	   	this.aSBox[11,4] := "0a"
   		this.aSBox[11,5] := "49"
	   	this.aSBox[11,6] := "06"
   		this.aSBox[11,7] := "24"
	   	this.aSBox[11,8] := "5c"
   		this.aSBox[11,9] := "c2"
	   	this.aSBox[11,10] := "d3"
   		this.aSBox[11,11] := "ac"
	   	this.aSBox[11,12] := "62"
   		this.aSBox[11,13] := "91"
	   	this.aSBox[11,14] := "95"
   		this.aSBox[11,15] := "e4"
	   	this.aSBox[11,16] := "79"
   		this.aSBox[12,1] := "e7"
	   	this.aSBox[12,2] := "c8"
   		this.aSBox[12,3] := "37"
	   	this.aSBox[12,4] := "6d"
   		this.aSBox[12,5] := "8d"
	   	this.aSBox[12,6] := "d5"
   		this.aSBox[12,7] := "4e"
	   	this.aSBox[12,8] := "a9"
   		this.aSBox[12,9] := "6c"
	   	this.aSBox[12,10] := "56"
   		this.aSBox[12,11] := "f4"
	   	this.aSBox[12,12] := "ea"
   		this.aSBox[12,13] := "65"
	   	this.aSBox[12,14] := "7a"
   		this.aSBox[12,15] := "ae"
	   	this.aSBox[12,16] := "08"
   		this.aSBox[13,1] := "ba"
	   	this.aSBox[13,2] := "78"
   		this.aSBox[13,3] := "25"
	   	this.aSBox[13,4] := "2e"
   		this.aSBox[13,5] := "1c"
	   	this.aSBox[13,6] := "a6"
   		this.aSBox[13,7] := "b4"
	   	this.aSBox[13,8] := "c6"
   		this.aSBox[13,9] := "e8"
	   	this.aSBox[13,10] := "dd"
   		this.aSBox[13,11] := "74"
	   	this.aSBox[13,12] := "1f"
   		this.aSBox[13,13] := "4b"
	   	this.aSBox[13,14] := "bd"
   		this.aSBox[13,15] := "8b"
	   	this.aSBox[13,16] := "8a"
   		this.aSBox[14,1] := "70"
	   	this.aSBox[14,2] := "3e"
   		this.aSBox[14,3] := "b5"
	   	this.aSBox[14,4] := "66"
   		this.aSBox[14,5] := "48"
	   	this.aSBox[14,6] := "03"
   		this.aSBox[14,7] := "f6"
	   	this.aSBox[14,8] := "0e"
   		this.aSBox[14,9] := "61"
	   	this.aSBox[14,10] := "35"
   		this.aSBox[14,11] := "57"
	   	this.aSBox[14,12] := "b9"
   		this.aSBox[14,13] := "86"
	   	this.aSBox[14,14] := "c1"
   		this.aSBox[14,15] := "1d"
	   	this.aSBox[14,16] := "9e"
   		this.aSBox[15,1] := "e1"
	   	this.aSBox[15,2] := "f8"
			this.aSBox[15,3] := "98"
	   	this.aSBox[15,4] := "11"
   		this.aSBox[15,5] := "69"
	   	this.aSBox[15,6] := "d9"
   		this.aSBox[15,7] := "8e"
	   	this.aSBox[15,8] := "94"
   		this.aSBox[15,9] := "9b"
	   	this.aSBox[15,10] := "1e"
   		this.aSBox[15,11] := "87"
	   	this.aSBox[15,12] := "e9"
   		this.aSBox[15,13] := "ce"
	   	this.aSBox[15,14] := "55"
   		this.aSBox[15,15] := "28"
	   	this.aSBox[15,16] := "df"
   		this.aSBox[16,1] := "8c"
	   	this.aSBox[16,2] := "a1"
   		this.aSBox[16,3] := "89"
	   	this.aSBox[16,4] := "0d"
   		this.aSBox[16,5] := "bf"
	   	this.aSBox[16,6] := "e6"
   		this.aSBox[16,7] := "42"
	   	this.aSBox[16,8] := "68"
   		this.aSBox[16,9] := "41"
	   	this.aSBox[16,10] := "99"
   		this.aSBox[16,11] := "2d"
	   	this.aSBox[16,12] := "0f"
   		this.aSBox[16,13] := "b0"
	   	this.aSBox[16,14] := "54"
   		this.aSBox[16,15] := "bb"
	   	this.aSBox[16,16] := "16"
	return

	function cypher_final_round()
		class::bytesub()
		class::shiftrow()
		class::addroundkey(this.Nr+1)
	return

	function cypher_round(nRoundNumber)
		class::bytesub()
		class::shiftrow()
		class::mixcol()
		class::addroundkey(nRoundNumber)
	return

	function decypher_final_round()
		class::invshiftrow()
		class::invbytesub()
		class::addroundkey(1)
	return

	function decypher_round(nRoundNumber)
		class::invshiftrow()
		class::invbytesub()
		class::addroundkey(nRoundNumber)
		class::invmixcol()
	return

	function decrypt(cCypherText)
	//	oRef.get_key() MUST be called at least once before this function is invoked.
		local i,j,cTmpText,cPlainText
      cPlainText = ""
		if argcount() == 0
			msgbox("No Cyphertext Supplied.","Error",32)
			return
		elseif this.Nb == 0
			msgbox("oRef.get_key() has not been called during this session.","Error",32)
			return
		endif
		for i=len(cCypherText)/(this.Nb*8) to 1 step -1
			cTmpText = substr(cCypherText,(i-1)*this.Nb*8+1,this.Nb*8)
			class::load_cypher_aState(cTmpText)
			class::addroundkey(this.Nr+1)
         for j= this.Nr to 2 step -1
				class::decypher_round(j)
			next
			class::decypher_final_round()
			cPlainText := class::unload_plain_aState()+cPlainText
		next
	return(trim(ltrim(cPlainText)))

	function encrypt(cPlainText)
	// oRef.get_key() MUST be called at least once before this function is invoked.
		local i,j,cTmpText1,cTmpText2,cCypherText
      cCyphertext = ""
		if argcount() == 0
			msgbox("No Plaintext Supplied.","Error",32)
			return
		elseif this.Nb == 0
			msgbox("oRef.get_key() has not been called during this session.","Error",32)
			return
		endif
		cTmpText2 = cPlainText+space(1+this.Nb*4-(len(cPlainText)%this.Nb*4))
		for i=1 to len(cTmpText2)/(this.Nb*4)
			cTmpText1 = substr(cTmpText2,(i-1)*this.Nb*4+1,this.Nb*4)
			class::load_plain_aState(cTmpText1)
			class::addroundkey(1)
   		for j=1 to this.Nr-1
				class::cypher_round(j+1)
			next
			class::cypher_final_round()
			cCypherText := cCypherText+class::unload_cypher_aState()
		next
	return(cCypherText)

	function get_key(cKey, nKeySize, nBlockSize)
	/*
	nKeySize is the size of the Key in bits.  It must be 128, 160,
		192, 224, or 256, which corresponds to 16, 20, 24, 28,
		or 32 characters, respectively.  Larger Key sizes are
		supported by the class (in steps of 32 bits/4 characters),
		but may not be crytographically secure (Rijndael is not
		specifed for larger Key sizes).  If nKeySize or nBlockSize
		is not defined, a value of 128 will be used.
	nBlockSize is the size of the Block in bits.  It must be 128, 160,
		192, 224, or 256, which corresponds to 16, 20, 24, 28,
		or 32 characters, respectively.  Larger Block sizes are
		supported by the class (in steps of 32 bits/4 characters),
		but may not be crytographically secure (Rijndael is not
		specifed for larger Block sizes).  If nBlockSize is not
		defined, a value of 128 will be used.
	*/
		local i
      i = argcount()
      if i == 0
			msgbox("No Key Value Supplied.","Error",32)
			return
		elseif i == 1
			this.Nb := 4
			this.Nk := 4
		elseif i == 2
			this.Nb := 4
			this.Nk := max((argvector(2)-(argvector(2) % 32))/32,4)
		else
			this.Nb := max((argvector(3)-(argvector(3) % 32))/32,4)
			this.Nk := max((argvector(2)-(argvector(2) % 32))/32,4)
		endif
		this.Nr := max(this.Nb,this.Nk)+6
		this.aExpandKey = new array(this.Nb*(this.Nr+1))
		this.aState = new array(4,this.Nb)
		class::key_expand(ckey)
	return

	function init()
		class::create_aSBox()
		class::create_aInvSBox()
      class::create_aRcon()
	return

	function invbytesub()
		local i,j
		for j = 1 to this.Nb
			for i = 1 to 4
				this.aState[i,j] := this.aInvSBox[htoi(substr(this.aState[i,j],1,1))+1,htoi(substr(this.aState[i,j],2,1))+1]
			next
		next
	return

	function invmixcol()
		local i,j,h
      atmpState = new array(4)
      atmp1 = new array(4)
      atmp2 = new array(4)
      atmp3 = new array(4)
      atmp4 = new array(4)
		for j=1 to this.Nb
			for i=1 to 4
				atmp1[i] := htoi(this.aState[i,j])
            atmp2[i] := bitlshift(atmp1[i],1)
            atmp3[i] := bitlshift(atmp1[i],2)
            atmp4[i] := bitlshift(atmp1[i],3)
			next
		 	atmpState[1] := bitxor(bitxor(bitxor(bitxor(atmp4[1],atmp3[1]),bitxor(atmp2[1],atmp4[2])),bitxor(bitxor(atmp2[2],atmp1[2]),bitxor(atmp4[3],atmp3[3]))),bitxor(bitxor(atmp1[3],atmp4[4]),atmp1[4]))
			atmpState[2] := bitxor(bitxor(bitxor(bitxor(atmp4[1],atmp1[1]),bitxor(atmp4[2],atmp3[2])),bitxor(bitxor(atmp2[2],atmp4[3]),bitxor(atmp2[3],atmp1[3]))),bitxor(bitxor(atmp4[4],atmp3[4]),atmp1[4]))
			atmpState[3] := bitxor(bitxor(bitxor(bitxor(atmp4[1],atmp3[1]),bitxor(atmp1[1],atmp4[2])),bitxor(bitxor(atmp1[2],atmp4[3]),bitxor(atmp3[3],atmp2[3]))),bitxor(bitxor(atmp4[4],atmp2[4]),atmp1[4]))
			atmpState[4] := bitxor(bitxor(bitxor(bitxor(atmp4[1],atmp2[1]),bitxor(atmp1[1],atmp4[2])),bitxor(bitxor(atmp3[2],atmp1[2]),bitxor(atmp4[3],atmp1[3]))),bitxor(bitxor(atmp4[4],atmp3[4]),atmp2[4]))
			for i=1 to 4
				h = bitand(atmpState[i],0x800)
            if (h == 0x800)
					atmpState[i] := bitxor(atmpState[i],0x8d8)
            else
					bitxor(atmpState[i],0x8d8) //A dummy lne to help prevent timing attacks - it actually does nothing except waste some CPU cycles.
            endif
				h = bitand(atmpState[i],0x400)
            if (h == 0x400)
					atmpState[i] := bitxor(atmpState[i],0x46c)
            else
					bitxor(atmpState[i],0x46c) //A dummy lne to help prevent timing attacks - it actually does nothing except waste some CPU cycles.
            endif
				h = bitand(atmpState[i],0x200)
            if (h == 0x200)
					atmpState[i] := bitxor(atmpState[i],0x236)
            else
					bitxor(atmpState[i],0x236) //A dummy lne to help prevent timing attacks - it actually does nothing except waste some CPU cycles.
            endif
				h = bitand(atmpState[i],0x100)
            if (h == 0x100)
					atmpState[i] := bitxor(atmpState[i],0x11b)
            else
					bitxor(atmpState[i],0x11b) //A dummy lne to help prevent timing attacks - it actually does nothing except waste some CPU cycles.
            endif
	 			this.aState[i,j] := itoh(atmpState[i],2)
			next
		next
		release atmpState
		release atmp1
		release atmp2
		release atmp3
		release atmp4
	return

	function invshiftrow()
		local i,j
      atmp = new array(3)
		for i=2 to 4
			for j= 1 to i-1
				atmp[j] := this.aState[i,this.Nb-j+1]
			next
			for j= this.Nb-i+1 to 1 step -1
				this.aState[i,j+i-1] := this.aState[i,j]
			next
			for j=1 to i-1
				this.aState[i,j] := atmp[i-j]
			next
		next
		release atmp
	return

	function key_expand(ckey)
		local i,tmp
		for i=0 to this.Nk-1
			this.aExpandKey[i+1] := itoh(asc(substr(cKey,i*4+1,1)),2)+itoh(asc(substr(cKey,i*4+2,1)),2)+itoh(asc(substr(cKey,i*4+3,1)),2)+itoh(asc(substr(cKey,i*4+4,1)),2)
		next
		for i=this.Nk to this.Nb*(this.Nr+1)-1
			tmp = this.aExpandKey[i]
			if i%this.Nk == 0
				tmp := itoh(bitxor(htoi(class::subWord(substr(tmp,3,6)+substr(tmp,1,2))),htoi(this.aRcon[i/this.Nk])))
			elseif (this.Nk > 6 and i%this.Nk == 4)
				tmp := class::subWord(tmp)
			endif
			this.aExpandKey[i+1] := itoh(bitxor(htoi(this.aExpandkey[i-this.Nk+1]),htoi(tmp)),8)
		next
	return

	function load_cypher_aState(cText)
		local i,j
		for j=1 to this.Nb
			for i=1 to 4
				this.aState[i,j] := substr(cText,1+(i-1)*2+(j-1)*8,2)
			next
		next
	return

	function load_plain_aState(cText)
		local i,j
		for j=1 to this.Nb
			for i=1 to 4
				this.aState[i,j] := itoh(asc(substr(cText,i+(j-1)*4,1)),2)
			next
		next
	return

	function mixcol()
		local i,j,h
      atmp1 = new array(4)
      atmp2 = new array(4)
		for j=1 to this.Nb
			for i=1 to 4
				atmp1[i] := htoi(this.aState[i,j])
				h = bitand(atmp1[i],0x80)
            atmp2[i] := bitlshift(atmp1[i],1)
            if (h == 0x80)
					atmp2[i] := bitxor(atmp2[i],0x11b)
            else
					bitxor(atmp2[i],0x1b) //A dummy lne to help prevent timing attacks - it actually does nothing except waste some CPU cycles.
            endif
			next
			this.aState[1,j] := itoh(bitxor(bitxor(bitxor(atmp2[1],atmp1[4]),bitxor(atmp2[2],atmp1[2])),atmp1[3]),2)
			this.aState[2,j] := itoh(bitxor(bitxor(bitxor(atmp2[2],atmp1[1]),bitxor(atmp2[3],atmp1[3])),atmp1[4]),2)
			this.aState[3,j] := itoh(bitxor(bitxor(bitxor(atmp2[3],atmp1[2]),bitxor(atmp2[4],atmp1[4])),atmp1[1]),2)
			this.aState[4,j] := itoh(bitxor(bitxor(bitxor(atmp2[4],atmp1[3]),bitxor(atmp2[1],atmp1[1])),atmp1[2]),2)
		next
		release atmp1
		release atmp2
	return

	function shiftrow()
		local i,j
      atmp = new array(3)
		for i=2 to 4
			for j=1 to i-1
				atmp[j] := this.aState[i,j]
			next
			for j=1 to this.Nb-i+1
				this.aState[i,j] := this.aState[i,j+i-1]
			next
			for j=1 to i-1
				this.aState[i,this.Nb-i+j+1] := atmp[j]
			next
		next
		release atmp
	return

	function subWord(cWord)
		local i,wrdtmp
		wrdtmp = ""
		for i=0 to 3
			wrdtmp := wrdtmp + this.aSBox[htoi(substr(cWord,i*2+1,1))+1,htoi(substr(cword,i*2+2,1))+1]
		next
	return wrdtmp

	function unload_cypher_aState()
		local i,j,cText
      cText = ""
		for j=1 to this.Nb
			for i=1 to 4
				cText := cText+this.aState[i,j]
			next
		next
	return(cText)

	function unload_plain_aState()
		local i,j,cText
      cText = ""
		for j=1 to this.Nb
			for i=1 to 4
				cText := cText+chr(htoi(this.aState[i,j]))
			next
		next
	return(cText)
endclass
