/*

Description:

The function setLdriverTable() is used for setting the language driver in the
table header of a level 4, 5 or 7 dbf table to a new value. If the table has one
or more index files only the production index file will have it's language
driver ID changed correpondingly. 

When the new language driver has been set the table is reindexed. 

The function includes a parameter for selecting whether or not table values
will be transliterated in accordance with the new language driver. For example
when changing the table language driver from DB437US0 to DBWINWE0 and the
parameter bTranslit is set to false, i.e. no transliteration is done, the 
string 'Andr Mller' will be displayed as 'Andr Mller' in the changed table.

Syntax:

setLdriverTable(<tablename expC>, <ldrivername expC>, <transliteration expL>)

tablename expC			: The table name including extension, "Fish.dbf"
ldrivername expC  	: The short name of the table language driver, "BLLT1DA0"
transliteration expL	: true, 	transliteration will take place
							  false, transliteration will not take place 

Examples of use:

	setLdriverTable("selfield.dbf", "DBWINWE0", false)
	setLdriverTable("Level5.dbf", "DBWINWE0", true)

Programmer:

	Ivar B. Jessen 

History:

22-10-2006 -- Original version

*/

function setLdriverTable
parameters cTable, cLdriver, bTranslit


// Create copy of original table if transliteration of data is required
	close tables

	if bTranslit = true
  		if file("trans.dbf")
			drop table trans
		endif
		copy table ( cTable ) to "trans.dbf"
   	use
	endif

// Get Table version
	vers = getVersion()

// Open table header 
	ft = new file()
	ft.open(cTable, "RW")

// Get the load driver ID corresponding to parameter cLdriver
	ldID = getLdriverID(cLdriver)

	do case
   	case vers = 3					// dBASE III table not valid here
   		return msgbox("dBASE III Table", "Can't do it")
		case vers = 4 or vers = 5	// Level 4 or 5 table. Write ID in table header 
			writeLdriverID(ldID)		// and in mdx if present
	   case vers =	7					// Level7 table. Write new ldriverName in table
   	   writeLdriverName(ldID)	// header and in mdx if present
	   otherwise
      	return null
   endcase

	old_safety = set('safety')

	if bTranslit
		// Zap table and append from saved table trans.dbf
		set safety off
		use ( cTable ) excl
   	zap
		set safety &old_safety
		append from "trans.dbf" reindex
		drop table trans
	else
   	// Reindex table
		use ( cTable ) excl
	   reindex
	   use
	endif
	close all
	set safety &old_safety

return


function getversion
*******************

	t = new tabledef()
	t.tablename = ( cTable )
	t.load()
   vers = t.version
   release object t
   release t

return vers


function writeLdriverName
*************************

// Blank out existing language driver name in table header
	ft.seek(0x020, 0)
	cNull = replicate(chr(0), 32)	
	ft.write(cNull)

// Write new language driver name in table header
	ft.seek(0x020, 0)
	ft.write(cLdriver)
	ft.close()

   cMdx = substr(cTable, 1, rat(".", cTable)) + "mdx"
   if not file(cMdx)
   	return
   endif

// Set LdriverID in mdx file
   setLdriverIDinMdx(cMdx)
   
return


function writeLdriverID
***********************

// Write new language driver ID in table header
	ft.seek(0x1D, 0)
	ft.write(chr(htoi(ldID)))
	ft.close()

   cMdx = substr(cTable, 1, rat(".", cTable)) + "mdx"
   if not file(cMdx)
   	return
   endif

// Set LdriverID in mdx file
   setLdriverIDinMdx(cMdx)

return


function setLdriverIDinMdx(cMdx)
********************************

	f = new file()
	f.open(cMdx, "RW")

// Get number of tags
	f.seek(0x1C, 0)
	nTags = asc(f.read(1))

// Get Block size adder
	f.seek(0x16, 0)
	b1l = asc(f.read(1))
	b1h = asc(f.read(1))
	blockSizeAdder = b1h*256 + b1l
	blockOffset = blockSizeAdder * 2

	f.seek(0x001F, 0)
	f.write(chr(htoi(ldID)))
	f.seek(0x0C0B, 0)
	f.write(chr(htoi(ldID)))

	for i = 1 to nTags - 1
   	f.seek(blockOffset - 1, 1)
   	f.write(chr(htoi(ldID)))
	next

	f.close()

return  


function getLdriverID(cLd)
*********************************

	cHex = ""
	for i = 1 to 8
		cHex += itoh(asc(substr(cLd, i, 1)))
	next

// create dummy version 7 table
	cHex = ""
	for i = 1 to 8
		cHex += itoh(asc(substr(cLDriver, i, 1)))
	next

   cStr = ""
	cStr += "046A0A14000000007500050000000000"
	cStr += "00000000000000000000000000000000"
	cStr +=  cHex + "0000000000000000"
	cStr += "00000000000000000000000000000000"
   cStr += "00000000494400000000000000000000"
	cStr += "00000000000000000000000000000000"
	cStr += "00000000490400000000000000000000"
	cStr += "000000000D1A"

   f = new file() 
   f.create("dummy.dbf")
   f.open("Dummy.dbf","RW")

   for n=1 to len(cStr) step 2
		f.write(chr(htoi(substr(cStr,n,2))))
   next

   f.close()
   release object f

	use dummy excl
	index on ID tag ID
	use

   if not file("dummy.mdx")
   	return
   endif

// Read language driver ID from dummy.mdx 
	f = new file()
	f.open("dummy.mdx", "R")
	f.seek(0x01F, 0)
	ldID = itoh(asc(f.read(1)), 2)
	f.close()
	drop table dummy

return ldID






